#!/usr/bin/env python
# coding=utf-8
import os, sys
import glob # For filename search with wildcards
import numpy as np
import sapatt as sp # scattering pattern
import sapiawin as sw # presentation windows

'''all_saf2patt.py: Collect all SAXS frames which belong to an experiment in a
    matrix, i.e. a 2D array which makes a mapp of the experiment. The curves
    are stored in a Patt() -- i.e. a 2D scattering image which has been
    designed to store 2D scattering data.

    This script has been written for the visualization of experiments
    with isotropic materials, in which the temperature changes with time.
    (ALBA experiments in 2017). Temperature and time are stored in
    attribute vectors.

SYNTAX:  @@all_saf2patt,<frameset>[''],<variant>[''],ulim[0.25]
        <frameset>: A file description (with wildcards) which specifies a set
                    of data files (extension: ".dat")
        <variant>:  A name prefix which indicates a special pattern usually
                    presenting simply a different range in s
        <ulim>:     The upper s-limit in which the data are presented
EXAMPLE: @@all_saf2patt,nmbsy02*
         @@all_saf2patt,nmbsy02*,sh1,0.1

Continue with ...'''


## Script procedures

def banner():
    # print the name of the running script
    myname = os.path.basename(sys.argv[0])
    #self.poutput("Finished: {!r}".format(myname))
    self.poutput("Finished: {!r}".format(sys.argv[0])) # complete
    pass
#END banner()

def doit():
    '''doit does the work'''
    #argument lister
    #for i, arg in enumerate(sys.argv[1:]):
    #    print("arg {}: {!r}".format(i+1, arg))

    ### Argument parsing
    argus = len( sys.argv ) - 1 # Number of provided arguments

    ## (1) The frame set to be worked on
    frameset = sys.argv[1]

    # If the user has given no extension, then append '.dat'
    test = frameset.split('.')
    if len(test) < 2:
        frameset = frameset+'.dat'

    ## (2) Variant name of the output
    variant = ''
    if argus > 1:
        # Indicate a variant of the usual pattern
        variant = sys.argv[2]

    ## (3) Different upper limit of displayed s
    ulim = 0.25
    if argus >2:
        try:
            ulim = float(sys.argv[3])
        except:
            return 'Error converting upper s-limit "%s"'%sys.argv[3]


    # alphabetically sorted list of the specified files
    allfiles = sorted(glob.glob(frameset))
    if len(allfiles) < 1:
        return 'No files matching "%s"'%frameset
    ### End Argument parsing

    # prepare collection

    ### Start doing the work
    for i, thefile in enumerate(allfiles):
        # NOTICE impossible to call a largo script from within a pyscript. Emulate
        #        emulate it by calls to cmd() or call the methods directly.
        '''
        cmd('read,%s'%thefile)
        cmd('del,1,x0.01')
        cmd('frontex')
        cmd('del,1,,x0.25') # NOTICE data are available up to 0.353/nm
        cmd('stretch,256')
        '''
        thecurve = self.ops[1]
        thecurve.readascii( thefile )   # read
        thecurve.delinterval(None,0.01) # trim front
        thecurve.delinterval(ulim,None) # trim tail
        if thecurve.n < 10:
            return 'ERROR: Trimming deleted all points. Script is not good for WAXS data.'

        # Extrapolate towards 0: 2nd deg., 5 ref.points, no trsf., only parabola
        # thecurve.frontex(2,5,0,True)
        # Option: straight line down into the origin
        thecurve.frontex(0,5,0)
        # Option: power law from 10 points into the origin
        # thecurve.front_powex(10,0)
        thecurve.comb(250) # Stretch the curve to 250 points

        thecurve = self.ops[1]
        elapsed = thecurve.env['elapsed']
        temp = thecurve.env['T'] # Specific for the ALBA experiment
        print('%3i  %6.1f s, %5.1f °C'%(i,elapsed,temp))
        span = thecurve.n # die 128!!!
        yvec = thecurve.y[:span]
        # TODO Interpolate the middle between the values for coherence
        #      with the pixel definition of a pattern (curves start from 0!)
        #      Applies both to yvec and xvec
        if i == 0:
            themap = yvec # scattering data of first frame
            xvec = thecurve.x[:span] # Identical for all frames
            # NOTICE and other initiating tasks ...
            timel = [elapsed] # Starting list of related times
            templ = [temp] # Starting list of related temperatures
        else:
            themap = np.vstack((themap,yvec)) # newer frame below
            timel.extend( [elapsed] ) # newer time at end
            templ.extend( [temp] )
        pass
    # postprocessing
    # NOTICE Eliminate negative values (from extrapolation etc.)
    themap[themap < 0.0] = 0.0

    coll = sp.Patt( themap ) # Generate a simple pattern
    # NOTICE serialization problem of json is solved, if numpy.ndarray are
    #        only handled via the Patt.puta() and Patt.geta() methods.
    # WARNING in Patt._dd['env'] ndarrays are stored as tagged lists.
    timea = np.asarray( timel )
    tempa = np.asarray( templ )
    spant = len( timel )

    deltime = (timel[-1] - timel[0])/(spant-1.0) # Should be 3 for ALBA1711
    dels = (xvec[-1] - xvec[0])/(span-1.0) # delta-s computed the cautious way

    coll.puta('center', (0,0)) # Curves and time start at 0. WARNING not after interpolation!
    coll.vaxras = [dels, deltime, 1] # last: restricted to positive values
    xname, xunits = thecurve.geta('axis1'), thecurve.geta('units1')
    zname, zunits = thecurve.geta('axis2'), thecurve.geta('units2')
    yname, yunits = 'frame', '$~$' # a LaTeX space NOTICE There must always be some string
    coll.vaxvars = [xname, yname, zname]
    coll.vaxunits = [xunits, yunits, zunits]
    # Define special attributes
    coll.puta('elapsed',timea)
    coll.puta('units_elapsed','s')
    coll.puta('T', tempa)
    coll.puta('units_T','°C')
    # NOTICE all environment arrays must be converted to lists
    coll.puta('s', xvec) # The real s-values related to each pixel
    coll.puta('units_s','nm^${{-1}}$')
    # Specify the experimental run
    theset = allfiles[0].split('_')
    stamp = variant + theset[0]

    self.patt = coll # Make available from "ipy"

    coll.puta( 'fn', stamp )
    coll.save( stamp+'.sjs' ) # simple json
    coll.showi(self.pia,wintit=stamp)
    print('Last pattern available as self.patt')
#END doit()

# END script procedures

# ****************** Main *******************************************
# Make a window dictionary, if it does not exist
try:
    self.piawinds
except:
    self.piawinds = {}
self.pia = sw.piawins( self.piawinds )
err = doit()
if err:
    self.poutput( err )
    # No signaling of error to the caller
banner()
