#!/usr/bin/env python
# coding=utf-8
import os, sys
import glob # For filename search with wildcards
import copy # For true copies of variables
import fabio # For the detector image file format IO
import sapatt as sp # For the sapatt image file format
import numpy as np  # For conversion of numerical types
import time         # For time formats and arithmetics
# import h5py  # For access to DESY hdf5 environmental data

'''cbf2sapatt.py: Generation of sapatt-files by conversion of 
                  cbf-files collected at P03 of DESY.


Info on FabIO: ~/python/FabIO/00LIESMICH.txt


SYNTAX: @@cbf2sapatt,<frameset>
EXAMPLE: @@cbf2sapatt,a1_2_*.cbf
    <frameset>: A (set of) image files addressed by the file designation.
                The frameset MUST contain an extension (here: .cbf),
                because fabio discriminates the different formats by their
                extension.
'''

## Script procedures

def banner():
    # print the name of the running script
    myname = os.path.basename(sys.argv[0])
    self.poutput("Finished: {!r}".format(myname))
    pass
#END banner()

def convertfiles():
    '''convertfiles does the work'''
    #argument lister
    #for i, arg in enumerate(sys.argv[1:]):
    #    print("arg {}: {!r}".format(i+1, arg))
    
    ### Argument parsing
    argus = len(sys.argv) - 1
    if argus != 1:
        return "Script argument missing: the (set of) file(s) to be converted"

    ## (1) The set of files to work on
    inframeset = sys.argv[1]
    
    # alphabetically sorted list of the specified files
    infiles = sorted(glob.glob(inframeset))
    if len(infiles) < 1:
        return 'No files found matching "%s"'%inframeset

    # FabIO discriminates file formats by the extension of the file name
    # So the user MUST supply an extension
    test = inframeset.split('.')
    if len(test) < 2:
        return 'No file extension supplied in the input frameset (e.g. ".cbf")'


    ### start the work
    
    for i, thefile in enumerate(infiles):
        thefilelist = thefile.split('.')
        outfilelist = thefilelist
        # The extension of the output file according to sapatt.
        # sjs: scattering pattern in json format.
        outfilelist[len(outfilelist)-1] = 'sjs'
        outfile = '.'.join(outfilelist)
        # print( outfile )
        
        fab = fabio.open( thefile )
        mapp = np.float32( fab.data ) # The pixel matrix

        #WARNING This oversimplified generation is only good for raw data from
        #        a PILATUS detector
        sap = sp.Patt(mapp, (0.172,'mm'))
        
        # Set filename attribute: infile without extension
        err = sap.puta('fn','.'.join(thefilelist[:-1]))
        # Set the center attribute to the center of the pattern
        err = sap.puta('center',[sap.width/2,sap.height/2])
        
        # Access the PILATUS header (text lines).
        # splitlines() returns the list of lines.
        pilhl = fab.header['_array_data.header_contents'].splitlines()
        
        # The fifth line contains the exposure time. Let us check the
        # format of the line in detail in order to detect headers that
        # do not conform to this program. The "exposure time list":
        etl = pilhl[4].split(' ')
        if etl[1] != 'Exposure_time':
            return '%s: Wrong header. "Exposure_time" not in the expected place'%thefile
        if etl[3] != 's':
            return '%s: Wrong header. The unit of Exposure_time is not "s"'%thefile
        # WARNING JSON cannot serialize numpy-floats!
        expo = float(etl[2]) # without further checking
        err = sap.puta('exposed', expo)
        
        # The sixth line contains the cycle time (time between frames).
        # This interval is longer than the exposure.
        etl = pilhl[5].split(' ')
        if etl[1] != 'Exposure_period':
            return '%s: Wrong header. "Exposure_period" not in the expected place'%thefile
        if etl[3] != 's':
            return '%s: Wrong header. The unit of Exposure_period is not "s"'%thefile
        cycl = float(etl[2]) # without further checking
        if cycl < expo:
            return '%s: Wrong header. Cycle time < exposure'%thefile
        # err = sap.puta('exposure_cycle', cycl)        

        # The second line contains the recording time in the format
        # "# 2019-09-29T13:16:24.756"
        timestring = pilhl[1][2:] # "2019-09-29T13:16:24.756"
        toj = time.strptime(timestring,'%Y-%m-%dT%H:%M:%S.%f') # ms are ignored!
        # Handling of the milliseconds
        tss = timestring.split('.')
        ms = int(tss[1])
        #NOTICE Definition of the "time-tuple" to be stored in sapatt
        # A component "day-of-the-year" is added to facilitate easy recognition
        # of experiments running across midnight. All components are integers!
        tt = ( toj.tm_year, toj.tm_mon, toj.tm_mday,
               toj.tm_hour, toj.tm_min, toj.tm_sec, ms, toj.tm_yday )
        err = sap.puta('recorded', tt)
        
        # The pixel area is described in line 3 of the header, but this is
        # already hard coded above in the definition of sap.
        
        # --- Retrieve external environmental data --------------------
        # not done, but below is how such data can be put down in the sapatt:
        '''
        sap.puta('progress_time', progr_t) # in seconds
        sap.puta('progress_force', progr_f) # in arbitrary units
        
        print( progr_f ) #NOTICE log force to screen
        '''
        # --- End retrieval of external environmental data ------------
        
        # Most of the important parameter values of a cbf file are not in
        # attributes in the fabio-file, but in a text header. This text
        # header varies from file to file and from company to company.
        # NOTICE The following statement would simply copy the header into the sapatt
        # err = sap.puta('textheader',fab.header["_array_data.header_contents"])

        err = sap.save( outfile )
               
        err = ''
        #if err != '':
        #    return 'Error reading file "%s" in origset'%thefile
        # Foot of the loop

    #NOTICE on completion
    # self.poutput('bla')
    return ''
#END convertfiles()

# END script procedures

# Main:
err = convertfiles()
if err != '':
    self.perror( err, traceback_war=False)
    # No special treatment in case of error
banner()
