#!/usr/bin/env python
# coding=utf-8
import os, sys
import glob # For filename search with wildcards
import copy # For true copies of variables
import sapatt as sp # For the sapatt image file format
import numpy as np  # For conversion of numerical types

'''sp_wharm2008.py: Preprocess a background corrected WAXS
pattern and generate a harmonized pattern for the special
experiment of August 2020 at P03. Some parameters are
hard-coded and were determined by manual processing.

wharmmask.sjs erases a circle with radius 34 pixels in the 
              center of the pattern.

SYNTAX:   @@sp_wharm2008,<framefile>
EXAMPLE:  @@sp_wharm2008,qa17a_w1up_499

Output: A file with "harm_" prepended in the file name
        Let us call the string before the first underscore
        the "flavor" of the file.
'''

## Script procedures
def argexpand( specs ):
    '''Expand the argument list, if the user has given less than specs
    arguments'''
    result = sys.argv[1:]
    while len( result ) < specs:
        result.append( '' )
    return result
#END argexpand()

def banner():
    # print the name of the running script
    myname = os.path.basename(sys.argv[0])
    self.poutput("Finished: {!r}".format(myname))
    pass
#END banner()

def harmThePattern():
    '''harmThePattern does the work'''
    #argument lister
    #for i, arg in enumerate(sys.argv[1:]):
    #    print("arg {}: {!r}".format(i+1, arg))

    cenw, cenh = 271, 301 # WAXS: Center of primary beam in pixels
    canw, canh = 640, 640 # WAXS: New canvas size of the pattern
    IntensityFactor = 0.001 # Just to obtain numbers easy to plot
    FillSize = 80 # Edge length of central cutout for hole fill
    wavel = 0.1  # wavelength in nm
    dist  = 155 # distance to detector in mm
    # The pixel size in mm is already stored in
    # b.vaxras: [0.172, 0.172, 1]
    
    ### Fetch hard-coded data
    ##  Detektor mask
    maskfile = 'wharmmask.sjs'
    mask = sp.Patt()
    if mask.load( maskfile ):
        return '%s: reading failed'%maskfile    
    
    thefile = sys.argv[1]    
    b = sp.Patt()
    if b.load( thefile ):
        return '%s: Loading error'%thefile
    
    b.logentry('=== Start sp_wharm2008 (Aug 20) ===')
    
    # PILATUS 300K
    # for rotation and the 4 quadrants
    b = b.widen( canw, canh, True ) # absolute values
    # Move the center of the primary beam into the center of the mapp
    b.centerit( cenw, cenh, '' ) # use the values
    # rotangle = 5
    # b.rotate( rotangle ) # set the equator parallel to the mapp
    
    # Kill data inside a small circle around the center
    b.mapp = b.mapp*mask.mapp

    b.harmony() # Fill blind areas from 4-quadrant harmony
    
    # Adapt intensity level
    b.mapp = IntensityFactor*b.mapp
    
    # Turn stretching axis up
    b.rot90()
    
    # Cut to final size
    # b = b.cutwin(1200,1200)
    
    '''
    # Coordinates to scattering vector
    delx, dely, deli = b.vaxras
    delx = delx/(wavel*dist)
    dely = dely/(wavel*dist)
    b.vaxras = [ delx, dely, deli ]
    b.logentry('To "s" with dist = %6.1f mm and lambda = %6.3f nm'%(dist, wavel))
    b.vaxvars =  ["$s_{{12}}$", "$s_3$", "$I$"]
    b.vaxunits[:2] = ["nm$^{{-1}}$", "nm$^{{-1}}$"]
    b.logentry('Axes set to designate fiber symmetry with s_{{12}} horizontal')
    '''

    # Print progress indicator?
    # print('something')
    
    # NOTICE Fill the central hole (xcenter, ycenter, edge length)
    # b.fillhole((cenw,cenh,FillSize))

    # Bookkeeping
    b.logentry('--- sp_wharm2008.py ---')
    # Make the result accessible from ipy and saskia
    self.patt = b
    # Store result in standard scratch file zzz.sjs
    # b.save('zzz')

    # Does the filename start with 'prep_'? - Then remove it
    parts = thefile.split('_')
    if parts[0] == 'prep':
        del parts[0]
        thefile = '_'.join(parts)
    # Store result in unique pattern file
    b.save('harm_'+thefile)

    # Show the result
    #b.showi(self.pia, wintit=thefile, ulev=None, clut='CMRmap')
    #print('***IS ROTANGLE CORRECT???***')
    return ''
#END harmThePattern()

# END script procedures

# Main:
err = harmThePattern()
if err != '':
    self.perror( err, traceback_war=False)
    # No special treatment in case of error
banner()
