#!/usr/bin/env python
# coding=utf-8
# Almut Stribeck almut@stribeck.de
'''
sacurve.py -- define, what is a curve and simple operations on them.
NOTICE: No class may have the same name as the file name that is hosting the classes.
        Simple solution: Capitalize Class names.
'''
from __future__ import print_function
import os # To check files
import copy # To generate true copies of variables instead of generating pointers only
import numpy as np # Well, for standard numerics
import scipy
from scipy.interpolate import UnivariateSpline # Smooth curves by spline
import warnings # Is needed in _make_dy. Is sketched in qinpol
import matplotlib.pyplot as plt # presentation of curves
import six # Differences between PY2 and PY3 ( isstring() )

class CurveClass( object ):
    '''
    Defines a curve together with its environmental parameters and operations on
    them: status report, clean out, read from text file, write to text file

    A curve record has the fields:
        x,
        y : arrays of float32 numbers containing the data columns of the curve
        n : an integer storing the number of points of the curve
        headers : A list of strings which collects general comments related to the curve
        env :     A dictionary which contains environmental parameters and other attributes
                  related to the curve.
        lastfile : A string holding the name of the file in which the newest data of the
                   curve are stored (will be set by writecurve and readcurve)
    '''
    def __init__(self, maxpoints):
        '''
        Initiate curve which may contain a maximum of maxpoints (x,y)-pairs
        '''
        self.x = np.zeros(maxpoints,np.float32)
        self.y = np.zeros(maxpoints,np.float32)
        self._maxpoints = maxpoints
        self.clean()
    #END Constructur __init___

    def clean(self):
        '''
        cleans out the curve; i.e. it erases the header list, sets lastfile="" and n=0
        '''
        self.lastfile = '' # String showing from which file the curve came
        self.headers  = [] # List of strings to hold the header comment lines
        self.env      = {} # Dictionary of environmental parameters and attributes
        self.n = 0 # The actual number of points in the curve
    #END clean()

    def isempty(self):
        '''The curve is empty.
        '''
        return self.n < 1

    def status(self, showhead):
        '''returns a list of strings that inform about the status of the curve
        The complete list of headers is only shown when showhead is not empty.
        '''
        if self.n < 1:
            return ['is empty']
        else:
            result = ['has %i points. It is related to the file %s' %(self.n,self.lastfile)]
            if len(showhead) > 0:
                result.extend( self.headers )
        return result
    #END status()

    def _setdefext(self, fn, defext):
        '''
        Handles the default extension of a curve file name
        '''
        result = fn
        test = fn.split('.')
        if len(test) == 1:
            result = result+'.'+defext
        return result
    #END _setdefext() of CurveClass

#----------------------------------------------------------------------
    def perror(self, errmsg):
        # Error message printing using the same syntax as that of module cmd2
        err = "ERROR: {}\n".format(errmsg)
        sys.stderr.write( err )
    #END perror() of CurveClass

    def isstring( self, item ):
        # work around the missing unicode type in Python3
        if six.PY2:
            return isinstance( item, (str, unicode) )
        else:
            return isinstance( item, str )
    #END isstring() of CurveClass

    def puta( self, key, value ):
        '''Put or replace an attribute in the attribute dictionary
    CurveClass.env  Call: curve.puta( 'myparam', 1000 )'''
        keys = key.strip()
        if self.isstring( keys ):
            self.env[keys] = value
            return False # without error
        else:
            self.perror('Parameter key is no string. Quote, please: "key"!')
            return True
    #END puta() of CurveClass

    def geta( self, key ):
        '''Get an attribute value from the dictionary CurveClass.env
        Call: value = curve.geta('wavelength').'''
        result = None
        if self.isstring(key):
            if key in self.env:
                result = self.env[key]
            else:
                self.perror('Key "%s" does not exist'%key)
            return result
        else:
            self.perror("Key is not a string -- quote it: 'name'!")
    #END geta() of CurveClass

    def hasa( self, key=None ):
        """Is key in the CurveClass attribute dictionary self.env?"""
        if key is None:
            key = ''
        if not self.isstring( key ):
            self.perror("Key is not a string -- quote it: 'name'!")
            return None
        if len(key.strip()) < 1:
            # Dump the whole attribute dictionary to the screen
            print('Attributes and their values:')
            attr = self.env
            for k in attr:
                print('%s  =  %s'%(k,attr[k]))
            return None
        return key in self.env
    #END hasa() of CurveClass

    def dela( self, key ):
        '''Delete the entry of key in the attribute dictionary of the pattern'''
        if not self.isstring(key, string):
            self.perror("Key is not a string -- quote it: 'name'!")
        if key in self.env:
            del(self.env[key])
    #END dela() of CurveClass

    def attribute( self, key, valpar=None, default=None ):
        '''An environmental attribute of the curve is read, set or deleted.
        Checking can simply be done by issueing a read-call after setting an
        attribute. Attributes are stored in the dictionary "self.env"
        This elder method mimics the attribute command, but implements a
        different access to the attributes of a curve as the access implemented
        with the patterns. Therefore I ADDITIONALLY implement the methods
        hasa(), dela() geta() and puta()'''
        if self.isempty():
            return # Do not work on empty curves
        if valpar is None:
            # Return the attribute value, if it exists
            res = default
            if key in self.env:
                res = self.env[key]
            return res
        if valpar == 'delete':
            if key in self.env:
                del(self.env[key])
        else:
            # Set or reset a simple attribute
            value = valpar # Preset: keep the type
            if self.isstring( valpar ):
                # May come directly from a user cmd. For this case:
                # try to convert the string to a float number
                try:
                    value = float( valpar )
                except:
                    # Keep it a string
                    value = valpar
            self.env[key] = value
            return
        pass
    #END attribute() of CurveClass

# START===========Curve I/O=====================================================
    def readascii(self, filename, strip=None, condense=None):
        '''Reads curve data from the ASCII file provided as the argument.
           The result of the function is an error string. An empty string
           indicates: no error. In this case the instance should be filled
           with the curve data from the file.
           strip: If true, all comments will be stripped from the curve header.
           This is reasonable during the first reading of raw data, when all
           environmental parameters are collected separately from log files.
        '''
        err = '' # No error
        if strip is None:
            strip = False
        if condense is None:
            condense = False
        fn = filename
        # First try the name as provided
        if not os.path.isfile(fn):
            fn = fn+'.dat' # Append default extension
        if not os.path.isfile(fn):
            err = 'File %s was not found or is unreadable' %fn
            # Error handling will be done by the calling program
            return err
        else:
            # Read the file
            f = open(fn, 'r')
            # erase the curve memory
            self.clean()
            i = 0
            enter = False # For condense mode
            for line in f:
                line = line.strip()
                if len(line) > 1: # Skip empty lines
                    if line[0] == '#':
                        if line[1] == '{': # Entry for env in python-dict format ...
                            # ... is appended to the dictionary
                            # TODO error checking
                            # NOTICE datetime-records are read as strings in the
                            # default numerical format. If arithmetics shall be done
                            # on them, they must be converted according to
                            #   valstr = self.env[key]
                            #   dto = datetime.strptime(vals,'%Y-%m-%d %H:%M:%S.%f')
                            #   self.env[key] = dto
                            try:
                                item = eval(line[1:])
                            except:
                                err = 'Error reading attribute: {}'.format(line[1:])
                                return err
                            self.env.update( item )
                        else:
                            if not strip:
                                # Do not strip this comment
                                self.headers.append( line )
                    else:
                        if i < len(self.x):
                            column = line.split()
                            if condense:
                                if not enter:
                                    xscr1, yscr1 = np.float32(column[0:2])
                                    enter = not enter
                                else:
                                    xscr2, yscr2 = np.float32(column[0:2])
                                    self.x[i] = 0.5*(xscr1 + xscr2)
                                    self.y[i] = 0.5*(yscr1 + yscr2)
                                    i = i + 1
                                    enter = not enter
                            else:
                                self.x[i], self.y[i] = np.float32(column[0:2])
                                i=i+1
                            pass # The curve array was not yet completely filled
                        if i >= len(self.x):
                            err = 'Curve has too many points. Truncated!'
                        pass  # Entry of a (x,y)-pair
                    pass # Discriminate between extra info and (x,y)-pairs
                pass # ignore empty lines
            f.close()
            self.n = i
            self.lastfile = fn
            return err
    #END readascii()

    def writeascii(self, filename, force):
        '''Writes the curve into an ASCII file and does some bookkeeping:
             (1) Appends a header comment line
                   # Origin: <filename>
             (2) (only after success:) replaces self.lastfile = <filename>
        '''
        if self.n <= 0:
            return 'Curve array is empty'
        fn = self._setdefext(filename, 'dat')
        isnew = not os.path.isfile(fn)
        if isnew or force:
            # NOTICE remove hash to save the originating file in a header comment
            # self.headers.append('# Origin: %s' %self.lastfile)
            # write into the file
            f = open(fn, 'w')
            # Write the headers
            for i in range(len(self.headers)):
                f.write( self.headers[i]+'\n')
            # Write the environment dictionary
            # item by item to the ASCII file
            for key in self.env:
                keystr = str(key).strip()
                value  = self.env[key]
                valstr = str(self.env[key]).strip()
                if self.isstring( value ):
                    # encapsulate the string in quotation marks
                    valstr = '"'+valstr+'"'
                line = '#{"'+keystr+'": '+valstr+'}\n'
                f.write( line )
            # Write the curve data to the ASCII file
            for i in range(self.n):
                f.write('%14.7e %14.7e\n' %(self.x[i],self.y[i]))
            f.close()
            self.lastfile = fn
            return '' #No error
        else:
            return 'File %s exists and overwriting is not forced' %fn
    #END writeascii()
# END===========Curve I/O=====================================================

# START==============Deletion and insertion of curve points===================
    def delirange(self, i1, i2):
        '''
        Delete index range, i.e. the points between i1 and i2 including both i1 and i2
        '''
        # Normalize the indices
        if self.n == 0:
            return
        n1 = max(i1, 0)
        n2 = min(i2, self.n-1)
        # xl = self.x[:n1] # all points which remain below
        xh = self.x[n2+1:] # all points which remain above
        yh = self.y[n2+1:] # all points which remain above
        # Copy shifted
        ll = len( xh )
        self.x[n1:n1+ll] = xh
        self.y[n1:n1+ll] = yh
        # Bookkeeping -- no need to zero the invalid data above
        self.n = self.n - (n2 - n1 +1)
        if self.n == 0:
            # If all points are gone, erase also all metadata
            self.clean()
    #END delirange()

    def delinterval(self, lox=None, hix=None):
        '''
        Delete all points for which the x-values are between lox and hix including
        lox and hix. If lox is None, the deletion starts from the beginning of the
        curve. If hix is None, the deletion extends up to the end of the curve.
        '''
        if self.isempty():
            return 'Curve array is empty'
        i1 = 0
        if lox is not None:
            while (self.x[i1] < lox) and (i1 < self.n-1):
                i1 = i1 +1
        if i1 == self.n-1:
            # Nothing to do: Deletion interval starts above the last point in the curve
            return
        if hix is None:
            i2 = self.n - 1
        else:
            i2 = i1
            while (self.x[i2] < hix) and (i2 < self.n-1):
                i2 = i2 + 1
        if (i2 == 0) and (self.x[0] > hix):
            # Nothing to do: Deletion interval ends below first point in the curve
            return
        self.delirange( i1, i2 )
        pass
    #END delinterval()

    def insert(self, xp, yp):
        '''Insert a point with the coordinates (xp, yp) into the curve'''
        if self.isempty():
            self.n = 1
            self.x[0] = xp
            self.y[0] = yp
            self.lastfile = '00nofile.dat'
            return
        # find the insertion position
        span = self.n
        nn = self._catchpos( xp )
        if self.x[nn] == xp:
            self.y[nn] = yp # Overwrite old point on position
            return
        if self.x[nn] < xp:
            nn = nn+1 # nn may become n --> simple append
            pass
        # Point to be inserted must be put at index nn
        if nn < span:
            # make space for the point
            upx = self.x[nn:span]
            self.x[nn+1:span+1] = upx
            upy = self.y[nn:span]
            self.y[nn+1:span+1] = upy
        # insert the new point
        self.x[nn] = xp
        self.y[nn] = yp
        # bookkeeping
        self.n = span + 1
    #END insert()
# END==================Deletion and insertion of curve points==========================

# START ===============Interpolation and smoothing=====================================
    def qinpol(self, xsearch, si):
        '''
        Determine the value of the curve at the position xsearch by a quadratic interpolation
        on the surrounding points and start at searching at the start index si. Because this
        business is normally carried out "from left to right" a variable si helps to shorten
        the search.
        '''
        ei = self.n # search must stop at the end of the curve
        # Move a window of 5 points upwards, until xsearch is in its center
        while (xsearch > self.x[si+2]) and (si+4 < ei):
            si = si+1
        # Cut out the window
        xwin = self.x[si:si+5]
        ywin = self.y[si:si+5]

        ## Something general: Many well-known solutions of mathematical problems
        ## can only dwell in digital space, if they are encapsulated in a
        ## "pressure cabin" which guarantees conditions under which they can
        ## survive. One of the most important pressure cabins is, to map the
        ## input data on the central interval of digital floating point numbers
        ## with exponent 0. This is the interval [0.5 .. 1). Doing so avoids
        ## accumulation of rounding errors and, thus, ill-conditioned matrices
        ## by supplying the algorithm with input data that exploit the complete
        ## range of the floating-point binary mantissa. Simultaneously, the
        ## binary exponent of 0  provides maximum dynamics in the directions of
        ## both big and small numbers in digital representation.

        # Prepare mapping to [0.5 .. 1]
        zf = 0.5
        xlo  = min(xwin)
        xlen = max(xwin) - xlo
        ylo  = min(ywin)
        ylen = max(ywin) - ylo
        # map intervals and search point
        xwinm = (xwin - xlo)*zf/xlen + zf
        ywinm = (ywin - ylo)*zf/ylen + zf
        xseam = (xsearch - xlo)*zf/xlen + zf

        # Carry out a quadratic regression and return the coefficients
        quadratic = 2
        # with warnings.catch_warnings():
        if True: # Just to permit keeping the indentation
        # In the unlikely event of pressure loss, warnings will thrown on you
        # Then firmly  pull the hash-sign dangling from the "with" phrase above
        # down towards the "if" clause and try to find better living conditions.
            warnings.simplefilter("ignore")
            coeff = np.polyfit( xwinm, ywinm, quadratic ) # Get the coeff vector, hi first
            p = np.poly1d(coeff) # encapsulate for numerical operations on polynomials
            res = p(xseam) # compute the polynomial at the position xsearch
        # inverse mapping of resulting ordinate
        res = (res - zf)*ylen/zf + ylo
        return res # ... and return also si by Python convention, if it is a variable
    #END quinpol()

    def _condition(self):
        '''Returns a conditioned version of the input curve and the corresponding
        mapping parameters. This is the first step of smoothing of a curve
        '''
        # Set up the mapping parameters (for reasoning cf. qinpol)
        span = self.n
        zf = 0.5
        xlo  = min(self.x[:span])
        xlen = max(self.x[:span]) - xlo
        ylo  = min(self.y[:span])
        ylen = max(self.y[:span]) - ylo
        # The conditioned curve is identical.
        # Only the x and y arrays are mapped
        conditioned = copy.deepcopy(self)
        conditioned.x = (self.x - xlo)*zf/xlen + zf
        conditioned.y = (self.y - ylo)*zf/ylen + zf
        # Make dictionary of the mapping
        condict = {'xlo': xlo, 'xlen': xlen, 'ylo': ylo, 'ylen': ylen}
        return conditioned, condict
    #END _condition()

    def _recondit( self, number, xmin, xlen ):
        '''Recondition a single number inverting the 1D conditioning described
        by xmin, xlen'''
        res = xmin + (number - 0.5) * xlen / 0.5
        return res
    #END _recondit()

    def _hasinfelems( self, arr ):
        ''' Returns True, if at least one of the elements of arr is either
        not a number or +-infinite'''
        #TODO This routine may perhaps become unnecessary
        test = np.isfinite( arr )
        return min(test) == False # At least 1 is infinite or nan
    #END _hasinfelems()

    def _recondition(self, con, cd):
        '''Recondition self using the mapping described by the dictionary cd
        and store the result into self. Environmental variables of self are
        not touched, but the number of points is.'''
        span = con.n
        self.x[:span] = cd['xlo'] + (con.x[:span] - 0.5) * cd['xlen'] / 0.5
        self.y[:span] = cd['ylo'] + (con.y[:span] - 0.5) * cd['ylen'] / 0.5
        self.n = span
    #END _recondition()

    def _make_dy(self):
        '''Estimate the error of each data point by a measure of correlation
        with respect to its neighbors. The result is an array dy which adds
        to the instance. This is no drawback, because the original curve has
        already been copied and transformed (Fl subtraction, spower). -- and
        this transformation must be reverted, anyway.'''
        span = self.n
        self.dy = np.zeros(self._maxpoints,np.float32)
        # Boxcar average by convolution with a box of integral 1 and length "box"
        box = 3
        ave = np.convolve(self.y[:span], np.ones((box,))/box, mode='same')
        # Deviations from these averages are taken as a measure of error:
        #  A straight line is drawn through the points i+1 and i+3 and the deviations
        # of the data points from these lines are taken as a measure of error and
        # accumulated into self.dy

        for i in range(span-4):
            # Compute the equation of the line
            m = (ave[i+3] - ave[i+1])/(self.x[i+3] - self.x[i+1])
            b = ave[i+1] - m*self.x[i+1]
            for j in range(i,i+5):
                dis = abs(m*self.x[j] + b - self.y[j])
                #TODO If this is insufficient, we can introduce an eps parameter
                #     to mark outliers, for which the error dis may be doubled.
                self.dy[j] += dis # add dis into the correct slot of dy
                pass
            pass
        # Normalize to the frequency by which errors have been registered
        # in each slot
        for i in range(span):
            freq = min(span-i,min(i+1,5))
            # softweight weights errors more moderately around the borders
            # and corrects for underweighting (factor 1.5)
            softweight = 1.5*(5.0+freq)/(10.0*freq)
            self.dy[i] = self.dy[i]*softweight
            pass
        #NOTICE identical to TOPAS version
        return # The result is in self.dy
    #END _make_dy()

    def _spline( self, sp, points ):
        '''self._spline returns the smoothed curve and its 1st derivative.
        It is fed with the user-friendly smoothing parameter sp (natural
        smoothing at sp approx. 1).
        On return, the derivative is in self.dy[:self.n]
        Parameters: <sp>: smoothing parameter. "Natural smoothing" is obtained
                          for <sp> around 1.0
                    <points>: If > 2, then the support of the curve will be
                              convered by <points> equidistantly spaced data
                              points.'''
        if (points > self._maxpoints):
            print('+++ spline: Generation of %i > %i points impossible'%(points,self._maxpoints))
            points = self._maxpoints
        span = self.n
        # Condition the data into the work curve and remember the
        # condition parameters in condict
        work, condict = self._condition()
        work._make_dy() # Estimate the errors of each point
        # Convert the errors to weights
        # Make result curve: still conditioned and with the dy's
        # later to be used for the derivative
        smoothie = copy.deepcopy( work )
        with warnings.catch_warnings():
            # catch division-by-zero warnings
            warnings.simplefilter("ignore")
            work.dy[:span] = 1.0/work.dy[:span]
        # Has there been a division by zero?
        goodies = np.isfinite(work.dy[:span])
        # Set infinite weights first to zero
        work.dy[np.where(goodies==False)] = 0.0
        # .. and finally to the maximum weight of the others
        work.dy[np.where(goodies==False)] = max(work.dy[:span])
        spi = 1.0*sp*self.n # Convert from user-friendly smoothing parameter
        # Compute the smoothing cubic spline
        smo = UnivariateSpline(work.x[:span], work.y[:span], w=work.dy[:span], k=3, s=spi)
        # Compute the derivative
        smod = smo.derivative()
        if points > 2:
            # Generate vector of equidistant points
            n = points
            xs = np.linspace(work.x[0],work.x[span-1],n)
        else:
            n = span # Number of points in the smoothed curve
            xs = work.x[:n]
        # make the smooth y-values
        ys = smo(xs)
        work.x[:n] = xs[:n]
        work.y[:n] = ys[:n]
        work.n = n
        # recondition work into smoothie using condict
        smoothie._recondition( work, condict )
        # Recondition the derivative
        yss = smod(xs)
        smoothie.dy[:n] = yss[:n]*condict['ylen']/condict['xlen']
        return smoothie
    #END _spline()

    def average(self, overp):
        '''Averages the first "overp" points in the curve, if "overp" is positive.
        For negative "overp", the average is perfomed over the last "overp"
        points. This method helps to define a constant region in a noisy
        curve.'''
        overp = int(overp)
        if overp == 0:
            return # The average over 0 points is the original curve
        span = self.n
        if span == 0:
            return # The average of an empty curve remains the empty curve
        pts = abs(overp)
        sign = overp/pts
        if abs(overp) > span:
            overp = sign*span # Not more than all valid points can be averaged
            pts = span
        yvalid = self.y[:span]
        if overp > 0:
            yset = yvalid[:overp] # first overp points
        else:
            yset = yvalid[overp:] # last overp points
        av = np.average(yset)
        yset = np.ones(pts)*av # Vector with pts-times the average
        if overp > 0:
            yvalid[:overp] = yset # replace first overp points
        else:
            yvalid[overp:] = yset # replace last overp points
        # Store back into self
        self.y[:span] = yvalid
        return
    #END average()

# END =================Interpolation and smoothing=================================

# START ===============Helper operations===============================================
    def _hasenoughpowerof2points(self):
        '''is true, if the curve has a number of points (self.n) which is a power of 2 (2**n)'''
        minpts = 32
        res = False
        testpower = minpts
        res = (self.n == testpower)
        while (not res) and (testpower < 2048):
            testpower = testpower << 1 # Shift left, i.e. multiply by 2
            res = (self.n == testpower)
        return res
    #END _hasenoughpowerof2points

    def _catchpos( self, xval ):
        ''' return the index of the position in x which is closest to xval
            was GREIFSTELLE in TOPAS
        '''
        # Vector operation for horizontal distances
        dist = abs( self.x[0:self.n] - xval )
        # grab index of minimum in array dist
        gri  = np.argmin( dist )
        return gri
    #End _catchpos

    def _catchpoint( self, xval, yval ):
        ''' return the index of the curve point in cur
        which is closest to the point (xval, yval)
            was GREIFPUNKT in TOPAS
        '''
        span=self.n
        # Vector operations for metrics
        xxv = (self.x[:span] - xval)**2
        yyv = (self.y[:span] - yval)**2
        metrics = xxv + yyv # Square of the distance
        # Grab index of the minimum in array metrics
        gri = np.argmin( metrics )
        return gri
    #End _catchpoint


# END =================Helper operations===============================================

# START ====================== One-Curve operations====================================
    def subcy(self, value):
        '''
        Subtracts a constant from all valid y values
        '''
        span = self.n # Only from the valid points
        self.y[:span] = self.y[:span] - value
    #END subcy()

    def subcx(self, value):
        '''
        Subtracts a constant from all valid x values
        '''
        span = self.n # Only from the valid points
        self.x[:span] = self.x[:span] - value
    #END subcy()

    def multiply(self, factor, ypower, xpower):
        '''
        Transform the curve according to the equations
          y = y*factor * x**ypower
          x = x**xpower
        '''
        span = self.n # Only the valid points
        # Transformation of y-values (1st equation)
        self.y[:span] = self.y[:span]*factor
        if ypower < 0:
            # Cut off the negative branch
            self.delinterval(None,0.0)
            if span != self.n:
                dp = span - self.n
                print('The first %i points have been deleted to enable the transformation.'%dp)
            span = self.n # Only the valid points
        if ypower != 0.0: # Nothing to to for ypower = 0.0
            self.y[:span] = self.y[:span] * self.x[:span]**ypower
        # Transformation of x values (2nd equation)
        if xpower != 0.0:
            self.x[:span] = self.x[:span]**xpower
    #END multiply()

    def morph( self, width, task=None ):
        '''Perform a morphological operator on the curve. Default: Smoothing by
        a median filter.
        Example: thecurve.morph(7) smoothes the curve by application of a
                     median filter extending over 7 data points.
                 thecurve.morph(7,'erode') replaces by the lowest intensity
                     in a sliding box of length 7
            <width>: a number defining the width ww of the sliding box, with
                     ww = round( width ). In case of an even box length and
                     median, a width < ww makes the median more erosive,
                     whereas a width > ww makes the median more dilative.
            <task>:  "erode":  pick the lowest value from the box
                     "median": pick the central value from the box list (default)
                     "dilate": pick the highest value from the box.'''
        hw = int(width/2.0)     # The center of the list
        ww = int(width + 0.5)   # The rounded effective width of the box
        # NOTICE These definitions permit control of the "center", if the
        #        effective ww is an even number: For width < ww the center
        #        is considered the lower of the two possibilities (erosive),
        #        for width > ww the center is considered higher (dilative)

        if task == 'erode':
            pick = 0
        elif task == 'dilate':
            pick = ww-1
        else:
            pick = hw # median filter is default

        span = self.n
        ycop = copy.copy(self.y[:span]) # because y is overwritten
        for i in range(span-ww):
            box = copy.copy(ycop[i:i+ww])
            box = np.sort(box)
            self.y[i+hw] = box[pick]
        pass
    #END morph() of CurveClass

    def closing(self, width):
        '''Morphological operator of the second kind: close "holes"
        which are smaller than width by first dilating the mask curve
        and second eroding.
        '''
        self.morph(width,'dilate')
        self.morph(width,'erode')
        pass
    #END closing() of CurveClass

    def opening(self, width):
        '''Morphological operator of the second kind: open tiny
        spots to a minimum "width" by first eroding the mask curve
        and second dilating.
        '''
        self.morph(width,'erode')
        self.morph(width,'dilate')
        pass
    #END opening() of CurveClass

    def similarto( self, other, threshold=None, fringe=None, minextent=None ):
        '''Returns a mask curve which is 1 where this and the other curve are
    similar and 0 where they are not The curves must have the same number
    of points and must be rasterized on the same x-values. The latter is
    not tested.
    other :    The other curve
    threshold: The relative deviation abs((self.y - other.y)/other.y)
               which is considered the threshold between similar and
               not similar.
    fringe:    An integer boxlen which determines the width of the range
               around a region of bad values which is considered to be
               influenced by the artefact in self (or in other)
    minextent: The minimum number of dissimilarities in succession which
               render the defect considerable
    DEFAULTS: threshold = 1.0 (i.e. 100% deviation)
              fringe    = 100
              minextent = 3
        '''
        # The parameters
        if threshold is None:
            threshold = 1.0
        threshold = abs( threshold )
        if fringe is None:
            fringe = 100
        fringe = int( fringe )
        if minextent is None:
            minextent = 3
        minextent = int( minextent )

        result = copy.deepcopy(self)
        n = self.n
        if n != other.n:
            # Curves of different length are not similar at all
            result.y[0:n] = np.zeros(n)
            return result
        if self.n < 2:
            # Do not work on single points or empty curves
            return result

        # result.y[0:n] = np.ones(n) # consider all points similar

        divi = other.y[:n]
        divi[divi<=0] = 0.0 # Prohibit division by zero
        rc = abs((self.y[:n] - other.y[:n])/divi)
        w = np.ones(n) # consider all points similar
        w[rc > threshold] = 0.0 # mark the invalid ones
        # copy to a "curve" for morph to work
        result.y[:n] = w
        if minextent > 0:
            # Consider tiny discrepancies tolerable
            result.closing( minextent )
        if fringe > 0:
            # extend the relevant discrepancies by fringe points
            result.morph( fringe, 'erode' )
        return result
    #END similarto() of CurveClass

    def bridgewith( self, mask, smp=None ):
        '''Replace bad points (where the mask is 0) by points interpolated
        from the rest of the curve.'''
        if smp is None:
            smp = 0.1
        n = self.n
        if n != mask.n:
            return True # with error

        my = copy.copy(mask.y[:n])
        mh = max(my)
        if mh <= 0:
            mh = 1.0
        my = my/mh
        m = int(sum(my)) # Good values
        pro = 100.0*m/n
        if pro < 50.0:
            print('Only %5.1f%% of the %i points are good. No bridging.'%(pro,n))
            return True # with error

        # Make a tuple of indices where good points are
        goodidx = np.where( mask.y[:n] > 0 )

        # Make feed arrays contain only the good points
        fedx = self.x[goodidx]
        fedy = self.y[goodidx]
        spi = smp*m # For the spline "m" is the natural parameter
        smo = UnivariateSpline(fedx, fedy, k=3, s=spi)
        # make and insert the smooth y-values
        self.y[:n] = smo(self.x[:n])
        return False
    #END bridgewith() of CurveClass

    def resample( self, template, smp=None ):
        '''Resample the curve so that it becomes defined at exactly the
        same points as is the curve in the template. smp is an optional
        smoothing parameter which normally is very small (interpolation)'''
        if smp is None:
            smp = 0.1
        n = self.n
        t = template.n
        if (n < 4) or (t < 4):
            return 'Curves too short for resampling'

        # We need at least some overlap
        mint = min(self.x[0], template.x[0])
        mino = max(self.x[0], template.x[0])
        maxo = min(self.x[n-1], template.x[t-1])
        maxt = max(self.x[n-1], template.x[t-1])
        if maxo <= mino:
            return 'Curves do not overlap. No resampling.'
        if (maxo-mino)/(maxt-mint) < 0.5:
            return 'Overlap < 50%. No resampling'

        fedx = self.x[:n]
        fedy = self.y[:n]
        spi = smp*t # For the spline "m" is the natural parameter
        smo = UnivariateSpline(fedx, fedy, k=3, s=spi)
        # make and insert the smooth y-values
        self.y[:t] = smo(template.x[:t])
        self.x[:t] = template.x[:t]
        self.n = t
        return ''
    #END resample() of CurveClass

    def comb( self, pts ):
        '''make the curve a comb function with pts points, starting from 0
        and extending to its present end at equidistantly spaced x-values.
        Externally the comb operator is realized by the stretch command.
        '''
        if pts > self._maxpoints:
            pts = self._maxpoints
        if self.x[0] < 0.0:
            # Remove a negative branch
            self.delinterval( None, 0.0 )
        if self.isempty():
            return True # with error
        if self.x[0] > 0.0: ##?
            self.insert(0.0,0.0) # insert the origin
        span = self.n
        xlen = self.x[span-1]
        xdel = xlen/(pts-1)
        # Interpolate
        newx = np.arange(pts)*xdel
        self.y[:pts] = np.interp(newx,self.x[:span],self.y[:span])
        self.x[:pts] = newx
        # Bookkeeping
        self.n = pts
        return False # without error
    #END comb() of CurveClass

    def ext15( self, pdeg, pts ):
        '''Extrapolate the curve tail using a polynomial of degree pdeg computed from the
        last pts points in the curve. Considering a curve which describes
        isotropic data this is analogous to the extrapolation "into the
        corners of an image". pdeg is the degree of the polynomial and pts
        is the number of points to use for the extrapolation.
        Special case: If pdeg==1 and pts==0, then the new endpoint is on
        the x-axis. It is connected to the last point of the curve by a
        straight line.'''
        span = self.n
        xbeg = self.x[span-1]
        xend = xbeg*1.5
        xdel = xbeg - self.x[span-2]
        xact = xbeg + xdel

        straight = (pdeg == 1) and (pts == 0)
        if not straight:
            thex = self.x[span-pts:span]
            they = self.y[span-pts:span]

            cond = max(they) # Mapping factor
            they = they / cond # relax a possibly ill-posed numerical problem
            a = np.polyfit( thex, they, pdeg ) # The poly-coefficients in the mapped world
            p = np.poly1d(a) # The polynomial shaped for extrapolation
            while xact <= xend:
                self.x[span] = xact
                self.y[span] = p(xact)*cond
                xact = xact + xdel
                span = span + 1
        else:
            # Connect to zero by a straight line
            ybeg = self.y[span-1]
            m = ybeg/(xbeg - xend)
            while xact <= xend:
                self.x[span] = xact
                self.y[span] = ybeg - m*(xbeg - xact)
                xact = xact + xdel
                span = span + 1
        self.n = span
        pass
    #END ext15()

    def frontex( self, pdeg, pts, spower, parabolic=None ):
        '''Extrapolate the curve head using a polynomial of degree pdeg computed from the
        first pts points in the curve. spower describes a transformation of the curve
                      I'(s) = I(s) * s**spower
        to be performed before the extrapolation and to be reversed after. If spower is
        not 0, the point at x=0 is approximated by the point at xstep/10, which is 1/10
        of the step width in order to be able to revert the transformation.
        For pdeg = 0 the origin is connected to the first point by a straight line.
        If parabolic==True, the fit function is a simple parabola with
                      f(s) = A * s**pdeg + B. '''
        span = self.n
        thex = self.x[:pts]
        they = self.y[:pts]
        xbeg = thex[0]
        delx = thex[1]-xbeg
        the0 = 0.1*delx
        newp = int(thex[0]/delx + 0.5)

        spow = min( int(abs(spower)), 4 ) # Permit only integers between 0 and 4
        transy = spow > 0
        if transy:
            they = they * thex**spow
        if parabolic:
            para = pdeg       # Remember the order of the parabola
            thex = thex**pdeg # parabolic stretch
            pdeg = 1          # mapped to a linear problem
        cond = np.longfloat(max(they)) # Mapping factor
        they = they / cond # relax a possibly ill-posed numerical problem
        if pdeg == 0:
            # Special treatment: straight connection line
            a = [ they[0]/thex[0], 0.0 ] # slope and starting height
        else:
            a = np.polyfit( thex, they, pdeg ) # The poly-coefficients in the mapped world
        p = np.poly1d(a) # The polynomial shaped for extrapolation
        xact = np.longfloat(xbeg - delx)
        while xact > 0.0:
            xpol = xact # position in the polynom
            if parabolic:
                xpol = xpol**para
            yact = p(xpol)*cond # value from polynomial re-conditioned
            if transy:
                yact = yact / xact**spow # revert transformation
            self.insert( xact, yact ) # insert point into curve
            xact = xact - delx # next point
        if  self.x[0] > 0.0:
            if transy:
                # close to, but not really zero
                xact = 0.1*delx
            else:
                xact = 0.0
            xpol = xact # position in the polynom
            if parabolic:
                xpol = xpol**para
            yact = p(xpol)*cond # value from polynomial re-conditioned
            if transy:
                yact = yact / xact**spow # revert transformation
            self.insert( 0.0, yact ) # Insert this always at x=0
        pass
    #END frontex()

    def front_pow( self, fp, ap=None ):
        '''Helper for a front extrapolation: take the first fp points of the
        curve and fit the function y = a*x**p. This function will definitively
        start at the origin. The problem is linearized by taking the logarithm.
        If a value for the Porod asymptote ap is given, the extrapolation is
        also good for the extrapolation of interference functions which are
        already shifted down by ap. Then the extrapolation does not touch
        (0.0, 0.0), but (0.0, -ap).'''

        #WARNING without conditioning the input data. Anyway, only limited
        #        conditioning in the interval [0..0.5] were possible
        #WARNING If self.x[0] is not positive, extrapolation is impossible
        if ap is None:
            ap = 0.0
        mat = np.zeros(5,np.float64) # The matrix elements
        d = np.zeros(3,np.float64) # The determinants
        mat[0] = fp
        for i in range(fp):
            xact = np.log( self.x[i] )
            yact = np.log( self.y[i] + ap )
            mat[1] += xact
            mat[2] += yact
            mat[3] += xact**2
            mat[4] += xact*yact
        # Determinants and linear equation
        d[0] = mat[0]*mat[3] - mat[1]**2
        d[1] = mat[2]*mat[3] - mat[4]*mat[1]
        d[2] = mat[0]*mat[4] - mat[1]*mat[2]
        a = d[1]/d[0]
        p = d[2]/d[0]
        # undo logarithm
        a = np.exp( a )
        return a, p
    #END front_pow() of CurveClass

    def front_powex( self, fp, ap ):
        '''Carry out the power-law front extrapolation of a curve'''
        # Compute the parameters of the power law by regression over
        # the first fp points
        a, p = self.front_pow( fp, ap )
        # Prepare for looping
        xbeg = self.x[0]
        delx = self.x[1] - xbeg
        xact = np.longfloat(xbeg - delx)
        while xact > 0.0:
            yact = a*xact**p - ap
            self.insert( xact, yact )
            xact = xact - delx # next point
        if self.x[0] > 0.0:
            # Last inserted point was still positive, so insert the origin
            self.insert( 0.0, -ap ) # Insert this always at x=0
        # correct for rounding error
        self.y[0] = 0.0
    #END front_powex() of CurveClass

    def integral( self ):
        '''The integral of the curve between self.x[0] and self.x[self.n-1]
        is returned. This is required for the computation of scattering power.
        It is not necessary to have the curve equidistantly spaced.'''
        gral = np.float64(0.0)
        span = self.n
        if span > 0:
            for i in range(span-1):
                # i iterates all delta-x. Safe and slow algorithm.
                dx = np.float64( self.x[i+1] - self.x[i] )
                y2 = np.float64( self.y[i+1] + self.y[i] ) # double y
                gral += y2*dx
            gral = gral/2.0
        return gral
    #END integral() of CurveClass

#END====================One curve operations=======================================

# START=================Two curve operations=======================================
    def addcurve(self, asecond, factor ):
        '''
        Adds a second curve multiplied by factor to the present curve
        '''
        err  = '' # No error
        # Existence checks must have been performed
        # Compute the common support of both curves
        suppmin = max( self.x[0], asecond.x[0] )  # The first elements
        n = self.n-1
        n2 = asecond.n-1
        suppmax = min(self.x[n], asecond.x[n2] ) # The last VALID elements
        if suppmin >= suppmax:
            return 'Add/subtract curves: The curves do not overlap'
        if suppmin > self.x[0]:
            print('Warning: Missing overlap on the left: %9.3e .. %9.3e'%(self.x[0],suppmin))
        if suppmax < self.x[n]:
            print('Warning: Missing overlap on the right: %9.3e .. %9.3e'%(suppmax,self.x[-1]))
        # Check, if both curves sampled almost identically
        if n2 == n:
            if (asecond.x[0] == self.x[0]) and (asecond.x[n-1] == self.x[n-1]):
                # The fast method: similar sampling of both curves
                self.y[:n] = self.y[:n] + factor*asecond.y[:n]
                return
        startat = 0 # The index in asecond to start with interpolation
        # loop all points in self
        for i in range(self.n):
            yestimate = asecond.qinpol(self.x[i], startat)
            # Carry out the addition
            self.y[i] = self.y[i] + factor*yestimate
        pass
    #END addcurve()

    def lowpass(self, thesource, cutoff, order, cleanme=None):
        '''Generate a lowpass filtered curve from the curve in thesource. The applied
    filter is a Butterworth filter with a cutoff frequency cutoff/256 and order order.
    The number of points in thesource must be a power of 2.
    Concerning the construction of other common filters see, e.g.
    PV-WAVE 8.5 User's Guide, p. 145-146'''
        span = thesource.n
        if (not thesource._hasenoughpowerof2points()):
            print('Source curve with %i points must have 2**n points [32 .. 2048]'%span)
            return
        # thesource.y is the right branch of an even and equidistantly
        # spaced function
        # We work with both the positive and the negative branch
        work = np.zeros(2*span)
        # Some condidtioning of the user data
        # Ease the computational problem from poorly conditioned user data
        factor = 1.0*span / max(thesource.y[:span])
        work[span:2*span] = factor * thesource.y[0:span] # Right branch is is easy
        # NOTICE Older numpy versions do not have the flip method, so we do it by hand
        span1 = thesource.n - 1
        for i in range(span):
            work[i] = work[2*span1-i]
        # Enter the reciprocal space
        krow = scipy.fft(work)
        # Generate the Butterworth low-pass filter
        cutchan = cutoff*span/256.0 # Normalize cutoff for curves with different span
        dist = np.zeros(2*span) # becomes the PV-WAVE, IDL, matlab ... "dist function" in 1D
        dist[:span]       = np.linspace(0,span-1,span) # left branch increasing
        dist[span:2*span] = np.linspace(span-1,0,span) # right branch decreasing
        buttlowpassfilt = 1.0 / (1.0 + (dist/cutchan)**(2*order))
        # Apply it to krow
        krow = krow*buttlowpassfilt
        # Back to space of origin
        work = scipy.ifft( krow )
        # Due to computational inaccuracy there are tiny imaginary parts
        # which can be ignored
        work = np.real( work ) / factor  # ... and redo the mapping
        # Now fill self, i.e. the result
        if cleanme:
            self.clean()
        self.y[:span] = work[span:2*span]
        self.x[:span] = thesource.x[:span]
        self.n = span
        pass
    #END lowpass()

    def abel( self ):
        '''Carry out the Abel transformation on the isotropic curve. This operation
    projects the isotropic function in n dimensions on an (n-1)-dimensional
    subspace. In "the old days" of scattering this operation was known as
    "slit smearing". For inversion one would probably have a look on
    the module PyAbel or my work concerning computed tomography of the
    SAXS of fibers. All input checking must be performed by the caller.
    Literature: Bracewell, R. N., The Fourier Transform and Its Applications,
    3rd ed, (2000), McGraw-Hill, Boston, chap. 13, "The Abel Transform",
    p. 351-356.
    Procedure has been tested with input: J0(2*r) and output: cos(2*x).
    Choose many points in Interval r=[0,300] because J0() decreases
    very slowly.'''
        span = self.n
        res = copy.deepcopy( self ) # Transfer everything
        res.y = np.zeros( len(res.x) ) # Erase all y-data
        dr = self.x[span-1]/(span-1) # average step size
        for Yi in range(span-1): # over all Y-indices (x**2 + y**2 = r**2
            Y = res.x[Yi] # The place at which the projected intensity is computed
            accu = np.longfloat( 0.0 )
            YY = Y*Y
            for Ri in range(Yi+1,span):
                R = self.x[Ri]
                DN = np.sqrt(R*R - YY)
                # a = a + f(r)*r/sqrt(r**2 - y**2)
                accu = accu + self.y[Ri]*R / DN
                pass
            accu = 2.0*accu*dr
            res.y[Yi] = accu
        return res
        pass
    #END abel()

class SmootherClass( object ):
    '''This class is only necessary, if it becomes necessary to smooth
    a curve in several sets. The state of the former work on the corresponding
    commands is documented in the file 00smoothcmds.sav.py'''

    def __init__( self ):
        self.setbounds = [] # Empty list of set boundaries
        self._reset()
    #END __init__ SmootherClass

    def _reset( self ):
        '''Reset the SmootherClass'''
        self.curve = None # no curve in smoothing memory
        self.newp = 0 # No generation of equidistantly spaced points
        self.raster = None # No linearly spaced raster of x-values available
        self.fl = 0.0 # No fluctuation bg subtracted
        self.spower = 0 # y-values not transformed
        # self.setbounds, the lis of set boundaries shall persist.
        pass
    #END _reset() SmootherClass

    def loadcurve( self, cur, newp, bg, spow ):
        '''Load cur into the smoother, transform it according to bg and spow,
        and prepare for generation of equidistantly spaced data points'''
        self.curve = copy.deepcopy( cur )
        n = cur.n
        if newp > 2:
            self.newp = newp
            self.raster = np.linspace(cur.x[0], cur.x[n-1], newp)
            print('Smooth curve will contain %i equidistantly spaced points.'%newp)
        self.fl = bg
        self.spower = spow
        print('Transformation for smoothing: y = (y - %f) * x**%i'%(bg,spow))
        self.curve.y[:n] = self.curve.y[:n] - self.fl
        if self.spower != 0:
            self.curve.y[:n] = self.curve.y[:n] * self.curve.x[:n]**self.spower
        pass
    #END loadcurve() SmootherClass
