#!/usr/bin/env python
# coding=utf-8
# Almut Stribeck. almut@stribeck.de
# This is the module file sacupaces.py . It contains objects which
# COMPOSE curves to transform them from one space to another.
# (Composition instead of inheritance)

from __future__ import print_function
import os # To check files
import copy # To generate true copies of variables instead of generating pointers only
import numpy as np
from scipy.special import j0 as BesselJ0 # j0(x) is faster than jn(0,x)
from scipy.special import jn as BesselJn # We need jn(2,x), as well

from sacurve import * # all the curve properties -- remember maxpoints!

#START Generic Classes================================================================
class KernelClass( object ):
    '''The generic Kernel for a Fourier-type transformation. It is called by instances
    of CurveTransformerClass. Subclasses must be defined for each kind of
    transform, and an instance of such a subclass must be supplied to
    CurveTransformer.transform() as the kernel to define the specific
    transform. The predefined methods are the neutral ones.'''

    def __init__(self):
        '''Nothing to do'''
        pass
    #END __init__ KernelClass

    def norm(self, r):
        '''Multiplication factor to be applied at the end of each summation over s.
        This may be a function of r, as in the case of the 2D F-Transform of an
        isotropic function'''
        res = 1.0 # The trivial norm
        return res
    #END norm()

    def kernel(self,arg):
        '''The kernel of the integral transform as a function of arg,
        where arg is, in general, 2*np.pi*r*s'''
        res = np.longfloat(0.0)
        return res # The trivial kernel
    #END kernel()

class CurveTransformerClass( object ):
    '''A class that houses a transcurve where the result of a transformation
    to a different space is created. Its working methods take a ciscurve and
    convert it to transcurve.
    The ciscurve is some kind of an interference function. It is an even
    function. Only its positive branch is stored in it. It vanishes at the
    upper limit. It starts x=0, it is evenly spaced ahd has a number of points
    which is a power of 2.'''

    def __init__(self):
        # Just to show the attribute.
        # In fact, transcurve will be defined by self.transform
        self.transcurve = CurveClass( 128 )
        self.maxpoints = 128
    #END __init__ CurveTransformerClass

    def transform(self, ciscurve, thekernel, m, dr):
        '''The generic transformer of a CurveClass-curve. Typical call:
        err, transcurve = transform( ciscurve, thekernel, arguments )
            ciscurve: The input curve
            mykernel: an instance of KernelClass which supplies the
                      kernel-related information of the integral transform
            m       : Number of points to be generated in transcurve
            dr      : delta-r to use in transcurve
            '''
        err = '' # No error
        if not isinstance(ciscurve, CurveClass):
            err = 'First argument is not an instance of CurveClass'
            return err, transcurve # Transcurve is the default
        if not isinstance( thekernel, KernelClass ):
            err = 'Second argument is not an instance of KernelClass'
            return err, transcurve
        self.maxpoints = len( ciscurve.x )
        # Generate a curve which will contain the result.
        # Be sure to copy.deepcopy() this latent variable in the outer world!
        self.transcurve = CurveClass( self.maxpoints )
        span = ciscurve.n # The number of points in cisspace
        spant = m # The number of points to make in transspace

        # Preset: trivial result curve
        self.transcurve.x[:spant] = dr*np.arange(spant)
        self.transcurve.y[:spant] = np.zeros(spant)
        self.transcurve.n = spant

        # ds is the line element in cisspace. We assume that
        # cisspace.x is linspaced. We do not check, but ease
        # nasty input a little bit by computing the average delta-s:
        ds = ciscurve.x[span-1]/(span-1) # The average spacing
        for ri in range( spant ): # Looping over the result index
            r = self.transcurve.x[ri]   # Pick the r value by its index
            # The first value at s=0 counts only half in the positive branch.
            # Moreover, longfloat helps to minimize rounding errors
            accu = 0.5*ciscurve.y[0]*np.longfloat(thekernel.kernel(0.0))
            pref = np.longfloat(2*np.pi)*r
            for si in range(1,span): # integrate over the following source indices
                arg = pref*ciscurve.x[si] # argument now: 2*Pi*r*s
                accu = accu + ciscurve.y[si]*thekernel.kernel( arg )
            # Postprocessing the integral: multiply by ds and the norm
            accu = accu*ds*thekernel.norm(r)
            # Put down the result value now in low precision
            self.transcurve.y[ri] = np.float( accu )
            # inform user
            if ri%10 == 0: # only every 10 computed points (modulo operator)
                print('Working %i\r'%ri,end='') # Show process on the same line
            pass
        print('%i points generated'%spant) # Show number of generated points and do line feed
        return err, self.transcurve
    #END transform()
#END Generic classes==================================================================

#START Kernels of specific transforms=================================================
class FouIso1KernelClass( KernelClass ):
    '''For 1D transformations of isotropic curves'''

    def kernel(self, arg):
        res = np.cos( arg )
        return res

    def norm(self, r):
        return 2.0

class FouIso2KernelClass( KernelClass ):
    '''For 2D transformations of isotropic curves'''

    def kernel(self, arg):
        res = BesselJ0( arg )
        return res

    def norm(self, r):
        return 2.0*np.pi*r

class FouIso3KernelClass( KernelClass ):
    '''For 3D transformations of isotropic curves'''

    def kernel(self, arg):
        res = np.sinc( arg )
        return res

    def norm(self, r):
        return 2.0

class G11DKernelClass( KernelClass ):
    '''The interference funtion G1 of a one-dimensional intensity measured with
    pointfocus is converted into an IDF by a simple (negative) Fourier cosine
    transform. Because of the SHORT-RANGE CORRELATION among the morphological
    entities, the interference function vanishes at the upper end of the curve
    array and it is allowed to compute the function at any point that is
    deliberately chosen.'''

    def kernel(self, arg):
        res = np.cos( arg )
        return res

    def norm(self, r):
         # Sign: negative Trsf; Factor 2 considers the left branch of the even curve
        return -2.0

class G22DKernelClass( KernelClass ):
    '''The 2D IDF from the projection of the fiber diagram on the equatorial plane
    is computed with a kernel which is the second derivative of the standard
    2D Fourier kernel with respect to r.'''

    def kernel(self, arg):
        res = BesselJ0(arg) - BesselJn(2,arg)
        return res

    def norm(self, r):
        '''Normalized with pi the one-sided integral
            $\int{0,\infty} g_2(x_{12} dx_{12} = A_{p,2}$ becomes
           normalized to the Porod Asymptote in two dimensions.
           Anyway, the Porod law in this case scales with s^{-3}'''
        res = np.pi
        pass