#!/usr/bin/env python
# -*- coding: utf-8 -*-
# coding=utf-8
# sapiawin.py
# Almut Stribeck almut@stribeck.de
from __future__ import print_function
import matplotlib.pyplot as plt
import matplotlib.image as mpimg # For loading of png's
from matplotlib.widgets import Cursor
import matplotlib
from matplotlib.lines import Line2D
from matplotlib.widgets import AxesWidget
import sys
import numpy as np
import savecsel as vs # My vector selectors
import six # for discrimination between PY2 and PY3


class piawins( object ):
    '''Plot interactive windows. These windows are NOT classical collections
    of classical graphs ("figures"), but representations of functions which
    are defined on a 2D raster. They may be (scattering) patterns or photos.
    '''
    def matplotlib_version(self):
        vs = matplotlib.__version__.split('.')
        for i in range(len(vs)):
            vs[i] = int(vs[i])
        return vs
    #END matplotlib_version() of piawins

    def __init__( self, windict=None, winsize=None, verbose=None ):
        '''
        windict : dictionary to manage the interactive windows. If it is
                  provided by the caller, the caller has also control. If
                  not provided, the class INSTANCE MUST stay ALIVE, while
                  interactive windows are on the screen. '''

        if winsize is None:
            self._winsize = 8.0 # "inches" i.e. somewhat smaller than screen
        else:
            self._winsize = winsize
        if verbose is None:
            self._verbose = False
            # Do not report much
        else:
            self._verbose = verbose

        if self._verbose:
            message = 'VERBOSE mode.'
            if windict is None:
                message = message+' Internal windict! Some messages impossible.'
            print(message)

        if windict is None:
            self.wind = {} # GENERATE window dictionary
        else:
            self.wind = windict # POINT at the provided dictionary

        # ** plt.rcParams contains (also) key mappings for the matplotlib
        # ** toolbar, some of which must be disabled during interactive
        # ** work with graphs.
        self.keyfuncd = {} # Dictionary to restore former key functions

        # Preset status variables
        self._keytask = '' # message from keyboard to selector
        self._crosson = True # The crosshair cursor is on
    #END constructor of piawins

    def _disable_kbd_func(self, func, rest='' ):
        ''' ... of the matplotlib window   check:   plt.rcParams
        e.g.:  self._disable_kbd_func('xscale') # logarithmic scaling off
               self._disable_kbd_func('yscale')
               # No keyboard access to toolbar modes:
               self._disable_kbd_func('pan')
               self._disable_kbd_func('zoom')'''
        kfun = 'keymap.%s'%func # Only keyboard entries
        if kfun in self.keyfuncd:
            return # Do not overwrite what has been saved
        value = plt.rcParams[kfun]  # Old value(s)
        self.keyfuncd[kfun] = value # Save for later restore
        plt.rcParams[kfun] = rest   # Disable this function except for rest
    #END _disable_kbd_func() of piawins

    def _recover_kbd_functions( self ):
        saved = self.keyfuncd
        for kfun in saved:
            if saved[kfun] != ['']:
                # Has relevant info, so play back!
                plt.rcParams[kfun] = saved[kfun]
        self.keyfuncd = {}
        # clean dict shows: Free to save, again
    #END _recover_kbd_functions() of piawins

    def gwa( self, winnum ):
        """Get the window's main axes"""
        winstat = self.wind[winnum]
        thefig  = winstat['fg']
        theaxes = winstat['ax']
        if isinstance(theaxes, tuple):
            theaxes = theaxes[0]
        return theaxes
    #END gwa() of piawins

    def gwc( self, winnum, cidkey ):
        """Get the window's connector ID stored under cidkey"""
        winstat = self.wind[winnum]
        thecids = winstat['ci'] # Dictionary of connector ID's
        result = None
        if cidkey in thecids:
            result = thecids[cidkey]
            pass
        return result
    #END gwc() of piawins

    def pwc( self, winnum, cidkey, cid ):
        """Put a cid for the window in the window status' cid dictionary using
        the key cidkey"""
        winstat = self.wind[winnum] # Get window status from manager
        thecids = winstat['ci'] # Get cids from window status
        thecids[cidkey] = cid   # Enter the new cid
        winstat['ci'] = thecids # Put cids in window status
        self.wind[winnum] = winstat # book in windows dictionary
    #END pwc() of piawins

    def _onclose(self, event):
        """Bookkeeping after a window has been closed. After the last window
        has been closed, the interactive plotting is switched off and the
        disabled keyboard functions are recovered."""
        winnum = event.canvas.figure.number # Which window was closed?
        if self._verbose:
            print('Window [%s] closed.'%winnum)
        if winnum in self.wind:
            del self.wind[winnum]
        if len( self.wind ) < 1:
            plt.ioff()
            if self._verbose:
                print('Interactive plotting finished')
            self._recover_kbd_functions()
    #END _onclose() of piawins

    def _onkeyboard(self, event):
        '''Handle key_press_event'''
        sys.stdout.flush()
        if event.key is None:
            #The Shift-Lock key was pressed
            return

        winnum = event.canvas.figure.number # What's my number?

        if self._verbose:
            lowerkey = event.key.lower()
            print('[%s] Keystroke: %s (%s)'%(winnum,event.key,lowerkey))

        # Handle close event
        if event.key == 'ctrl+w':
            if self._verbose:
                print('Window [%s] closed by "ctrl+w"'%winnum)
            return

        # This window's status
        winstat = self.wind[winnum]
        thefig  = winstat['fg']
        theaxes = winstat['ax']
        if isinstance(theaxes, tuple):
            theaxes = theaxes[0]
        thecids = winstat['ci'] # Dictionary of connector ID's

        # Toolbar mode changes by keyboard are forbidden. Almut switched them off
        toolbar = plt.get_current_fig_manager().toolbar
        if toolbar.mode != '':
            print('[%s] KEYBOARD DISABLED while "%s" is active!'%(winnum,toolbar.mode))
            return # Do not react on kbd-keys, if in toolbar mode

        # not needed: allcids['close'] and allcids['keyboard']
        CrossHair = thecids['xhair']    # The crosshair cursor of this window
        VecSelect = thecids['sel_vec']  # The vector selector of this window
        KeyToWin  = thecids['onwkey']   # keystrokes for change of display mode

        segs = event.key.split('+')
        if len( segs ) == 2:
            modifier, key = segs
        else:
            modifier, key = '', event.key

        lowerkey = key.lower()

        # Call the display mode handler if the key does not trigger a
        # vector selector operation (circle, ring, line, bar, box, crosshair on/off)
        if lowerkey not in ['c','g','l','r','x','#']:
            KeyToWin( event, self.wind ) # pass event and window dictionary
            #return # blocks extension without changing the list above

        if lowerkey == '#':
            # Toggle visibility of crosshair cursor
            self._crosson = not self._crosson
            CrossHair.visible = self._crosson
            CrossHair.clear('')
        if lowerkey == 'l': #*** LINE ***
            # Enter a line. May also become the spine of a bar
            CrossHair.clear('')
            CrossHair.visible = not self._crosson
            VecSelect.set_flavor('line')
            self._keytask = event.key # Message to the service handler
            VecSelect.set_active(True) # Input the line
        elif lowerkey == 'x': #*** BOX ***
            # Enter a box (edges parallel to the image edges)
            CrossHair.clear('') # appears to avoid nasty cursor echoes
            CrossHair.visible = not self._crosson
            VecSelect.set_flavor('box')
            self._keytask = event.key # Message to the service handler
            VecSelect.set_active(True)
        elif lowerkey == 'r': #*** BAR ***
            # Enter a bar
            # Check, if a spine is available
            if VecSelect._roiis != 'line':
                print('Use "line" first to enter the spine of the inclined bar!')
                return
            #NOTICE Cf. feed: use the line from the previous call as the backbone
            CrossHair.clear('')
            CrossHair.visible = not self._crosson
            VecSelect.set_flavor('open_bar',feed=VecSelect._roi)
            self._keytask = event.key  # Message to the service handler
            VecSelect.set_active(True)
        elif lowerkey == 'c': #*** CIRCLE by radius ***
            # Enter a circle
            CrossHair.clear('')
            CrossHair.visible = not self._crosson
            VecSelect.set_flavor('circle')
            self._keytask = event.key # Message to the service handler
            VecSelect.set_active(True)
        elif lowerkey == 'g': #*** RING around image center ***
            # Enter a ring
            CrossHair.clear('')
            CrossHair.visible = not self._crosson
            VecSelect.set_flavor('ring')
            # self.ringxc, self.ringyc = VecSelect.set_flavor('ring') #was: TEST
            # print( self.ringxc, self.ringyc )
            self._keytask = event.key # Message to the service handler
            VecSelect.set_active(True) # Input the vector

        pass
    #END _onkeyboard() of piawins

    def _onvecselect(self, linefrom, lineto, roi):
        """A vector input action has been performed in a window. If linefrom
        (a button-press event) is None, the input has been aborted by the
        user (holding alt-key while releasing the button). In any case,
        lineto is a well-defined button-release event."""
        winnum = lineto.canvas.figure.number
        aborted = linefrom is None

        # ** switch from line selector to crosshair cursor
        VecSelect = self.gwc(winnum, 'sel_vec') # Get the window's cid
        VecSelect.set_active(False) # Switch vector selector off
        CrossHair  = self.gwc(winnum, 'xhair')
        CrossHair.visible = self._crosson # Make CrossHair visible
        if aborted or (self._keytask == ''):
            # Do not bother the images action handler
            return

        # Simplify: The vector which has been input
        x1, y1 = linefrom.xdata, linefrom.ydata # in data coordinates
        x2, y2 = lineto.xdata, lineto.ydata

        # serve "verbose" mode
        if self._verbose:
            print('[%s] Vector in flavor "%s"'%(winnum,VecSelect._flavor))
            print('[%s]  from (%5.1f,%5.1f) to (%5.1f,%5.1f). ROI:\n'%(winnum,x1,y1,x2,y2))

        #Extend event by new fields for the handler of window actions
        event = lineto
        event.winnum = event.canvas.figure.number # just for simple access
        event.polyroi = roi              # The polygonal roi
        event.flavor = VecSelect._flavor # line, box, bar, circle
        event.keytask = self._keytask    # What to do with the roi
        event.vector = [ x1, y1, x2, y2 ] # Mind the "classical" arrangement!

        # Grab the pointer to the outer window action handler
        onwindowselection = self.gwc(winnum, 'onwsel')
        # Call it!
        onwindowselection( event )
    #END _onvecselect() of piawins


    def _crosshair( self, ax ):
        '''Crosshair cursor for ax. Visibility can be controlled by
        setting, e.g., cidcur.visible = False && cidcur.clear()'''
        if matplotlib.get_backend() == 'TkAgg':
            cidcur = Cursor(ax, useblit=True, color='green', linewidth=1,
                            vertOn=True, horizOn=True)
        else:
            cidcur = Cursor(ax, color='green', linewidth=1,
                            vertOn=True, horizOn=True)
        return cidcur
    #END _crosshair() of piawins

    def _freshup( self, img, winnum, wintit=None ):
        '''Change content of an open window.
        WARNING If the new img does not have the same size of pixels in both
        directions, it will be squeezed i.e. mapped onto the former data coordinate
        system, and selectors will not work correctly.'''
        if winnum in self.wind:
            winstat = self.wind[winnum]
            pltim  = winstat['im']
            pltim.set_array( img )
            if wintit:
                thefig = winstat['fg']
                thefig.canvas.set_window_title('[%s] %s'%(winnum,wintit))
            plt.show() # Makes new content pop up immediately (no mouse action)
        pass
    #END _freshup() of piawins


    def popup( self, img, winnum=None, wintit=None, clut=None, clutop=None,
              clubott=None, onwsel=None, onwkey=None ):
        '''Show a raster image (e.g. a scattering pattern) in an interactive
        window. img:     the bitmap as a 3D numpy array conforming to matplotlib
                winnum:  the preferred window number of the user
                wintit:  An (additional) title string
                clut:    a color lookup table for "grayscale" data like scattering patterns
                clubott: A number which is mapped on the bottom entry of the clut
                clutop:  A number which is mapped on the top of the clut
                onwsel:  An outer handler which knows what to do with a selection
                onwkey:  an outer handler which can change the view of the image as
                         response to a pressed key'''

        #** Compute the extension of the new piawindow
        imgdims = len(img.shape)
        if imgdims == 2:
            height, width = img.shape
        elif imgdims == 3:
            height, width, channels = img.shape
            if channels != 3:
                # No support of the extra alpha channel!
                raise ValueError('Photo image must have exactly 3 channels (r,g,b)')
        else:
            raise TypeError('img is not two-dimensional')
        ispattern = imgdims == 2

        if ispattern:
            # Mapping of grayscale data (scattering patterns) on a clut
            if clut is None:
                clut = 'CMRmap' # default CLUT
            if clubott is None:
                clubott = np.min(img)
            if clutop is None:
                clutop = np.max(img)

        longest = max( height, width )
        hi = self._winsize*height/longest # window and png height in inches
        wi = self._winsize*width/longest  # window and png width in inches

        switchion = len( self.wind ) == 0 # Remember to set plt.ion()
        # Prepare to register this figure in self.wind
        if winnum is None:
            slot = 1 # The first figure shall be figure 1
            if not switchion:
                # Search for the lowest free slot
                while slot in self.wind:
                    slot += 1
        else:
            if winnum in self.wind:
                #close open window
                if self._verbose:
                    print('Closing old window [%s]'%winnum)
                plt.close( winnum )
            slot = winnum

        thefig = plt.figure(slot,figsize=(wi,hi)) # This also numbers in plt
        # Set the title string of the window
        if wintit is None:
            thefig.canvas.set_window_title('Window [%s]'%slot)
        else:
            thefig.canvas.set_window_title('[%s] %s'%(slot,wintit))

        if switchion:
            if self._verbose:
                print('Switching matplotlib to "interactive", disconnect some keys from toolbar.')
            plt.ion() # Switch interactive plotting on
            self._disable_kbd_func('xscale') # no logarithmic scaling
            self._disable_kbd_func('yscale')
            self._disable_kbd_func('pan')    # Enter pan only by toolbar
            self._disable_kbd_func('zoom')   # Enter zoom only by toolbar
            self._disable_kbd_func('all_axes') # disable: key "a"
            self._disable_kbd_func('back',['left', 'backspace']) # .. key "c"
            self._disable_kbd_func('forward', 'right') # .. key "v"
            self._disable_kbd_func('grid') # .. key "g"
            self._disable_kbd_func('home', 'home')  # .. keys "h", "r"
            # NOTICE save (the plot) interferes with the cursor. It must be
            #        removed from the event chain (and served later: inside
            #        the own keyboard handler) -- but how? NOTICE.
            self._disable_kbd_func('save') # keys "s", "ctrl+s"
            # I cannot remove the icon -- and pressing it causes the mentioned
            # trouble. I tried the following and some more:
            # thefig.canvas.manager.remove_tool('save') #WARNING
            # matplotlib.backend_managers.ToolManager.remove_tool('save')

        # Change the default axes to use the full space (0 .. 1)
        # of the canvas
        lop, hip = 0.0, 1.0
        fitinto = (lop, lop, hip-lop, hip-lop) # start and lengths
        ax = thefig.add_axes( fitinto ) # The frame-filling image
        #** Connect to event handlers
        cidclo = thefig.canvas.mpl_connect('close_event',self._onclose)
        cidkbd = thefig.canvas.mpl_connect('key_press_event',self._onkeyboard)
        cidvec = None # The vector selector
        cidcux = None # The x-hair cursor
        '''onwsel is the cid of the vector selector. It is entered below.
        onwkey handles presentation modes of the image (logscale, bottom up,
        only positive, -- but also: store this pattern! (alt+s)
        '''
        blit = matplotlib.get_backend() == 'TkAgg' #TkAgg knows how to blit
        # Initiate a vector selector in the initial flavor 'line'
        cidvec = vs.VSelector(ax,self._onvecselect, 'line', blit)
        cidvec.set_active(False)
        cidcux = self._crosshair( ax )


        #** Call the draw artist(s)
        if ispattern:
            # Scattering patterns need a CLUT to become colorful
            # The lower end of the colormap (clut) is mapped on
            # vmin, the upper end is mapped on vmax.
            pltim = plt.imshow(img,cmap=clut,vmin=clubott,vmax=clutop,interpolation='none')
        else:
            pltim = plt.imshow(img,interpolation='none')
        # NOTICE How to suppress tick labels
        #plt.xticks([]), plt.yticks([])
        # NOTICE the bounding box of the axes is in screen pixels!
        #print( ax.bbox.get_points() )
        # NOTICE the xlimits and ylimits are in pixels, but the
        #        y-axis is inverted:
        # xmin, xmax = ax.get_xlim()
        # ymin, ymax = ax.get_ylim()
        # print('xlim: (%5.1f,%5.1f), ylim: (%5.1f,%5.1f)'%(xmin,xmax,ymin,ymax))
        # For a 512 x 512 pattern we obtain the output:
        # xlim: ( -0.5,511.5), ylim: (511.5, -0.5)

        # NOTICE matplotlib 1 MUST NOT call plt.show() for the first window.
        # NOTICE matplotlib 2 MUST CALL plt.show() for the first window.
        #        Otherwise: extra empty canvas in mpl1 or in mpl2 first window
        #        will not come up.
        if switchion:
            major, minor, micro = self.matplotlib_version()
            if major > 1:
                plt.show()

        # Display a message in a window overwriting some part of the mapp
        if six.PY2:
            fs = 14 # The font size
        else:
            fs = 12
        hint = '"ctrl+()":inside, "alt+()":out  low=kill, UP=FILL  (l)ine,bo(x),ba(r),(c)ircle,rin(g) [l(o)g,+,-]'
        plthint = plt.figtext(0.01, 0.99, hint, style='normal', color='red', figure=thefig,
                            verticalalignment='top',fontsize=fs, bbox={'facecolor':'white', 'alpha':0.7, 'pad':3})
        plthint.set_visible( False ) # on popup: hide the hint

        # A text box displaying when the function is displayed bottom up
        pltbohi = plt.figtext(0.01, 0.01, "Bottom up", style='normal', color='red', figure=thefig, fontsize=fs, bbox={'facecolor':'white', 'alpha':0.7, 'pad':3})
        pltbohi.set_visible( False ) # is not bottom up - so hide the box

        # collect all cids
        allcids = {'close'    : cidclo,
                   'keyboard' : cidkbd,
                   'xhair'    : cidcux,
                   'sel_vec'  : cidvec,
                   'onwsel'   : onwsel,
                   'onwkey'   : onwkey}

        #** Set up a state dictionary for this window
        # NOTICE content change: pltim.set_array( img )
        state = {'fg' : thefig,   # The figure window itself
                 'ax' : (ax),     # (tuple of) axes, full ax first.
                 'ci' : allcids,  # dictionary of connector id's
                 'im' : pltim,    # for image exchange in existing window
                 'hi' : plthint,  # The helping hint in the window
                 'bo' : pltbohi}  # The bottom-up indicator in the window

        #** register the window in wind
        self.wind[slot] = state

        #** present the window(s)
        # plt.show() #NOTICE Not necessary while interactive
        return slot # The window number which has actually been used
    #END popup() of piawins
