#!/usr/bin/env python
# coding=utf-8
"""
Almut Stribeck. almut@stribeck.de
SASkia: Small-Angle Scattering (k)it for (i)nterpretation and (a)nalysis

As a descendent of TOPAS, there are some placeholders for old TOPAS commands
in the program code, which are not (yet) implemented. 

Developed and tested with
    Python 2.7.6 (default, Oct 26 2016, 20:32:47) [GCC 4.8.4]
        in conjunction with matplotlib 1.3.1 on Linux (Kubuntu),
    Python 3.6.3 (compiled from sources on Linux (CentOS)) [GCC 4.7.7]
        in conjunction with matplotlib 2.1.0
        
    WARNING Did not work with Python 3.4 (pyparse was missing)

    WARNING For MS Windows the Python versions > 3.5.1 are broken!
            No interactive windows are possible, because of a
            fundamental bug introduced into the basic cpython. As
            soon as an INTERACTIVE window is opened, it shows the
            message "(Not Responding)" - and the program
            hangs. Cf.: http://bugs.python.org/issue31546
                PyOS_InputHook is not called when waiting for input()
    Python 3.5.1 (on Windows10 running Anaconda 4.4.0 64 bit)
            conda create -n py351 python=3.5 anaconda
            ... went broke after several tries to find a working Python 3
        Every time before starting python on Windows one must call
            activate py351

    To check the Python version:
        import sys
        sys.version
    To check the matplotlib version:
        import matplotlib
        matplotlib.__version__

    After installation of python open a console (Anaconda Prompt under
    Windows, text console under Linux) and install extra modules:
        pip install cmd2
        pip install pygame

    If Python must be installed from the sources, my hints in
    saskia_READMEjjmm.txt may be helpful.

General:
    This is a cmd2-based program. Such programs require Python 2.7
    or Python 3.5 and higher
Work on the command language. So far:
(1) As additional argument separators commata are allowed. Only this
    facilitates skipping arguments.
(2) The user can call linear scripts (no branching, but formal arguments
    in scripts are replaced by actual ones). Scripts are text files with
    the default extension .lrg (for my script-loader command "largo".
    largo means: "Load script with ARguments and GO"). The shortcut command
    character "@" for script loading here points at largo.
(3) Errors in scripts which are called by largo do only flush the
    command queue and return to the user prompt. They do not stop the
    program.
(4) cmd2's command-line commands which contain scripts with errors will only
    be offered to the interpreter once. Thereafter postloop switches off
    allow_cli_args
(5) Script nesting is possible
(6) Multi-commands per line supported on the user's direct input line.
    The separator character is '&'
Not yet available:
    lists of file names as arguments of commands or scripts

Work on the graphics:
    The interactive graphics is tested for the Tk-backend ("TkAgg").
    Because graphical tools vanish under Python 2.7 when the curves
    are presented in a differently transformed view, critical state
    changes are realized by completely closing the plot and opening
    a new one. This is a bit nasty.
    SASkia's interactive graphics assumes that the standard keypress
    events of matplotlib have not been reassigned by the user, because
    it has to take care for special handling of those keypress events.

    In order to achieve compatibility across the matplotlib versions,
    an individual LineSelector is in curveselector.py. This selector
    will move the complete latent line, while the Ctrl-key is pressed.
"""
# The following line makes "print" from Python2 work as in Python3.
from __future__ import print_function

from savcmd2 import Cmd, set_use_arg_list
import six  # To handle differences between Python2.7 and Python3.6
import copy # To generate true copies of variables instead of pointers
import numpy as np # For arrays etc.
# for nonlinear regression in do_extexp
import warnings
from scipy.optimize import curve_fit, OptimizeWarning


# DO NOT IMPORT ANY matplotlib RELATED STUFF BEFORE THIS LINE!
import matplotlib # For the graphics
import matplotlib.pyplot as plt # for 2D-Plots
plt.switch_backend('TkAgg') # Relaxed way to enforce using a backend
""" WARNING Explicitly select a screen-qualified backend. This is necessary
    in my Python3 environment which uses the non-screen "Agg" backend, if
    not specified explicitly.

    To compile backends for matplotlib is a poorly described task. I needed
    it for the environment on the CentOS computer. If Python 3 is compiled
    from the source tar ball, the related backend-related Linux system
    libraries are required to be already installed before compilation.
    BOTH the BINARIES AND the DEVELOPMENT HEADERS must have been installed.
"""

# import sys # For handling of keypress events in the interactive graphics
import shlex # For my own version of pyscript

from sacurve import * # Curves: Definitions, I/O and basic manipulations
from sacutray import *  # Curves display (and interactive manipulation)
from sacuspaces import * # Transformations between reciprocal and physical space
from datetime import datetime, timedelta # Arithmetics with time stamps

class SasCmds(Cmd):
    """ A collection of commands intended to evaluate small-angle X-ray
    scattering curves together with their related environmental data. """

    # Setting default_to_shell to "True" makes that unknown commands are 
    # transferred to the operating system
    #default_to_shell = True

    # In the example application constants are defined here. There they are
    # typed in capital letters (e.g. MUMBLES)


    def __init__(self):
        # WARNING The feature of cmd2 to use abbreviated command names has
        #         been withdrawn by the developers. It will only work, if my
        # frozen-in version savcmd2.py of cmd2 is used.
        # Commands may be abbreviated
        self.abbrev = True
        # Conmmands entered in the following list extend up to the next
        # End-of-Command token (";" or an empty line). The interpreter stops
        # interpreting, after such a token. Several commands can be put on
        # the same line of DIRECT user input by separating them by "&". This
        # is an extra feature of SASkia.
        #self.multilineCommands = ['orate']

        # Adjust for change concerning the default values of matplotlib
        # cf.: https://matplotlib.org/users/dflt_style_changes.html
        mv = matplotlib.__version__
        mvs = mv.split('.')
        if int(mvs[0]) < 2: # Assuming that there are digits before the dot
            self.fontfactor = 1.0 # No extra fontsize factor
        else:
            # Font sizes are optimized for the scheme of matplotlib 1.3.1,
            # so make them smaller for the newer matplotlib
            self.fontfactor = 0.7

        # For a variable which shall be settable we must add it to the OrderedDict settable
        self.settable['fontfactor'] = 'Adjust the size of label fonts used in the plot'
        # Now the user can call "set fontfactor 0.7" do decrease the font sizes in the plot

        # Possibilities e.g.: 001 = 1 --> display only OP0 (if it contains data)
        #                     011 = 3 --> display OP1 and OP0
        # 110 = 6 --> display OP2 and OP1, but not OP0 (even if it contains a curve)
        self.visicurves = 7 # Binary: 111 -- Display all curves
        # The graphics command shows only OP1 and OP2
        self.settable['visicurves'] = 'Bit pattern controlling the display of curves in cmd "graphic"'

        '''The path strings tell the program, where and in which order to
        search for files. (UNIX: echo $PATH$, MS Windows: echo %PATH%). in the
        definition used here different locations ar separated by ";", as is
        usual in MS Windows. Then the UNIX separator is free to be used under
        MS Windows in drive-specific paths. Also the directory separator
        differs among the operating systems. If the user wants to keep the
        path independent from the operating system, she may use "." as the
        directory separator, but must avoid to use the operating system
        specific separators "/" or "\" at all. Only then dots will be
        translated "on the fly" into the correct directory separator. The
        path is processed by the method _pathlist()'''
        self.datapath = '' # related to data files TODO Not yet implemented!
        self.settable['datapath'] = 'Where to search for (other) data files, if implemented'
        self.scriptpath = '' # related to script files
        self.settable['scriptpath'] = 'Where to search for (own) scripts'

        # Command shortcuts are defined here
        self.shortcuts.update({'@': 'largo'})
        self.shortcuts.update({'@@': 'pyscript'})

        # Now initialization of the ancestor
        # use_ipython permits to open the ipython console by the command "ipy" 
        Cmd.__init__(self, use_ipython=True)

        # above the prompt: 79 dashes plus a line feed as separator
        _delim=79*'-'+'\n'
        self.prompt = _delim+'SASkia> '

        # Presented to the commands is a string containing the specifications
        # for the arguments. This means: no list of specifications.
        set_use_arg_list(False)

        # =================The data block======================================
        # Initialize the operand stack with 3 empty curves
        self._MAXP = 4096
        self.cmdres = None # Commands may store a result for further processing
        self.curvea = CurveClass( self._MAXP ) # a curve with environmental dara
        # The "operand stack" self.ops must be made from really different
        # curves, not only by three pointers to the same curve self.curvea.
        self.ops = [ self.curvea, copy.deepcopy(self.curvea),
                    copy.deepcopy(self.curvea) ]
        # WARNING My curve bookkeeper will mixup or even break the pointer
        # identity of self.curvea. Therefore take care to only address the
        # operands via self.ops[]
        
        # A list of actual specifications for replacement of formal parameters
        # in loaded scripts
        self.largoargs=[]
        # An instance of the standard graphics from sacutray.py
        self.ThePlot = SasPlotClass()
        # Instance of the plot used for the inspection of raw and smoothed data
        # (used to interactively find a good smoothing parameter)
        self.SmoPlot = IaSmoothPlot()
        # Working environment of a curve smoother which does set-by-set
        # smoothing. Has been removed, because the internal smoother of
        # python works satisfactory for complete curves.
        # (Old source: 00smoothcmd.sav.py)
        # self.Smoother = SmootherClass()
        self.piawinds = {} # Dictionary of plot-interactive windows
        self.patt = None # No last 2D-pattern attached
    #END constructor

    def preparse(self, line):
        '''Extension of the command syntax by permitting commata as argument
        separators. Replaces empty arguments by a "=" (indicating a request
        to use a specification default). Replaces commata in the argument list
        by blanks.

        Extra functionality:
          - If the program is interpreting a script which is called by the largo
            command, preparse tries to replace formal arguments "$1" ... "$9"
            by the actual specs which have been following the script name in the
            largo command call (Remember: "@" is shortcut for "largo,").
          - If the program is in direct user line input: clear the list
            self.largoargs, which may still exist after having returned
            from largo.
          - Handle multi-command input on a single direct user input line by
            defining the ampersand "&" as the command separator. 
        '''

        if self.cmdqueue == []: # Is the program processing a direct line input?
            # Clear the list of largoargs for safety
            self.largoargs = []
            # Handle multi-command line input (separating char: "&")
            rawqueue = line.split('&')
            queue = []
            for i in range(len(rawqueue)):
                if rawqueue[i] != '':
                    queue.append(rawqueue[i])
                    pass
                pass
            # A nasty user may have entered only escape characters
            if queue == []:
                line = ''
            else:
                line = queue.pop(0) ; # No problem to overwrite outer line
                # Push rest of the queue to cmdqueue
                self.cmdqueue = copy.copy(queue) # no deepcopy required

        # Can formal arguments be replaced by specifications of a caller???
        if len(self.largoargs) > 0:
            lilic = line.split('$') #Are there escape characters?
            for i in range(1,len(lilic)):
                dummy = lilic[i]
                if len(dummy) > 0:
                    if dummy[0] in ('1','2','3','4','5','6','7','8','9'):
                        j = int(dummy[0])-1 #Index of actual specification
                        k = len(self.largoargs) - 1 # .. at the end of the list
                        specs = self.largoargs[k] # .. are the actual specs
                        if (len(specs) > j): #Replacement is possible
                            if specs[j] == '=': # emptiness indicator
                                specs[j] = '' # restore empty actual spec
                            lilic[i] = specs[j]+dummy[1:] # replace by actual spec
                        else:
                            lilic[i] = '$'+lilic[i] # restore formal arg
                            pass
                        pass
                    else:
                        lilic[i] = '$'+lilic[i] # restore the dollar-clause
                    pass
                else:
                    lilic[i] = '$'+lilic[i] # restore the dollar-clause
                pass
            line = u''.join(lilic) # The line after replacement
        # Old start of preparse
        lilic = line.split(',')
        # replace empty arguments by default indicator ('=')
        for i in range( len(lilic) ):
            if lilic[i] == '':
                lilic[i] = '='
            # If one would allow multi-word arguments, one could
            # encapsulate each argument in quotation marks here
            #lilic[i] = '"'+lilic[i]+'"'
        result = ' '.join(lilic) # To cmd2-syntax: separate arguments by blanks
        # Replace multiple spaces by single ones (Commata adjacent 
        # to spaces may have caused such multiple spaces)
        result = ' '.join(result.split())
        return result

    def postloop(self):
        '''Guarantee that the program will only stop if the quit command is
           explicitly called'''
        last_cmd = self.history.pop()
        if last_cmd not in ('quit','qui'):
            self.largoargs = [] # erase script arguments
            self.cmdqueue = [] # erase command queue
            # Prevent cmdloop() from re-processing the command line arguments
            # before restarting the interpreter loop
            self.allow_cli_args = False
            self.cmdloop() # Catch the error by entering a new cmdloop
        else:
            # SASkia stops here
            self.poutput('Good Bye')
    #END postloop

    def do_eos(self, _):
        """Handles cleanup when a script has finished executing.
        This extended version, additionally, does bookkeeping
        of the do_largo argument stack"""
        # Is there actually a file directory related to a script?
        if self._script_dir:
            self._script_dir.pop()
            if self.largoargs:
                self.largoargs.pop() # remove innermost arguments
    #END do_eos (end of script)


    #------------------- primitive commands for testing --------------------
    # My first primitive command (only for debugging)
    def do_argprint(self, arg, opts=None):
        '''Prints the specifications which have been passed to it --
        after the preparsing (preparse) has finished'''
        # join the list elements using the separator ''
        # This overwrites the variable arg!!!
        arg = u''.join(arg)
        self.poutput(arg)
    #END do_argprint

    def do_comment(self, arg, opts=None):
        '''Takes the arguments simply as a comment and prints it. It makes no
    sense to define the shortcut "#" for this command because the hashtag is
    treated as the escape character for a comment which is suppressed anyway'''
        result = '# '+ arg
        self.poutput(result)
    #END do_comment
    #--------------End of Primitive commands for testing -------------------

    #--------------Helper methods----------------------- -------------------
    def _longlist(self, arg, specs):
        '''splits the argument line into an argument list and fills this
        list with default strings until specs arguments are in the list.'''
        result = arg.split(' ')
        if result[0] == '':
            result[0] = '='
        while len( result ) < specs:
            result.append( '=' )
        return result

    def _setdefault(self, arg, fallback):
        '''
        Handles the default for a specification, i.e. it replaces a '='
        by fallback
        '''
        result = arg
        if (arg == '=') or (arg == ''):
            result = fallback
        return result

    def _tonumber(self, text ):
        '''Converts a string to a number and handles exceptions
        '''
        ErrDescr = '' # no error
        res = 0 # Default result on error
        try:
            res = float(text)
        except ValueError as verr:
            ErrDescr = '"%s" is not a number'%text
        except Exception as ex:
            ErrDescr = 'Exception occured while converting "%s" to float'%text
        return res, ErrDescr
    #END _tonumber

    def _toint(self, text):
        '''Converts a string to an integer and handles exceptions
        '''
        ErrDescr = '' # no error
        res = 0 # Default result on error
        try:
            res = int(text)
        except ValueError as verr:
            ErrDescr = '"%s" is no integer'%text
        except Exception as ex:
            ErrDescr = 'Exception occured while converting "%s" to integer'%text
        return res, ErrDescr
    #END _toint()

    def _toopi(self, text):
        '''Converts a string to an operand index which specifies a curve'''
        res = -1
        ErrDescr = ''
        if text in ('0','1','2'):
            res, ErrDescr = self._toint(text)
        else:
            ErrDescr = 'Curve index "%s" is not allowed.'%text
        return res, ErrDescr
    #END _toopi()

    def _todatetime(self, text):
        '''Converts the standard international date-time stamp to a datetime
        object. Then python can do date arithmetics. No error checking.
            input, e.g.: "2017-11-27 11:27:19.829261"
        '''
        res = datetime.strptime(text,'%Y-%m-%d %H:%M:%S.%f')
        return res

    def _normaltimestr(self, albatime):
        '''Convert the alba time string to international format which works
        on all computers irrespective of "locale" setting. The Alba time stamp
        looks:  "Nov 21 18:09:19 2017\n"
        international pythonic: "2017-11-21 18:09.19.000000"
        '''
        # modi translates month names in the locale en_US to numbers
        # irrespective of the language setting of the computer on which this
        # program is running.
        modi={'Jan': '01', 'Feb': '02', 'Mar': '03', 'Apr': '04', 'May': '05', 'Jun': '06',
              'Jul': '07', 'Aug': '08', 'Sep': '09', 'Oct': '10', 'Nov': '11', 'Dec': '12'}
        elem = albatime.split(' ')
        month = modi[elem[0]] # numeric month
        year  = elem[3][:4]   # remove trailing \n character (if still present)
        result = year+'-'+month+'-'+elem[1]+' '+elem[2]+'.000000' #
        return result
    #END _normaltimestr

    def _pathlist( self, pathstr ):
        # Convert a path string into a list of paths, in which all elements
        # conform to the user's wish to be os-dependent or independent, resp.
        if pathstr == '':
            return [pathstr]
        ipath = pathstr
        if ipath.find(os.sep) == -1:
            # No os-dependent directory separator is in path.
            # Replace dots by directory separators!
            pathlist = ipath.split('.')
            ipath = os.sep.join(pathlist)
        pathlist = ipath.split(';')
        for i, p in enumerate(pathlist):
            if p != '': # if p is not the present directory...
                if p[-1] != os.sep:
                    # os.sep is missing at the end. Append it!
                    pathlist[i] = p + os.sep
            # Expand "~"
            pathlist[i] = os.path.expanduser( pathlist[i])
        return pathlist
    #END _pathlist

    def _findscript( self, fn, ext ):
        '''Search for a script file according to the following search rules:
        (1) iterate the self.scriptpath. Remember: ";" is a SEPARATOR so if
            the current directory shall be searched first, self.scriptpath
            must start with a ";" - a";;" inserts the current directory
            at a later position in the search queue.
            (a) try the file name fn as it has been given.
            (b) try fn+ext
        (2) check the existence of the script in a distribution related
            script directory which is the subdirectory "scripts" of the
            directory in which the program itself resides.
            (a) try the file name fn as it has been given.
            (b) try fn+ext
            '''

        # Iterate the path list, if given
        userdirs = self._pathlist(self.scriptpath)
        for d in userdirs:
            found = d + fn
            if os.path.isfile(found):
                return found
            found = found + ext
            if os.path.isfile(found):
                return found

        # The distribution directory for scripts
        myself = sys.argv[0]   # The running script
        num = myself.rfind(os.sep) + 1
        myhome = myself[0:num] # its home directory
        myscripts = myhome + 'scripts' + os.sep # subdir scripts
        found = myscripts+fn
        if os.path.isfile(found):
            return found
        found = found+ext
        if not os.path.isfile(found):
            found = '' # not found
        return found
    #END _findscript

    def do_pyscript(self, arg, opts=None):
        """Runs a python script file from the console.
    SYNTAX: pyscript,<python-script>,<arguments>
    Example: pyscript,arg_printer,hello, here, I am

    Pyscripts cannot be called from within a pyscript (as can the largo
    scripts), but pyscripts can use all the features of python programs.

Inside a script the native console commands can be executed by calling
    cmd("<thecommand>"), e.g.  cmd("read,data.dat")
From inside the pyscript all internal variables can be addressed, e.g.
    self.ops[1].n = 0    erases the curve in operand 1.
Paths or arguments that contain spaces must be enclosed in quotes.

    Here an enhanced, local version of the "pyscript" command is running.
    It features a user-friendly script-file search. First it tries to find
    the file as specified by the user. Then it tries to append the default
    extension ".py". If this does not work, it searches the subdirectory
    "scripts" of the directory in which this program resides. This provides
    means to have important, general python scripts (as well as largo
    scripts) at hand from any data directory."""
        if not arg:
            self.perror("pyscript requires at least 1 argument ...", traceback_war=False)
            self.do_help('pyscript')
            return
        arglist = self._longlist(arg,1) # Command has at least 1 argument
        # Apply searchrule to the script name
        scriptloc = self._findscript( arglist[0], '.py' )
        if scriptloc == '':
            self.perror('Python script %s not found'%arglist[0], traceback_war=False)
            return True # Exit from cmd queue
        # Enter expanded script name
        self.poutput('Start python script:  %s'%scriptloc)
        arglist[0] = scriptloc

        # TODO find a way to circumvent the problem of blanks in file paths
        if scriptloc.rfind(' ') != -1:
            print('File path contains blank spaces! Cross fingers!!!')
            # repr formatting introduces operating-system related escape
            # sequences (needed for Windows)
            arglist[0] = "{!r}".format(script_path)
            set_use_arg_list( True )
            # Call my parent with the arguments as a list
            Cmd.do_pyscript( self, arglist )
            set_use_arg_list( False )
            return

        # re-join the argument string
        argstring = u' '.join( arglist )
        # Call my parent
        Cmd.do_pyscript( self, argstring )
    #END do_pyscript

# ------------------Script nesting and loading of scripts with arguments
    def lonest(self, file_path):
        '''LOad NESTed (scripts). This is a modified do_load. It is
        called by do_largo. In contrast to do_largo, here the commands
        from the inner script are PREpended to the cmdqueue. Thus the
        commands from the inner script are consumed one by one, and
        afterwards the command from the outer script is continued.'''

        expanded_path = os.path.abspath(os.path.expanduser(file_path.strip()))

        # Make sure expanded_path points to a file
        if not os.path.isfile(expanded_path):
            self.perror('{} does not exist or is not a file'.format(expanded_path), traceback_war=False)
            return

        # Make sure the file is not empty
        if os.path.getsize(expanded_path) == 0:
            self.perror('{} is empty'.format(expanded_path), traceback_war=False)
            return

        # Make sure the file is ASCII or UTF-8 encoded text
        if not self.is_text_file(expanded_path):
            self.perror('{} is not an ASCII or UTF-8 encoded text file'.format(expanded_path), traceback_war=False)
            return

        try:
            # Specify file encoding in Python 3, but Python 2 
            # doesn't allow that argument to open()
            if six.PY3:
                # Add all commands in the script to the command queue
                with open(expanded_path, encoding='utf-8') as target:
                    inner = target.read().splitlines() # The innermost commands
                    inner.append('eos') # schedule clean up by do_eos
                    inner.extend(self.cmdqueue) # append the rest of the command queue
                    self.cmdqueue = inner # ... and make it the new cmdqueue
            else:
                # Add all commands in the script to the command queue
                with open(expanded_path) as target:
                    inner = target.read().splitlines() # The innermost commands
                    inner.append('eos') # schedule cleanup by do_eos
                    inner.extend(self.cmdqueue) # append the rest of the command queue
                    self.cmdqueue = inner # ... and make it the new cmdqueue

        except IOError as e: # e is the error string
            self.perror('Problem accessing script from {}:\n{}'.format(expanded_path, e))
            return

        self._script_dir.append(os.path.dirname(expanded_path))


    def do_largo(self, arg):
        '''Load a script file and supply ARGuments to gO
    SYNTAX: largo,<scriptfile>,<arguments>
        <scriptfile> : ASCII file containing a list of commands.
             Inside the largo scripts specifications may be formal parameters
             "$1" .. "$9", which are replaced by the actual arguments of the
             script call itself.
             DEFAULT file extension: ".lrg"
        largo scripts are linear, but they can be nested "deliberately", and
        replacement of formal arguments is handled. If the script is not found
        in the current path, it is searched in the scriptpath (cf. saskia.rc)
        using the default file extension ".lrg".

self.largoargs is a list of lists. If self.largoargs[0] is empty, this means:
no argument replacement is possible. The set of arguments which is related to
the active script is in self.largoargs[len(self.largoargs)-1]. self.largoargs
is prolonged by this command and shortened by do_eos (end of script).'''
        argu = self.preparse( arg ) # make blank separated argument list
        arglist = self._longlist( argu, 1 )
        scriptin = self._setdefault(arglist[0],'') # No default
        scriptfile = self._findscript(scriptin,'.lrg')
        # Error if scriptfile is not found
        if not scriptfile:
            self.perror('lrg script %s not found'%scriptin, traceback_war=False)
            return False

        # strip name of largo script and keep the arguments
        arglist = arglist[1:]
        self.largoargs.append(arglist) # Provide script nesting
        # self.largoargs = [ arglist ] # No nesting in this case
        argstring = u' '.join( arglist )
        # self.poutput('Largo arguments: %s'%argstring)
        # WARNING  super(SasTools1D, self).do_load(scriptfile) does not work
        # lonest is my (modified) copy of do_load
        self.lonest( scriptfile )
        return False # Do not quit the cmdloop
    #END do_largo

    #------------- End of "Loading Scripts with arguments and nesting-------------

    # ------------ I/O related commands ------------------------------------------

    def do_readcurve(self, arg):
        '''Reads a curve from an ASCII file
    SYNTAX: readcurve, <filename>[.dat], <op>, [<header>,<conden>]
      <filename> is a string. If the string does not contain a ".",
                 then the default extension ".dat" is appended.
      <op> is the index of the curve register which may be 0, 1 or 2.
           DEFAULT: If <op> is missing: <op>=1
                    If <op> is not 0, 1 or 2: <op>=0
      <header> = "i": ignore header lines (except for environment
                      attribute entries)
      <conden> = something: The curve is condensed to half its points
                 by averaging pairs of points.
    Curves may not contain more than 4096 data points (self._MAXP).
    The format of a curve: Comment lines starting with # are allowed. If the
    second character of a comment line is a "{", The line contains a key-value
    pair according to the syntax of a python dictionary entry. The key is
    always a string. The value may be of any type. Examples:
        #{"I0": 163.918855591}
        #{"axis1": "s"}
    A line which does not start with "#" must hold at least two numbers
    separated by spaces. The first two numbers on the line are interpreted as
    the x- and the y-value, respectively.'''

        arglist = self._longlist(arg, 4) # be sure to have a list of 3 specs
        fn = self._setdefault(arglist[0], '')
        # Handle second argument (opi)
        opi = self._setdefault(arglist[1], '1')
        if opi not in ('0','1','2'):
            self.perror('opi=%s not allowed'%opi, traceback_war=False)
            return True
        opi = int( opi ) # Index in list ops[]
        stripheaders = self._setdefault(arglist[2],'no')
        stripheaders = stripheaders[0].lower() == 'i'
        condense = self._setdefault(arglist[3],'')
        condense = condense != ''
        # Call the read method of the curve operand
        _err = self.ops[opi].readascii( fn, stripheaders, condense )
        if _err != '':
            self.perror(_err, traceback_war=False)
            # With return True we ask to flush the command queue ( self.cmdqueue )
            return True
        else:
            self.poutput('Curve %s read into OP%1i with %4i points' %(fn,opi,self.ops[opi].n))
            if self.ops[opi].n >= len(self.ops[opi].x):
                self.poutput('*** Too many points. Curve truncated')
    #END do_readcurve()

    def do_writecurve(self, arg):
        '''Command to save a curve in an ASCII text file.
        SYNTAX: writecurve, <filename>[], <opi>[1], <force>[]
          <filename> is a string. If the string does not contain a ".",
                     then the default extension ".dat" is appended.
                     If no filename is specified, the filename is used
                     under which the curve is known to SASkia
                     ( self.ops[opi].lastfile ).
                     If the file already exists it is only overwritten,
                     if <force>=f
          <opi>      The index of the operand which contains the curve.
                     It may be 0, 1 or 2. DEFAULT: <opi>=1
          <force>    = force : overwrite an existing file with the same
                     name.'''

        arglist = self._longlist(arg, 3) # be sure to have 3 specifications

        ## (2) Operand to write
        opis = self._setdefault(arglist[1], '1')
        opi, err = self._toopi(opis)
        if err != '':
            self.perror( err, traceback_war=False )
            return True # Flush cmd queue and return to prompt
        curve = self.ops[opi]
        if curve.isempty():
            self.perror('Curve array OP%i is empty'%opi, traceback_war=False)
            return True # Flush cmd queue and return to prompt

        ## (1) Filename to write to. Default is "lastfile"
        fn = self._setdefault(arglist[0], curve.lastfile)
        ## (3): Force to overwrite an existing file?
        force = arglist[2] != ''
        if force:
            force = arglist[2][0] == 'f'
        _err = self.ops[opi].writeascii(fn, force)
        if _err != '':
            self.perror(_err, traceback_war=False)
            return True

        self.poutput('Curve from OP%1i with %4i points written to %s' %(opi,self.ops[opi].n,self.ops[opi].lastfile))
        if len(self.ops[opi].headers) > 0:
            self.pfeedback('%s'%self.ops[opi].headers[0])
    #END do_writecurve()

    def do_attribute( self, arg ):
        '''Get or set an attribute of a curve. Attributes are stored in the dictionary
    self.ops[opi].env -- The &write command also saves the dictionary in the
    ASCII data file. The &read command reads the dictionary back from the
    data file.
    SYNTAX: attribute,<opi>[1],<key>[],<value>[]
        <opi>   Index of the operand to work on.
        <key>   The key to the attribute. If <key> is missing, the complete
                attribute dictionary is dumped to the screen
        <value> If no value is provided, the command returns the value of the
                corresponding key also in the variable "self.cmdres". This is
                good for collecting attributes of data series in new curves.
                If <value> is the word "delete", the respective entry is
                deleted from the dictionary.
                If a value is provided, the corresponding entry is generated
                or overwritten. Values which can be converted into integers
                will be stored as integer. Values which can be converted into
                real numbers will be converted into real numbers. All other
                kind of values are stored as strings.'''
        arglist = self._longlist(arg, 3) # be sure to have a list of 3 specs
        opis = self._setdefault(arglist[0], '1')
        opi, err = self._toopi(opis)
        if err != '':
            self.perror( err, traceback_war=False )
            return True # Flush cmd queue and return to prompt
        if self.ops[opi].isempty():
            self.perror('Curve array OP%i is empty'%opi, traceback_war=False)
            return True # Flush cmd queue and return to prompt
        keys = self._setdefault(arglist[1], '')
        if keys == '':
            # Dump the whole attribute dictionary to the screen
            self.poutput('OP%i: Attributes and their values:'%opi)
            attr = self.ops[opi].env
            for key in attr:
                self.poutput('%s  =  %s'%(key,attr[key]))
            return
        valstr = self._setdefault(arglist[2], '')
        if valstr == '':
            res = self.ops[opi].attribute(keys) # Call the attribute method of the curve
            self.poutput("self.ops[%i].env['%s']: %s %s"%(opi,keys,res,type(res)))
            self.cmdres = copy.deepcopy(res)
        else:
            self.ops[opi].attribute(keys,valstr)
    #END do_attribute

    def do_headers( self, arg ):
        '''Display all headers of the curve or append an additional header line.
Some characters of an additional header line are swallowed by the command
interpreter (see below). If this must be circumvented, &ipy or a python script
must call  "self.ops[opi].headers.append(newhead)" directly.
    SYNTAX: headers,<opi>[1],<newheader>['']
        <opi>       Index of the operand to work on.
        <newheader> A new header to append. All follow-up arguments are
                    concatenated to form the new header. WARNING: Several
                    characters are stripped from the header without warning:
                    ",", ";" ... -- moreover, the leading "#" must not be
                    input by the user, because it introduces a comment even
                    to the command interpreter.
                    If <newheader> is missing, then all headers are shown.'''
        arglist = self._longlist(arg, 2) # be sure to have a list of 2 specs
        opis = self._setdefault(arglist[0], '1')
        opi, err = self._toopi(opis)
        if err != '':
            self.perror( err, traceback_war=False )
            return True # Flush cmd queue and return to prompt
        if self.ops[opi].isempty():
            self.perror('Curve array OP%i is empty'%opi, traceback_war=False)
            return True # Flush cmd queue and return to prompt
        newhead = self._setdefault(arglist[1], '')
        if newhead == '':
            # show all headers on the screen
            theheaders = self.ops[opi].headers
            self.poutput('OP%i headers:'%opi)
            if len(theheaders) > 0: # Only if there are headers
                for i, head in enumerate(theheaders):
                    self.poutput( head )
            return

        # Here the "arg" is a string in which the first character
        # is either "=" (for default) or "0", "1", "2".
        # WARNING The following code only works, if the highest
        #         operand index is a single character
        newhead = '#'+arg[1:] # Necessary to mark the header
        self.ops[opi].headers.append(newhead)
    #END do_headers

    #-------------- END I/O-related commands-----------------------------
    #-------------- Info commands----------------------------------------
    def do_status(self, arg):
        '''Shows the contents of the operand arrays.
    SYNTAX: status,<verbose>
        <verbose> : If this argument is not empty, then even all the headers
                    are shown.
    Hint: Access one of the operand curves after call of &ipy command via
          self.ops[opi] with opi from (0,1,2)
        '''
        arglist = self._longlist(arg,1)
        verbose = arglist[0]
        for opi in range(3):
            self.poutput('OP%i '%opi,'') # print string without line feed
            lines = self.ops[opi].status(verbose)
            for i in range(len(lines)):
                self.poutput(lines[i])
    #END do_status()

    def do_dumpop(self, arg):
        '''Show the curve data (i.e x-y-pairs) of an operand
        SYNTAX: dumpeop,<opi>
                <opi>: Operand index ( '0', '1,', '2')
                       DEFAULT: opi=1
    Hint: Access one of the operand curves after call of &ipy command via
          self.ops[opi] with opi from (0,1,2)
        '''
        arglist = self._longlist(arg,1)
        opi = self._setdefault(arglist[0], '1')
        if opi not in ('0','1','2'):
            self.perror('Operand index must be 0, 1 or 2', traceback_war=False)
            # This command never fails, so we do not need to erase the cmdqueue
            return
        opi = int(opi)
        self.poutput('OP%i '%opi,'') # print string without line feed
        line = self.ops[opi].status('') # short status
        self.poutput( line[0] ) # Now with linefeed
        points = self.ops[opi].n
        if points == 0:
            return
        ppl = 3 # pairs per line
        pitems = 20*ppl # pause after each 20 lines
        for i in range(self.ops[opi].n):
            modul = i % ppl
            if modul == 0:
                self.poutput('%4i| '%i,'') # Index only for first pair in each output line
            self.poutput( '%10.3e,%10.3e| '%(self.ops[opi].x[i],self.ops[opi].y[i]),'')
            if modul == (ppl-1):
                self.poutput('\n') # a Line feed, please
            #if ((i+1) % pitems) == 0: # To introduce a scroll pause
            #    self.poutput('pause')
        if (self.ops[opi].n % ppl) != 0:
            self.poutput('\n') # a last line feed was required
    #END do_dumpop()

    def do_opsjuggle(self, arg):
        '''Juggling with curves in the operands OP0, OP1, OP1
        SYNTAX: opsjuggle,<token>
         <token>  = '0', '1', '2' delete curve in operand OP<token>
                  = 'r' copy right, i.e. OP1 --> OP2
                  = 'l' copy left,  i.e. OP1 <-- OP2
                  = '.' exchange OP0 <--> OP1
                  = 'x' exchange OP1 <--> OP2
                  = '^' ring change: OP0 -> OP1, OP1 -> OP2, OP2 -> OP0
        access curves in operands in iPython console by
                  self.ops[opi] with opi from [0,1,2]
        E.g.:
        SASkia> ipy
        In [1]: thecurve = self.ops[1]
        In [2]: thecurve.isempty()
        Out[1]: False
        '''
        arglist = self._longlist(arg,1) # Only one argument in unicode
        inst = arglist[0]
        if (inst == '') or (len(inst) > 1):
            self.perror('Instruction must be a single character, not "%s"'%inst, traceback_war=False)
            return True # Flush cmdqueue
        if inst in ('0','1','2'):
            # erase curve in operand
            opi = int(inst)
            self.ops[opi].clean()
            self.poutput('OP%i has been erased'%opi)
        elif inst.lower() == 'r':
            # fill 2 from 1
            self.ops[2] = copy.deepcopy(self.ops[1])
            self.poutput('Copied: OP1 --> OP2')
        elif inst.lower() == 'l':
            # fill 1 from 2
            self.ops[1] = copy.deepcopy(self.ops[2])
            self.poutput('Copied: OP1 <-- OP2')
        elif inst.lower() == 'x':
            # exchange OP1 and OP2
            dummy = copy.deepcopy(self.ops[2])
            self.ops[2] = copy.deepcopy(self.ops[1])
            self.ops[1] = copy.deepcopy(dummy)
            self.poutput('Exchanged: OP1 <--> OP2')
        elif inst == '.':
            # exchange OP0 and OP1
            dummy = copy.deepcopy(self.ops[0])
            self.ops[0] = copy.deepcopy(self.ops[1])
            self.ops[1] = copy.deepcopy(dummy)
            self.poutput('Exchanged: OP0 <--> OP1')
        elif inst == '^':
            # perform circular change upwards
            dummy = copy.deepcopy(self.ops[2])
            self.ops[2] = copy.deepcopy(self.ops[1])
            self.ops[1] = copy.deepcopy(self.ops[0])
            self.ops[0] = copy.deepcopy(dummy)
            self.poutput('Circular change performed: OP0 --> OP1 --> OP2, OP2 --> OP0')
        else:
            self.perror('Unknown instruction token: "%s"'%inst, traceback_war=False)
            return True # Flush the cmdqueue
    #END do_opsjuggle

    def do_absolute(self, arg):
        '''Take the absolute value of the curve in OP<opi>.
    Written for the display of fitted standard variations which sometimes
    are iterated into the negative, but only the square is in the model.
    SYNTAX: absolute,<opi>[1]
        <opi> : operand array from (0,1,2)
        '''
        arglist = self._longlist(arg,1)
        opi = self._setdefault(arglist[0], '1')
        if opi not in ('0','1','2'):
            self.perror('Operand index must be 0, 1 or 2', traceback_war=False)
            return True
        opi = int(opi)
        n = self.ops[opi].n
        if n > 0:
            self.ops[opi].y[:n] = abs( self.ops[opi].y[:n] )
            self.poutput('OP%i: y = ABS(y) '%opi)
    #END do_absolute()

    def do_delpoints(self, arg):
        '''Delete data points from a curve.
    SYNTAX: delpoints,<opi>,<head>,<tail>
            <opi>: 0, 1 or 2. The curve to act on. default: 1
            <head>: How far to delete from the front of the curve. Default:
                    0 (do not delete)
            <tail>  How far to delete from the tail of the curve. Default:
                    0 (do not delete)
     Meanings of head and tail:
            value     meaning
            <number>  delete int(number+0.5) points starting from head or tail
            x<number> delete starting from head or tail until the x-value
                      reaches <number>
            -         delete starting from head or tail while the y-value <= 0
                      (I.e. remove insignificant points marked non-positive) 
        '''
        arglist = self._longlist(arg,3) # Command has 3 arguments
        # Handle the operand
        opis = self._setdefault(arglist[0], u'1')
        opi, err = self._toopi(opis)
        if err != '':
            self.perror( err, traceback_war=False )
            return True # Flush cmd queue and return to prompt
        if self.ops[opi].isempty():
            self.perror('Curve array OP%i is empty'%opi, traceback_war=False)
            return True # Flush cmd queue and return to prompt
        self.poutput('delpoints from OP%i'%opi)

        # Process the tail of the operand first
        tail = u''+self._setdefault(arglist[2], u'0')
        if tail == '-':
            # remove trailing irrelevant points (x-values <= 0)
            idx = self.ops[opi].n - 1 # upper valid index
            while (idx >= 0) and (self.ops[opi].y[idx] <= 0.0):
                idx = idx - 1
            # No shift necessary. Only bookkeeping
            era = self.ops[opi].n - 1 - idx
            self.poutput('Tail: %i insignificant points deleted'%era)
            self.ops[opi].n = idx + 1
            if idx + 1 == 0:
                # All points erased: clean and return
                self.poutput('Curve array became empty')
                self.ops[opi].clean()
                return
            pass
        elif tail[0] == 'x':
            tailval, err = self._tonumber(tail[1:])
            if err != '':
                self.perror( err, traceback_war=False )
                return True # Flush cmd queue and return to prompt
            # Find index with x-value
            idx = self.ops[opi].n - 1
            while (idx >= 0) and (self.ops[opi].x[idx] > tailval):
                idx = idx - 1
        else:
            idx, err = self._toint( tail )
            if err != '':
                self.perror( err, traceback_war=False )
                return True # Flush cmd queue and return to prompt
            idx = self.ops[opi].n - idx # counted from the end
        if tail not in ('0', '-'):
            # special cases already treated: Do nothing; scan insignificant
            # now idx is the lower bound of the index range
            points = self.ops[opi].n
            if idx in range( points ):
                self.ops[opi].delirange( idx, points-1 )
                era = points - idx # points must be the old number of points
                self.poutput('Tail: %i points deleted'%era)
                if self.ops[opi].n <= 0:
                    self.poutput('Tail: The operation removed ALL points')
            else:
                # idx is not a valid index
                self.poutput( 'Tail: Request could not be served' )
        # tail was no special case

        # Process the head argument in similar manner -- but here we MUST shift
        head = u''+self._setdefault(arglist[1], u'0')
        if head == '-':
            # remove irrelevant points (x-values <= 0) at the beginning of the curve
            idx = 0 # lower valid index
            points = self.ops[opi].n
            while (idx < points) and (self.ops[opi].y[idx] <= 0.0):
                idx = idx +1
            self.poutput('Head: Deleting %i insignificant points'%idx)
            if idx > 0:
                self.ops[opi].delirange( 0, idx-1 )
                if self.ops[opi].n <= 0:
                    # All points erased: clean and return
                    self.poutput('Head: The operation removed ALL points')
                    self.ops[opi].clean()
                    return
            pass
        elif head[0] == 'x':
            headval, err = self._tonumber(head[1:])
            if err != '':
                self.perror( err, traceback_war=False )
                return True # Flush cmd queue and return to prompt
            # Find index with x-value
            points = self.ops[opi].n
            idx = 0
            while (idx < points) and (self.ops[opi].x[idx] < headval):
                idx = idx + 1
            idx = idx - 1
        else:
            idx, err = self._toint( head )
            idx = idx - 1 # silly input is caught later
            if err != '':
                self.perror( err, traceback_war=False )
                return True # Flush cmd queue and return to prompt
        if head not in ('0', '-'):
            # Not a special case (Do nothing, scanned from the front)
            # now idx is the upper bound of the index range to be deleted
            if idx in range( self.ops[opi].n ): # Do nothing if input is silly
                self.ops[opi].delirange( 0, idx )
                idx = idx + 1
                if self.ops[opi].n <= 0:
                    self.poutput('Head: ALL points deleted')
                else:
                    self.poutput('Head: %i points deleted'%idx)
            else:
                # idx is not a valid index into the curve data
                self.poutput( 'Head: Request could not be served' )
            pass
        #end: Was no special case
    #END do_delpoints

    def do_condense(self, arg):
        '''Reduce the number of points in a curve by averaging groups of 2 to 20 points
    and storing each average in a single new point.
    SYNTAX: condense,<opi>,<combine>
            <opi> : operand index from (0,1,2)
            DEFAULT: <opi>=1
            <combine> number of points to be averaged (2..20)
            DEFAULT: <combine>=2
        '''
        arglist = self._longlist(arg,2) # Two arguments
        # Handle the operand
        opis = self._setdefault(arglist[0], u'1')
        opi, err = self._toopi(opis)
        if err != '':
            self.perror( err, traceback_war=False )
            return True # Flush cmd queue and return to prompt
        box = u''+self._setdefault(arglist[1], u'2')
        if not box.isnumeric(): # isnumeric() is just o.k. for integers
            self.perror('<combine> must be numeric, but is "%s"'%box, traceback_war=False)
            return True # Flush the cmdqueue
        box = int(box)
        if box < 2:
            self.perror('<combine> must be > 1, but is "%s"'%box, traceback_war=False)
            return True
        if box > 20:
            self.perror('<combine> must be < 21, but is "%s"'%box, traceback_war=False)
            return True
        points = self.ops[opi].n
        if 3*box >= points:
            self.perror('3*<combine> must be < points in OP%i. Points="%i"'%(opi,points), traceback_war=False)
            return True
        # start to average
        boxf = 1.0*box
        k = 1 # write index
        for i in range(1,points-box/2-1,box):
            sumx, sumy = [0,0]
            for j in range(i,i+box):
                sumx = sumx + self.ops[opi].x[j]
                sumy = sumy + self.ops[opi].y[j]
            sumx = sumx / boxf
            sumy = sumy / boxf
            self.ops[opi].x[k] = sumx
            self.ops[opi].y[k] = sumy
            k = k+1
        #NOTICE At the end a bit of the interval may be
        #       lost -- but not preserving the last point
        # makes the command more robust if it is applied
        # repeatedly on the same curve
        #bookkeeping
        self.ops[opi].n = k
        self.poutput('Curve in OP%i condensed to %i points'%(opi,k))
    #END do_condense

    def __filtint_iterwarn(self):
        self.pfeedback('WARNING: Iterating background correction several times')
        self.pfeedback('         will remove more and more scattering of rough')
        self.pfeedback('         or imperfect domains. Tell readers, if you do.')
    #END __filtint_iterwarn()


    def do_filtint(self, arg):
        '''Make interference function OP1 --> OP0 by spatial frequency filtering of
    I(s)*s**p. The curve must start at 0, must have a number of points that is
    a reasonable power of 2, and must be equidistantly spaced ("channeled").
    This can be achieved by &comb. Hint: Single-step inspection by &lowpass,
    which performs a single lowpass filter determination.
        SYNTAX: filtint, <ItStop>[1],<cutoff>[3].<order>[1]
                <ItStop>: a number that defines the condition which stops the
                          iterative determination.
                          <ItStop> > 0: Perform a defined number of iterations
                                        and stop.
                          <ItStop> < 0: Iterate, until the integral of the
                          interference function is smaller than this user
                          supplied fraction of the integral without
                          iteration. Thus "-1" is no iteration, "-0.2" is:
                          improve the integral to 20% of the initial value.
                <cutoff>: the low-pass cutoff channel with respect to a curve
                          of 256 channels (is internally adapted to return
                          the same filtering for shorter or longer curves)
                <order>:  The order of the low-pass Butterworth filter
        The command determines a Butterworth background, subtracts it, applies
        the Hanning operator and continues to do so until the <ItStop>
        condition is fulfilled. The result is put in OP0
        '''
        mcf = 128 # maximum cutoff channel
        mord = 20 # maximum filter order
        arglist = self._longlist(arg,3) # Command has 3 arguments
        # Check inputcurve
        opi = 1
        opo = 0
        inputcurve = self.ops[opi]
        outcurve = self.ops[opo]
        span = inputcurve.n
        if (not inputcurve._hasenoughpowerof2points()):
            self.perror( 'OP%i must contain 2**n points but has %i'%(opi,span), traceback_war=False )
            return True # Flush cmd queue and return to prompt
        itstops = self._setdefault(arglist[0],u'1')
        itstop, err = self._tonumber(itstops)
        if err != '':
            self.perror( err, traceback_war=False )
            return True # Flush cmd queue and return to prompt
        if abs(itstop) > 20.0:
            self.perror( 'ItStop: Absolute value >20 not allowed', traceback_war=False )
            return True # Flush cmd queue and return to prompt
        cuts = self._setdefault(arglist[1], u'3')
        cutchan, err = self._tonumber( cuts )
        if err != '':
            self.perror( err, traceback_war=False )
            return True # Flush cmd queue and return to prompt
        if cutchan < 1.0:
            cutchan = 1.0
            self.poutput('Cutoff channel increased to %i'%cutchan)
        if cutchan > mcf:
            cutchan = mcf
            self.poutput('Cutoff channel reduced to %i'%cutchan)
        ords = self._setdefault(arglist[2], u'1')
        order, err = self._toint( ords)
        if err != '':
            self.perror( err, traceback_war=False )
            return True # Flush cmd queue and return to prompt
        if order < 1:
            order = 1
            self.poutput('Order increased to %i'%order)
        if order > mord:
            order = mord
            self.poutput('Order decreased to %i'%order)
        # Argument checks finished

        # Generate the Hanning filter
        pilin = np.linspace(0,np.pi,span) # Linear spaced values 0 .. pi
        Hanning = 0.5 * (1.0 + np.cos(pilin)) # The Hanning "cosine-bell function"
        filcurve = CurveClass( self._MAXP ) # cache for the result of filtering
        self.ops[opo] = copy.deepcopy(inputcurve) # Copy input to output curve
        outcurve = self.ops[opo] # set the pointer after copy
        # Compute the first "curved background" in filcurve
        filcurve.lowpass(inputcurve,cutchan,order)
        # Subtract background and hann all valid curve points
        outcurve.y[:span] = (outcurve.y[:span] - filcurve.y[:span])*Hanning
        # Compute integral
        integ0 = outcurve.y[:span].sum()
        self.poutput('Start integral: %9.3e'%integ0)
        if itstop >= 0:
            # *** User has requested a fixed number of iterations
            countdown = int(itstop)
            endit = countdown
            if countdown > 0:
                if countdown > 1:
                    self.__filtint_iterwarn()
                self.poutput('No. rel integral')
                relint = integ0/abs(integ0)
                self.poutput('  0   %6.3f'%relint)
            else:
                self.poutput('Background corr without iteration')
            while countdown > 0:
                # Proceed on the already corrected curve
                filcurve.lowpass(outcurve,cutchan,order)
                # Subtract background and hann all valid curve points
                outcurve.y[:span] = (outcurve.y[:span] - filcurve.y[:span])*Hanning
                # Compute integral
                integral = outcurve.y[:span].sum()
                relint = integral / abs(integ0)
                it = endit - countdown + 1
                self.poutput('%3i   %6.3f'%(it,relint))
                countdown = countdown - 1
            pass
        else:
            # *** User has requested to iterate until the integral of the
            #     interference function becomes (almost) zero
            relintlevel = abs( itstop )
            relint = integ0/abs(integ0)
            it = 0
            self.__filtint_iterwarn()
            self.poutput('No. rel integral')
            self.poutput('  0   %6.3f'%relint)
            while (it < 20) and (abs(relint) > relintlevel):
                # Proceed on the already corrected curve
                filcurve.lowpass(outcurve,cutchan,order)
                # Subtract background and hann all valid curve points
                outcurve.y[:span] = (outcurve.y[:span] - filcurve.y[:span])*Hanning
                # Compute integral
                integral = outcurve.y[:span].sum()
                relint = integral / abs(integ0)
                it = it + 1
                self.poutput('%3i   %6.3f'%(it,relint))
        pass
    #END do_filtint

    def do_inter(self, arg):
        '''Make interference function (old INT cmd)
Not yet implemented'''
        #TODO
        self.poutput('Not (yet) implemented')
        pass
    #END do_inter

    def do_longperiod(self, arg):
        '''Was implemented in the sf-package.
The command is not yet implemented'''
        #TODO
        self.poutput('Not yet implemented')
        pass
    #END do_longperiod

    def do_lowpass(self, arg):
        '''Low-pass filtering of a curve in OP1 by means of a Butterworth filter.
    The curve must have 2**n equidistantly spaced data points starting from
    x=0. This requirement can be fulfilled by application of the command
    &comb to suitable scattering data (i.e. intensities multiplied by a
    power of s).
    SYNTAX: lowpass,opi[1],<cc>[3],<order>[1]
            <opi> (0,1,2)   index of the curve operand array
            <cc> (1..255)   cutoff channel of the Butterworth filter
               with respect to a pseudo curve of 256 equidistantly
               spaced points. Thus the effective relative cutoff
               frequency is cc/256
            <order> (1..30) order of the Butterworth filter
    Typical call: low,1,2 (filter OP1 with narrower a band)
    Notice: The Method of the command can be applied in the iPython console,
    if the original curve shall not be overwritten. E.g. for result in OP0:
    SASCUR> read,combedcurve
    SASkia> ipy
    In [1]: cc=2
    In [2]: order=1
    In [3]: self.ops[0].lowpass(self.ops[1],cc,order)'''
        mcf = 255.0 # maximum cutoff frequency
        mord = 30 # maximum filter order
        arglist = self._longlist(arg,3) # Three arguments
        opis = self._setdefault(arglist[0], u'1')
        opi, err = self._toopi(opis)
        if err != '':
            self.perror( err, traceback_war=False )
            return True # Flush cmd queue and return to prompt
        thecurve = self.ops[opi]
        span = thecurve.n
        if (not thecurve._hasenoughpowerof2points()):
            self.perror( 'OP%i must contain 2**n points but has %i'%(opi,span), traceback_war=False )
            return True # Flush cmd queue and return to prompt
        cuts = self._setdefault(arglist[1], u'3')
        cutchan, err = self._tonumber( cuts )
        if err != '':
            self.perror( err, traceback_war=False )
            return True # Flush cmd queue and return to prompt
        if cutchan < 1.0:
            cutchan = 1.0
            self.poutput('Cutoff channel increased to %i'%cutchan)
        if cutchan > mcf:
            cutchan = mcf
            self.poutput('Cutoff channel reduced to %i'%cutchan)
        ords = self._setdefault(arglist[2], u'1')
        order, err = self._toint( ords)
        if err != '':
            self.perror( err, traceback_war=False )
            return True # Flush cmd queue and return to prompt
        if order < 1:
            order = 1
            self.poutput('Order increased to %i'%order)
        if order > mord:
            order = mord
            self.poutput('Order decreased to %i'%order)
        # Filter the curve in place
        thecurve.lowpass(thecurve,cutchan,order)
        self.poutput('OP%i: Low-pass Butterworth filter of order %i applied.'%(opi,order))
        self.poutput('     Relative cutoff frequency: %.1f/256. (%i points)'%(cutchan,span))
        pass
    #END do_lowpass

    def do_similar( self, arg ):
        '''Assess similarity between the curves in OP1 and OP2. Result: a mask in
    OP0. It is zero where the two curves are considered not to be similar. For
    the ease of visual inspection the mask is multiplied by 90% of the maximum
    y-value of OP 2.
    The curves must contain the same number of points. It is assumed that they
    are neighboring frames from an experimental run and that therefore they
    should be rather similar, and that dissimilarities are from technical
    reason (cosmic ray spikes etc.).
    SYNTAX: similar,<threshold>[100],<fringe>[100],<minextent>=[3]
    <threshold>: The percentage of permitted change of a point in the curve
    <fringe>:    The number of points to the right and to the left which are
                 considered to be affected, as well.
    <minextent>: minimum extent of a defect. Narrower defects are not marked.

    The result in OP0 can be used to bridge the defects in the damaged curve.'''

        if self.ops[1].n != self.ops[2].n:
            err = 'Numbers of points in OP1 and OP2 are different'
            self.perror( err, traceback_war=False )
            return True

        if self.ops[1].n < 3:
            err = 'Trivial or empty curves!'
            self.perror( err, traceback_war=False )
            return True

        arglist = self._longlist(arg,3) # Three arguments
        ## (1)
        threshs = self._setdefault(arglist[0], '100')
        thresh, err = self._tonumber( threshs )
        if err != '':
            self.perror( err, traceback_war=False )
            return True # Flush cmd queue and return to prompt
        thresh = 0.01*thresh # From percentage to fraction
        ## (2)
        fringes = self._setdefault(arglist[1], '100')
        fringe, err = self._toint( fringes )
        if err != '':
            self.perror( err, traceback_war=False )
            return True # Flush cmd queue and return to prompt
        ## (3)
        minexs = self._setdefault(arglist[2], '3')
        minex, err = self._toint( minexs )
        if err != '':
            self.perror( err, traceback_war=False )
            return True # Flush cmd queue and return to prompt

        self.ops[0] = self.ops[1].similarto(self.ops[2],thresh,fringe,minex)
        res = self.ops[0]
        n = res.n
        goodies = sum(res.y[:n])
        persim = 100.0*goodies/n
        self.poutput('%5.1f%%  of the %i curve points in OP1 and OP2 are'%(persim,n))
        self.poutput('rendered similar. A bridging mask is now in OP0')
        res.y[:n] = res.y[:n] * 0.9*max(self.ops[2].y[:n])
        pass
    #END do_similar

    def do_bridge( self, arg ):
        '''Follow-up command to "similar". The mask in OP0 i used to mark and remove
    defective points in one of the other OPs. The missing data are bridged by
    spline interpolation.
    SYNTAX: bridge,<opi>[1],<smoothness>[0.1]
        <opi>        1 or 2 (result is in OP0!)
        <smoothness> is the smoothing parameter. A low value should be good.
        '''
        arglist = self._longlist(arg,2) # Two arguments
        opis = self._setdefault(arglist[0], u'1')
        opi, err = self._toopi(opis)
        if err != '':
            self.perror( err, traceback_war=False )
            return True # Flush cmd queue and return to prompt
        if opi == 0:
            self.perror( 'OP0 is not allowed!', traceback_war=False )
            return True # Flush cmd queue and return to prompt
        act  = self.ops[opi]
        mask = self.ops[0]
        if act.n != mask.n:
            err = 'Number of points in OP0 and OP%i differ'%opi
            self.perror( err, traceback_war=False )
            return True # Flush cmd queue and return to prompt

        smos = self._setdefault(arglist[1], u'1')
        smo, err = self._tonumber( smos )
        if err != '':
            self.perror( err, traceback_war=False )
            return True # Flush cmd queue and return to prompt
        smo = abs(smo)
        err = act.bridgewith( mask, smo )
        return err
    #END do_bridge

    def do_resample( self, arg ):
        '''Resample the curve in OP0 so that it becomes defined at exactly the
    same points as is the curve in the template <opi>. smp is an optional
    smoothing parameter which normally should remain very small
    (indicating interpolation).
    Intention: If backgrounds are defined at other positions than the sample
    curves, they must be resampled. This happened with ALBA data, because
    backgrounds had to be redone from the 2D-patterns and came out differently
    sampled as compared to what the online process did.

    SYNTAX: resample,<opi>[1],<smoothness>[0.1]
        <opi>: operand of the template
        <smoothness> is the smoothing parameter. A low one should be good.'''

        arglist = self._longlist(arg,2) # Two arguments
        opis = self._setdefault(arglist[0], u'1')
        opi, err = self._toopi(opis)
        if err != '':
            self.perror( err, traceback_war=False )
            return True # Flush cmd queue and return to prompt
        if opi == 0:
            self.perror( 'OP0 is not allowed as template!', traceback_war=False )
            return True # Flush cmd queue and return to prompt
        template  = self.ops[opi]
        work = self.ops[0]

        smos = self._setdefault(arglist[1], u'1')
        smo, err = self._tonumber( smos )
        if err != '':
            self.perror( err, traceback_war=False )
            return True # Flush cmd queue and return to prompt
        smo = abs(smo)
        err = work.resample( template, smo )
        if err != '':
            self.perror( err, traceback_war=False )
            return True # Flush cmd queue and return to prompt
        return
    #END do_resample

    def do_multiply(self, arg):
        '''Multiplication of curve data.
    SYNTAX: multiply, <factor>[1], <opi>[1], <ypower>[0], <xpower>[1]
        Default parameter values are shown in brackets. They are the neutral
        values, which do not change the curve.
        Typical call: mul,,,4 (replaces y-values in OP1 with y*x**4)
        The y-values are subjected to the transformation:
            y[i] = y[i] * <factor> * x[i]**<ypower>
        The x-values are subjected to the transformation:
            x[i] = x[i]**<xpower>
        A transformation with <xpower>=0 is refused.
        if <ypower> is < 0.0, points with x[i] <= 0.0 are deleted before the
        transformation.
        <factor>: here it is allowed to write "1/<number>",
                  e.g.     &multiply,1/2.01
                  In this case the inverse is computed. This is useful for the
                  normalization to the sample thickness (2.01 mm)
        '''
        arglist = self._longlist(arg,4) # Four arguments
        facts = self._setdefault(arglist[0], u'1')
        inverse = facts[0:2] == '1/'
        if inverse:
            facts = facts[2:]
        factor, err = self._tonumber( facts )
        if err != '':
            self.perror( err, traceback_war=False )
            return True # Flush cmd queue and return to prompt
        if inverse:
            if factor == 0.0:
                self.perror("I go home until you tell me the inverse of 0!", traceback_war=False)
                return True
            factor = 1.0/factor
        opis = self._setdefault(arglist[1], u'1')
        opi, err = self._toopi(opis)
        if err != '':
            self.perror( err, traceback_war=False )
            return True # Flush cmd queue and return to prompt
        thecurve = self.ops[opi]
        if thecurve.isempty():
            self.perror( 'OP%i is empty.'%opi, traceback_war=False )
            return True # Flush cmd queue and return to prompt
        ypows = self._setdefault(arglist[2],u'0')
        ypower, err = self._tonumber( ypows )
        if err != '':
            self.perror( err, traceback_war=False )
            return True # Flush cmd queue and return to prompt
        xpows = self._setdefault(arglist[3],u'1')
        xpower, err = self._tonumber( xpows )
        if err != '':
            self.perror( err, traceback_war=False )
            return True # Flush cmd queue and return to prompt
        thecurve.multiply( factor, ypower, xpower )
        # Work report
        self.poutput('OP%i:    y = y '%opi, ' ')
        if factor != 1.0:
            self.poutput('* %9.3e'%factor,' ')
        if ypower != 0.0:
            self.poutput(' * x**%9.3e'%ypower, ' ')
        if xpower == 1:
            self.poutput(' ') # New line
        else:
            self.poutput('   |   x  <--  x**%9.3e'%xpower)
        pass
    #END do_multiply

    def do_munfold(self, arg):
        '''Mellin-unfold the 2D chord distribution g2(x12) of an ensemble of
    uncorrelated circular disks from the intrinsic chord length distribution
    of any circular disk. The result is the diameter distribution hD(x12) of
    the disks. See also anacircles'''
        #TODO
        self.poutput('Not yet implemented')
        pass
    #END do_munfold

    def do_anacircles(self, arg):
        '''Analyze the 2D chord distribution of uncorrelated circles directly
    without munfolding by resorting to the property of the Mellin convolution
    that its moments multiply.'''
        #TODO
        self.poutput('Not yet implemented')
        pass
    #END do_anacircles

    def do_projiso(self, arg):
        '''Isotropic data from an D-dimensional space are projected on an
    (D-1)-dimensional subspace by means of the Abel transformation. E.g. a
    SAS curve I(s) which is isotropic in 3D is projected on a plane resulting
    in [I]_2(s_{12}. A second call projects on a line resulting in
    [I]_1(s_1) = [I]_1(s_2). It has not escaped me that this mechanism touches
    on a fundamental philosophical question: Only when we interpret an
    isotropic curve, we make a decision on the dimensionality of the space
    in which we consider it isotropic. n-dimensional data are projected on a
    line by (n-1)fold application of this command (e.g. the Abel transform).
        SYNTAX: projiso,<opi>[1]
            The curve must start at 0 and must be equidistantly spaced. This
            is achieved by &comb.
            PROCEDURE: The curve is first extrapolated to sqrt(2)*xmax by
                       a line in order to minimize cutoff effects.
        '''
        arglist = self._longlist(arg,1)
        opis = self._setdefault(arglist[0],u'1')
        opi, err = self._toopi(opis)
        if err != '':
            self.perror( err, traceback_war=False )
            return True # Flush cmd queue and return to prompt
        ciscurve = self.ops[opi]
        if ciscurve.n < 32:
            self.perror( 'OP1 is empty or has < 32 points', traceback_war=False )
            return True # Flush cmd queue and return to prompt
        if ciscurve.x[0] != 0.0:
            self.perror('Curve in OP1 does not start at s=0. &comb', traceback_war=False)
            return True
        span = ciscurve.n
        equids = ciscurve.x[span-1]/(span-1)
        test   = equids/ciscurve.x[1]
        if (test < 0.99) or (test > 1.01):
            self.perror('Curve in OP1 is not equidistantly spaced. &comb', traceback_war=False)
            return True
        ulim = ciscurve.x[ciscurve.n-1]
        ciscurve.ext15( 1, 0 ) # Prolong by a straight line "into the corners"
        # Apply Abel transformation
        self.ops[opi] = copy.deepcopy(ciscurve.abel())
        # Cut off the extrapolated points
        # Find index with ulim
        idx = self.ops[opi].n - 1
        while (idx >= 0) and (self.ops[opi].x[idx] > ulim):
            idx = idx - 1
        self.ops[opi].n = idx + 1
        self.poutput('OP%i: Isotropic curve data projected from D-space to (D-1) space.'%opi)
        pass
    #END do_projiso

    def do_physpace(self, arg):
        '''Transformations from reciprocal to physical space.
    reciprocal-space data in OP1  --->  OP2 receives real-space data.
    SYNTAX: physpace,<rmax>[100],<dr>[0.5],<kind>[1]
        <rmax>: maximum r value in the result (in units of nm)
        <dr>:   delta-r, the spacing between the data points in the result.
                Thus the number of points in the result becomes
                    self.ops[2].n = <rmax>/<dr> + 1
        <kind>  Result in OP2
        ------  -------------
           1    Simple negative Fourier transform g1(r) = - FT( G1(s) )(r).
                If the input is the interference function G1(s) of a
                one-dimensional scattering curve I1(s), the result is the
                IDF (Ruland). Apply to G1(s3) from projected [I(s)]1(s3)
                and 1D Porod-power law (s**2) or to G1(s) of a morphology
                comprising lamellae stacks (3D Porod-power law (s**4)).
                Computation of the 1D correlation function simply starts
                from the corresponding 1D intensity, but then the result
                must be flipped upside down (&mul,-1,2) and normalized.
           2    g2(x12), the 2D-IDF (chord length distribution) computed from
                G2(s12) of a fiber diagram projected on the equatorial plane
                s12. The Porod law of this projection scales with s^(-3)'''
        arglist = self._longlist(arg,3) # Three arguments
        # Is there a curve in OP1?
        ciscurve = self.ops[1]
        if ciscurve.n < 32:
            self.perror( 'OP1 is empty or has < 32 points', traceback_war=False )
            return True # Flush cmd queue and return to prompt
        if ciscurve.x[0] != 0.0:
            self.perror('Curve in OP1 does not start at s=0', traceback_war=False)
            return True
        span = ciscurve.n
        equids = ciscurve.x[span-1]/(span-1)
        test   = equids/ciscurve.x[1]
        if (test < 0.99) or (test > 1.01):
            self.perror('Curve in OP1 is not equidistantly spaced', traceback_war=False)
            return True
        rmaxs = self._setdefault(arglist[0], u'100')
        rmax, err = self._tonumber( rmaxs )
        if err != '':
            self.perror( err, traceback_war=False )
            return True # Flush cmd queue and return to prompt
        drs = self._setdefault(arglist[1], u'0.5')
        dr, err = self._tonumber( drs )
        if err != '':
            self.perror( err, traceback_war=False )
            return True # Flush cmd queue and return to prompt
        rmax = abs( rmax )
        dr   = abs( dr )
        if dr < 0.01:
            self.perror('dr smaller than 0.01 nm is not accepted', traceback_war=False)
            return True # Flush cmd queue and return to prompt
        spant = int(rmax/dr) + 1
        if spant < 10:
            self.perror('Result curve would contain less than 10 points', traceback_war=False)
            return True
        mxp = len(ciscurve.x)
        if spant > mxp:
            self.perror('Result would contain %i > %i points'%(spant,mxp), traceback_war=False)
        kinds = self._setdefault(arglist[2], u'1')
        kind, err = self._toint( kinds )
        if err != '':
            self.perror( err, traceback_war=False )
            return True # Flush cmd queue and return to prompt
        # End of parameter and state checks
        # Condition the ciscurve to avoid rounding errors
        up = abs( max( ciscurve.y[:span] ) )
        dn = abs( min( ciscurve.y[:span] ) )
        cond = 1.0 / max( up, dn ) # conditioning factor
        ciscurve.y[:span] = ciscurve.y[:span]*cond
        # Instantiate a curve transformer
        curtrans = CurveTransformerClass()
        # Answer the request as a function of "kind"
        self.poutput('Transforming the curve in OP1 from reciprocal to real space by')
        if kind == 1:
            # Instantiate the 1D Fourier transform of an isotropic curve
            thekernel = FouIso1KernelClass()
            # Transform the ciscurve
            self.poutput('a simple Fourier transform 1dimensional G1(s) to IDF g1(r),')
            self.poutput('or a 1D intensity I1(s) to a 1D correlation function. For the')
            self.poutput('latter the result must be multiplied by -1 and scaled.')
            err, transcurve = curtrans.transform( ciscurve, thekernel, spant, dr )
            transcurve.y[:spant] = -transcurve.y[:spant]
        elif kind == 2:
            thekernel = G22DKernelClass()
            self.poutput('the 2nd derivative of the Fourier kernel (2dimensional G2(s12)')
            self.poutput('as projected to the equatorial plane is transformed into the')
            self.poutput('2dimensional g2(r12) interface distribution.')
        else:
            ciscurve.y[:span] = ciscurve.y[:span]/cond # Revert conditioning
            self.perror('Transformation type kind=%i is not implemented'%kind, traceback_war=False)
            return True
        # Transformation finished
        if err != '':
            self.perror('An error occured during the transformation')
            ciscurve.y[:span] = ciscurve.y[:span]/cond # Revert conditioning
            self.perror( err, traceback_war=False )
            return True
        # Transformation successful
        ciscurve.y[:span] = ciscurve.y[:span]/cond # Revert conditioning
        self.ops[2] = copy.deepcopy( self.ops[1] ) # For all the metadata
        self.ops[2].n = transcurve.n # The curve data
        self.ops[2].y = transcurve.y
        self.ops[2].y[:spant] = transcurve.y[:spant]/cond # Insert and recondition t-data
        self.ops[2].x[:spant] = transcurve.x[:spant] # Insert valid t-data
        pass
    #END do_physpace

    def do_rename(self, arg):
        '''Rename a curve which resides in one of the operands by overwriting the
    self.ops[opi].lastfile string. The <opi> must be allowed and must contain
    a curve.
    SYNTAX: rename, <opi>[1], filename[]
        <opi>      0,1, or 2
        <filename> = '': Output the actual default name of the curve in <opi>
                   != '': New default name for the curve in <opi>. If the
                   string does not contain a dot, the default extension of a
                   curve file (".dat") is appended.
        '''
        arglist = self._longlist(arg,2) # Two arguments
        opis = self._setdefault(arglist[0], u'1')
        opi, err = self._toopi(opis)
        if err != '':
            self.perror( err, traceback_war=False )
            return True # Flush cmd queue and return to prompt
        filename = self._setdefault(arglist[1], '')
        thecurve = self.ops[opi]
        if thecurve.isempty():
            self.perror('No curve in OP%i'%opi, traceback_war=False)
            return True # Flush cmd queue and return to prompt
        if filename != '':
            # Append default extension, if no extension given
            if filename.rfind('.') < 0:
                filename = filename + '.dat'
            thecurve.lastfile = filename
            self.poutput('NEW default file name of curve in OP%i: "%s" '%(opi,filename))
        else:
            filename = thecurve.lastfile
            self.poutput('Default file name of curve in OP%i is "%s" '%(opi,filename))
    #END do_rename

    def do_sbx(self, arg):
        '''Subtract from x-values of a curve. This shifts the curve in x-direction
        SYNTAX: sbx, <value>, [<opi>]
            <value>: a number (x-value which becomes the new zero)
            <opi>  : '0', '1' or '2'. Defaults to '1' (The operand to work on)
        '''
        arglist = self._longlist(arg,2) # Two arguments
        value, err = self._tonumber(arglist[0])
        if err != '':
            self.perror( err, traceback_war=False )
            return True # Flush cmd queue and return to prompt
        opis = self._setdefault(arglist[1], u'1')
        opi, err  = self._toopi( opis )
        if err != '':
            self.perror( err, traceback_war=False )
            return True # Flush cmd queue and return to prompt
        if self.ops[opi].isempty():
            self.perror('OP%i is empty'%opi, traceback_war=False)
            return True # Flush cmd queue and return to prompt
        self.ops[opi].subcx( value )
        self.poutput('All x-values in OP%i reduced by %9.3e'%(opi,value))
        pass
    #END do_sbx

    def do_sby(self, arg):
        '''Subtract from y-values of a curve
           SYNTAX: sby, <value>, [<opi>]
               <value> : a number or "!", the system variable indicator
               <opi>    : '0','1','2'
           If <opi> not supplied: From OP1 subtract a machine background
           provided in OP2 and take <value> as the absorption factor
               sby,<value> : OP1 = OP1 - <value> * OP2
           if <opi> is supplied by the user: Subtract the constant <value>
           from all the y-values in OP<opi>:
               OP<opi> = OP<opi> - <value>
           If <value> is an exclamation mark, the value stored in the
           system variable self.ThePlot.backgr is used. This is the last
           value which has interactively been determined by the user.
           '''
        arglist = self._longlist(arg,2) # Two arguments
        grabsystem = arglist[0] == '!'
        if grabsystem:
            value = self.ThePlot.backgr
            self.poutput('Take the fluctuation background self.ThePlot.backgr: %9.3e'%value)
        else:
            value, err = self._tonumber(arglist[0])
            if err != '':
                self.perror( err, traceback_war=False )
                return True # Flush cmd queue and return to prompt
        opis = arglist[1]
        machineback = opis in ('=','') # missing or explicit default request
        if machineback:
            # Subtract machine background multiplied by absorption factor
            if grabsystem:
                self.perror( 'A fluctuation background is no absorption factor.', traceback_war=False )
                return True
            if self.ops[1].isempty() or self.ops[2].isempty():
                self.perror( 'Both OP1 and OP2 must contain data', traceback_war=False )
                return True
            # Multiply curve in OP2 by -value and add the result into OP1
            err = self.ops[1].addcurve(self.ops[2],-value)
            if err != None:
                self.perror( err, traceback_war=False )
                return True # Flush cmd queue and return to prompt
            else:
                self.poutput('Subtraction carried out; OP1 = OP1 - %9.3e * OP2'%value)
        else:
            # Subtract a constant "Fluctuation background" from the curve
            opi, err = self._toopi(opis)
            if err != '':
                self.perror( err, traceback_war=False )
                return True # Flush cmd queue and return to prompt
            thecurve = self.ops[opi]
            if thecurve.isempty():
                self.perror( 'OP%i is empty.'%opi, traceback_war=False )
                return True # Flush cmd queue and return to prompt
            span = thecurve.n
            thecurve.y[:span] = thecurve.y[:span] - value
            self.poutput('%9.3e subtracted from curve in OP%i'%(value, opi))
        pass
    #END do_sby

    def do_submin(self, arg):
        '''Preparation for the computation of a correlation function. Chris
        Vonk's advice: Cut off the intensity curve at the minimum between SAXS
        and WAXS and subtract the minimum. After that the curve becomes
        band-limited and is Fourier transformable. Moreover, the correlation
        function does not change much when more sophisticated pre-evaluation
        is chosen
        SYNTAX: submin,<opi>[1],<levelfact>[1.0],<rangefact>[1.0]
            <opi>: '0', '1' or '2'.  Which curve shall be treated
            <levelfact>: either ! (subtract the last interactively determined
                                    background from self.ThePlot.backgr)
                  or [0.0 .. 1.0]  multiply the minimal y-value in the curve
                                   by this value before subtraction
            <rangefact>: >1.0      multiply the x-value of the minimum by
                                   <rangefact> to determine the position of
                                   where to cut off
        The factor parameters make the command work less radical and may be
        useful in automated pre-processing (e.g. cutting off before the WAXS,
        subtracting the typical fluctuation background level) before smoothing
        (&median) or extrapolation (&ext15) is performed'''
        arglist = self._longlist(arg,3) # Three arguments
        opis = self._setdefault(arglist[0], u'1')
        opi, err  = self._toopi( opis )
        if err != '':
            self.perror( err, traceback_war=False )
            return True # Flush cmd queue and return to prompt
        if self.ops[opi].isempty():
            self.perror('OP%i is empty'%opi, traceback_war=False)
            return True # Flush cmd queue and return to prompt

        # Determine the minimum y-value (ymin) of the curve and
        # at which index (imin) and which x-value (xmin) it is found
        thecurve = self.ops[opi]
        span = thecurve.n
        ymin = min( thecurve.y[:span] ) # Only from the valid points
        mipos = np.where(thecurve.y[:span] == ymin) # array of lists
        mipol = mipos[-1] # the last list
        imin  = mipol[-1] # the last index in the curve where the minimum is taken
        xmin = thecurve.x[imin]

        # Prepare the subtraction business
        grabsystem = arglist[1] == '!'
        if grabsystem:
            level = self.ThePlot.backgr
            self.poutput('Subtract the fluctuation background self.ThePlot.backgr: %9.3e'%level)
        else:
            levels = self._setdefault( arglist[1],u'1.0')
            level, err = self._tonumber(levels)
            if err != '':
                self.perror( err, traceback_war=False )
                return True # Flush cmd queue and return to prompt
            level = abs( level )
            level = min( 1.0, level )
            # From relative to absolute level
            level = ymin*level

        # Prepare the cutting business
        cutfacts = self._setdefault(arglist[2],u'1.0')
        cutfact, err = self._tonumber(cutfacts)
        if err != '':
            self.perror( err, traceback_war=False )
            return True # Flush cmd queue and return to prompt
        cutfact = abs( cutfact )
        cutfact = max( 1.0, cutfact )
        if cutfact != 1.0:
            # Determine bigger imin
            nupos = np.where(thecurve.x[:span] < xmin*cutfact)
            nupol = nupos[-1]
            imin  = nupol[-1]
        span = imin+1
        self.ops[opi].n = span # Cutting off is done
        self.poutput('OP%i: Curve has now %i points'%(opi,span))

        # Do subtraction
        self.ops[opi].y[:span] = thecurve.y[:span] - level
        self.poutput('     ynew = yold - %8.3e'%level)
        pass
    #END do_submin

    def do_scatpower(self, arg):
        '''Determines the scattering power by integrating a curve, which is a
    one-dimensional intensity. This is the integrand. If the computation is
    successful, the result is stored in the attribute "Q" of the curve.
    Supported 1D-intensities are the projection of the intensity on a line or
    the "Lorentz-corrected", isotropic intensity in the notation s**2*I(s).
    So in the latter case the factor 2*Pi must still be considered.
    SYNTAX: scatpower,<opi>[1],<kind>[proj],<ap>[0.0]
        <opi>: Operand index (where the integrand is loaded) [0 .. 2], i.e.
               integrand = self.ops[opi]
        <kind> = proj: The curve is {I}_1(s), i.e. an intensity projected
                       on a line. In this case, with n=integrand.n,
                       Q_p = 2*(integrand.integrate() + ap/integrand.x[n-1])
               = ssI:  The curve is s**2*I(s), i.e. the "Lorentz-corrected"
                       intensity, and Q becomes Q = 2*Pi*Q_p
        <ap>:   The Porod asymptote ap = lim( s**2*integrand(s)). If a
                number is provided, the integral of the curve is increased by
                ap/s_{max}. Doing so extends the integral to infinity.
                
                If the string "est" is provided, <ap> = s_e^2 * y_e is esti-
                mated by assuming that the last point in the curve already
                fulfills Porod's law. Then the extension of the integral is
                simply y_e*s_e.'''
        arglist = self._longlist(arg,3) # Three arguments
        opis = self._setdefault(arglist[0], u'1')
        opi, err  = self._toopi( opis )
        if err != '':
            self.perror( err, traceback_war=False )
            return True # Flush cmd queue and return to prompt
        integrand = self.ops[opi]
        span = integrand.n
        if span < 3:
            self.perror('OP%i is (almost) empty'%opi, traceback_war=False)
            return True # Flush cmd queue and return to prompt
        if integrand.x[0]/integrand.x[span-1] > 0.01:
            self.perror('The curve in OP%i does not start at x=0'%opi, traceback_war=False)
            return True
        kind = self._setdefault(arglist[1], 'proj')
        aps  = self._setdefault(arglist[2], '0.0' )
        
        # Make Ap
        if aps[0:3].lower() == 'est':
            ap = integrand.x[span-1]**2 * integrand.y[span-1]
        else:
            ap, err = self._tonumber( aps )
            if err != '':
                self.perror( err, traceback_war=False )
                return True # Flush cmd queue and return to prompt

        kval = 2.0*(integrand.integral() + ap/integrand.x[span-1])
        if kind == 'ssI':
            kval = 2.0*np.pi*kval
        elif kind == 'proj':
            pass
        else:
            self.perror( 'Kind of integrand "%s" unknown'%kind, traceback_war=False )
            return True # Flush cmd queue and return to prompt
        self.ops[opi].env['Q'] = float(kval) # register the scattering power in the integrand
        self.poutput('Determined scattering power Q: %s'%kval)
        self.poutput('Hint: "Q" is now an &attribute of the integrand curve.')
        pass
    #END do_scatpower

    def do_smplain(self, arg):
        '''SMooth PLAIN: Smoothing of complete curves (i.e. not split into sets)
        by cubic splines. If no smoothing parameter <sp> is provided, the user
        asks for interactive determination.
        SYNTAX: smplain,<equi>[0],<derive>[],<fromOP>[1],<toOP>[2],<sp>[]
                <equi>: number of equidistant points to be generated along
                        the x-axis. For <equi> < 5 the original distribution
                        of the data points along the x-axis is not changed.
                <derive>: only if the string starts with "d", the output will
                          not be the smooth curve, but its first derivative.
                <fromOP>: The operand, in which the source of the smoothing
                          is found.
                <toOP>: The operand, where the result is stored. <toOP> is
                        overwritten.
                <sp>: Smoothing parameter. If given, the interactive part is
                      skipped.
                      <sp> = 0: Interpolating cubic spline
                      <sp> between about 1 and 10: natural smoothing.'''
        arglist = self._longlist(arg,5) # Five arguments
        fos = self._setdefault(arglist[2], u'1')
        fop, err  = self._toopi( fos )
        if err != '':
            self.perror( err, traceback_war=False )
            return True # Flush cmd queue and return to prompt
        if self.ops[fop].isempty():
            self.perror('OP%i is empty'%fop, traceback_war=False)
            return True # Flush cmd queue and return to prompt
        tos = self._setdefault(arglist[3], u'2')
        top, err  = self._toopi( tos )
        if err != '':
            self.perror( err, traceback_war=False )
            return True # Flush cmd queue and return to prompt
        ders = self._setdefault(arglist[1], 'x')
        derive = ders[0] == 'd'
        equis = self._setdefault(arglist[0], '0')
        equi, err = self._toint( equis )
        if err != '':
            self.perror( err, traceback_war=False )
            return True # Flush cmd queue and return to prompt
        if equi < 5:
            equi = 0
        sps = self._setdefault(arglist[4],'')
        interact = sps == ''
        if interact:
            sp = 1.0
        else:
            sp, err = self._tonumber( sps )
            if err != '':
                self.perror( err, traceback_war=False )
                return True # Flush cmd queue and return to prompt

        while interact:
            if six.PY2:
                spstr = raw_input('Enter smoothing parameter [%5.2f]: '%sp)
            else:
                spstr = input('Enter smoothing parameter [%5.2f]: '%sp)
            spstr.strip() # Remove white space on both sides
            good = spstr == '' # just pressing return is a good default
            if not good:
                good = True # What we expect
                try:
                    sp = float(spstr)
                except ValueError:
                    self.poutput("This was no number")
                    good = False # Expectation not fulfilled
            if good:
                answer = self.SmoPlot._showsmoothing(self.ops[1],self.fontfactor,sp,equi)
                interact = answer != 'y' # not yet accepted by the user
            pass
        # while interact
        # acceptable smoothing parameter sp has been found
        # smooth again using sp, retrieve the result and store it in OPtop
        smo = self.ops[fop]._spline( sp, equi )
        n = smo.n
        if top != fop:
            self.ops[top] = copy.deepcopy(self.ops[fop])
        self.ops[top].x[:n] = smo.x[:n]
        if derive:
            self.ops[top].y[:n] = smo.dy[:n]
        else:
            self.ops[top].y[:n] = smo.y[:n]
        self.ops[top].n = n
    #END do_smplain

    def do_average(self, arg):
        '''Average front or tail of a curve to a constant value.
    SYNTAX: average,<opi>[1],<pa>[0]
                <opi>: index of operand array to average
                <pa>:  points to average.
                       pa > 0: average the FIRST pa points
                       pa = 0: do nothing
                       pa < 0: average the LAST -pa points
            If abs(pa) is greater than the number of points in the
            curve, all points in the curve are averaged.'''
        arglist = self._longlist(arg,2) # Two arguments
        opis = self._setdefault(arglist[0],'1')
        opi, err  = self._toopi( opis )
        if err != '':
            self.perror( err, traceback_war=False )
            return True # Flush cmd queue and return to prompt
        if self.ops[opi].isempty():
            self.perror('OP%i is empty'%opi, traceback_war=False)
            return True # Flush cmd queue and return to prompt
        pas = self._setdefault(arglist[1],'0')
        pa, err = self._toint( pas )
        if err != '':
            self.perror( err, traceback_war=False )
            return True # Flush cmd queue and return to prompt
        self.ops[opi].average( pa ) # average!
        # Report what has been done
        if abs(pa) >= self.ops[opi].n:
            rs = 'All'
            pa = self.ops[opi].n
        else:
            if pa > 0:
                rs = 'The first'
            else:
                rs = 'The last'
                pa = -pa
        self.poutput('%s %i points in OP%i have been replaced by their average'%(rs,pa,opi))
    #END do_average()

    def do_comb(self, arg):
        '''Make a curve a "comb-rastered" function in the mathematical sense.
    The curve must (in principle) start in the origin. For compatibility with
    my old TOPAS program this command can be called by the alias "stretch".
    Afterwards the curve starts at x=0, it ends at the former end value and
    its x-values are equidistantly spaced. This means that between two
    consecutive x-values there is a constant "comb-tooth width" of delx.
    Because "combing" a function is essential to the discrete, fast Fourier
    transformation (DFFT), it is suggested to make the comb have m = 2**n
    points with m from  [32, 64, 128, 256, 512, 1024, 2048].
    Nevertheless, the command does its work also with any other m in the
    range [32 .. 2048].
    SYNTAX: comb,<teeth>[256],<opi>[1]
        <teeth>: The number of data points after combing
        <opi>:   Index of the operand array on which the comb is executed

    The default settings make a function which is good for spatial frequency
    filtering (&filtint). In this way an interference function G(s) can be
    made if
        OP1 = ( I(s) - backgr ) * s**spower
    and this curve is truncated at the end of the Porod's law region.
        spower is the power of Porod's law and
        backgr is an estimated fluctuation background. The parameters are
    found by graphical inspection (&graphic) and the curve is prepared by
    &sby and &multiply

    METHOD: A point with the coordinates (0.0,0.0) is inserted and then the
    interval from smin=0 to smax is divided into <teeth>-1 intervals. The
    y-values are computed by interpolation.'''

        arglist = self._longlist(arg,2) # Two arguments
        ## (2)
        opis = self._setdefault(arglist[1],'1')
        opi, err  = self._toopi( opis )
        if err != '':
            self.perror( err, traceback_war=False )
            return True # Flush cmd queue and return to prompt

        thecurve = self.ops[opi]
        if thecurve.isempty():
            self.perror('OP%i is empty'%opi, traceback_war=False)
            return True # Flush cmd queue and return to prompt
        if thecurve.x[0] < 0.0:
            self.perror('Curve has negative x-values', traceback_war=False)
            return True

        ## (1)
        teethstr = self._setdefault(arglist[0],u'256')
        teeth, err = self._toint(teethstr)
        if err != '':
            self.perror( err, traceback_war=False )
            return True # Flush cmd queue and return to prompt
        if teeth < 32:
            self.poutput('teeth increased to 32')
            teeth = 32
        if teeth > 2048:
            self.poutput('teeth decreased to 2048')
            teeth = 2048
        thecurve.comb( teeth )
        self.poutput('OP1: Curve combed to %i points'%teeth)
        pass
    #END do_comb

    # Define an alias name / a synonym
    do_stretch = do_comb

    def do_unfold(self, arg):
        '''Linear desmearing of a curve from a given profile (curve) (to do)'''
        #TODO
        self.poutput('Not yet implemented')
        pass
    #END do_unfold

    def do_xts(self, arg):
        '''Rescaling the curve from x [mm] to scattering vector s [1/nm]
        SYNTAX: xts,<dist>,<lambda>[0.15418],<opi>[1]
            typical call: xts,2453.0,0.1
            <dist>   : the distance [mm] between sample and detector
            <lambda> : the X-ray wavelength [nm]. Defaults to
                       the copper wavelength. (Synchrotrons often have 0.15
            <opi>    : The index of the operand array
        Here the trigonometric functions sin and tan are approximated by radians.
        The error is < 1% in small angle scattering'''
        arglist = self._longlist(arg,3) # Three arguments
        dists = self._setdefault(arglist[0], u'') # Default: value missing
        dist, err = self._tonumber(dists)
        if err != '':
            self.perror( 'Sample-detector distance "%s" is no number'%dists, traceback_war=False )
            return True # Flush cmd queue and return to prompt
        lams = self._setdefault(arglist[1], u'0.15418')
        lamb, err = self._tonumber(lams)
        if err != '':
            self.perror( err, traceback_war=False )
            return True # Flush cmd queue and return to prompt
        opis = self._setdefault(arglist[2], u'1')
        opi, err  = self._toopi( opis )
        if err != '':
            self.perror( err, traceback_war=False )
            return True # Flush cmd queue and return to prompt
        if ops[opi].isempty():
            self.perror('OP%i is empty'%opi, traceback_war=False)
            return True # Flush cmd queue and return to prompt
        thecurve = ops[opi]
        span = thecurve.n
        # Approximation of trigonometric functions (tan, sin) by radians
        # yields negligible error in small angle scattering.
        mulcon = 1.0/(1000.0*lamb*dist)
        self.output('OP1: Transforming x-axis from distance on detector [mm] to s [1/nm]')
        thecurve.x[:span] = mulcon*thecurve.x[:span]
        pass
    #END do_xts

    def do_tts(self, arg):
        '''Rescaling the curve from 2theta [°] to scattering vector s [1/nm]
        SYNTAX: tts,<lambda>[0.15418],<opi>[1]
            typical call: tts,0.15
            <lambda> : the X-ray wavelength in units of nm. Defaults to
                       the copper wavelength. (Synchrotrons often have 0.15
            <opi>    : The index of the operand array'''
        arglist = self._longlist(arg,2) # Two arguments
        lams = self._setdefault(arglist[0], u'0.15418')
        value, err = self._tonumber(lams)
        if err != '':
            self.perror( err, traceback_war=False )
            return True # Flush cmd queue and return to prompt
        opis = self._setdefault(arglist[1], u'1')
        opi, err  = self._toopi( opis )
        if err != '':
            self.perror( err, traceback_war=False )
            return True # Flush cmd queue and return to prompt
        if ops[opi].isempty():
            self.perror('OP%i is empty'%opi, traceback_war=False)
            return True # Flush cmd queue and return to prompt
        thecurve = ops[opi]
        span = thecurve.n
        lam2 = 2.0/value # s = (2/lambda) * sin(theta)
        mulcon = 3.1415926/360.0 # 2theta [°] to theta [radians]
        self.output('OP1: Transforming x-axis from 2theta [°] to s [1/nm]')
        thecurve.x[:span] = lam2 * np.sin(mulcon*thecurve.x[:span])
    #END do_tts

    def do_cts(self, arg):
        '''Channel-to-scattering vector -- rescale the x-axis of the input data in OP1
        according to a polynomial, the coefficients of which are the arguments
        of this command. *** See example for s-to-q conversion ***
        This is a general-purpose method to transform any kind of input grid
        (i.e. channels, 2-theta, centimeters) on any kind of (inclined, curved)
        SAXS detector to units of the scattering vector. For the transformation
        from 2theta [°] to s [nm] and a wavelength of 0.15 nm we have, e.g.,
        in the tangent-plane approximation (valid for SAXS):
            cts,0.11636,0
            The example indicates already that the highest order coefficient
            of the polynomial comes first. For the example this means that the
            scaling equation is
                x[i] = 0.11636*x[i] + 0.0
            from q to s: cts,0.1591549,0
            from s to q: cts,6.2831853,0
            A more general example:
            cts,1.5,2.3,3.7 this would mean
                x[i] = 1.5 * x[i]**2  +  2.3 * x[i]  +  3.7
        This convention reflects the way how the python module numpy handles
        polynomials and their mathematics'''
        #NOTICE Well, the next line is unnecessary, as preparse has handled this
        #       outcome of users who write "cts, 1, 2" instead of "cts 1 2" or "cts,1,2"
        #       but to be sure:
        argc= ' '.join(arg.split()) # Remove multiple spaces from the argument string
        arli = argc.split(' ') # split the string into the substrings separated by blanks
        argc = ','.join(arli) # join the substrings by commata
        argc ='['+argc+']' # Make it conform to list syntax
        try:
            argc = eval(argc)
        except:
            self.perror('This was not a list of plain numbers', traceback_war=False)
            return True
        p = np.poly1d(argc) # evaluate to a numpy polynomial for polynom maths
        thecurve = self.ops[1] # Works on OP1
        if thecurve.isempty():
            self.perror('OP1 is empty', traceback_war=False)
            return True # flush cmd queue and return to user prompt
        span = thecurve.n
        self.poutput('Channel-to-s (cts): The new x-values in OP1 become x_new = ')
        print(p) #WARNING This only works with print, not with self.poutput
        self.ops[1].x[:span] = p(thecurve.x[:span]) # Apply the polynomial to all the valid points
        pass
    #END do_cts


    def do_frontex(self, arg):
        '''Extrapolate a curve I(s) towards s=0 (cf. CurveClass.frontex() )."
    SYNTAX: frontex,<pdeg>[-2],opi[1],<points>[5],<spower>[0]
        <pdeg> [-5..5] Degree and kind of the fit. Prefer negative values!
           <pdeg> = 0 : connect the origin with the first point of the curve
                        by a straight line f(s) = m*s . Specifically:
                          m = cu.y[0]/cu.y[0] with cu = self.ops[opi]
           <pdeg> < 0:  fit using a PARABOLA of degree abs(pdeg). The property
                        "even function" is ideally preserved.
           <pdeg> > 0:  fit using a full POLYNOMIAL of degree pdeg. This kind
                        of extrapolation is smoother at the connection point.
           <pdeg> = "p"<ap> fit by a POWER LAW into (x,y) = (0.0, -<ap>)
                        e.g.: p0.037 fits down (0.0,-0.037). In scattering,
                        <ap> frequently is the Porod asymptote.
                        Power law: y = a*x**p - ap. For a "p"-extrapolation
                        <spower> is not considered!
           If <spower> is > 0 and <pdeg> is a number, The fit is on
                               I_(s) = I(s) * s**spower
                and the extrapolated data are re-mapped on I(s). Care is
                taken for the computation of an approximate value at s=0.

        <opi>    Index of the operand array which shall be extrapolated
        <points> The number of data points, which are used for the
                 determination of the extrapolation curve.
        <spower> [0..4] power of "Porod's law"'''
        arglist = self._longlist(arg,4) # Four arguments
        opis = self._setdefault(arglist[1], u'1')
        opi, err  = self._toopi( opis )
        if err != '':
            self.perror( err, traceback_war=False )
            return True # Flush cmd queue and return to prompt
        ptss = self._setdefault(arglist[2],u'5')
        points, err = self._toint( ptss )
        if err != '':
            self.perror( err, traceback_war=False )
            return True # Flush cmd queue and return to prompt
        pdegs = self._setdefault(arglist[0], u'-2')
        if pdegs[0] == 'p':
            if len(pdegs) == 1:
                ap = 0.0
            else:
                ap, err = self._tonumber(pdegs[1:])
                if err != '':
                    self.perror( err, traceback_war=False )
                    return True # Flush cmd queue and return to prompt
            span = self.ops[opi].n
            self.ops[opi].front_powex( points, ap )
            expts = self.ops[opi].n - span
            self.poutput('Finished powex. %i points extrapolated towards s=0'%expts)
            return # from power law extrapolation

        pdeg, err = self._toint(pdegs)
        if err != '':
            self.perror( err, traceback_war=False )
            return True # Flush cmd queue and return to prompt
        parabolic = pdeg < 0
        pdeg = abs(pdeg)   # Handle silly input
        pdeg = min(pdeg,5)
        if points < pdeg+1:
            self.poutput('<points> increased to <pdeg>+1')
            points = pdeg+1

        spowers = self._setdefault(arglist[3],u'0')
        spower, err = self._toint( spowers )
        if err != '':
            self.perror( err, traceback_war=False )
            return True # Flush cmd queue and return to prompt
        spower = max( 0, spower )
        spower = min( spower, 4 )
        thecurve = self.ops[opi]
        span = thecurve.n
        if span < points:
            self.perror( 'OP%i has less data points than <points>=%i'%(opi,points), traceback_war=False )
            return True
        if pdeg == 0:
            self.poutput( 'OP%i: Connecting the origin to the first point by a straight line.'%opi)
        else:
            if parabolic:
                self.poutput( 'OP%i: Front extrapolation using a %i-degree PARABOLA computed from'%(opi,pdeg) )
            else:
                self.poutput( 'OP%i: Front extrapolation using a %i-degree POLYNOMIAL computed from'%(opi,pdeg) )
            self.poutput( 'the first %i points using the (intermediate) curve f(s) * s^%i.'%(points,spower) )
        if thecurve.x[0] <= 0.0:
            self.poutput('NOTHING TO DO: Curve starts already at s=0.')
            return False
        self.ops[opi].frontex( pdeg, points, spower, parabolic )
        newspan = self.ops[opi].n
        expts = newspan - span
        self.poutput('Finished. %i points have been extrapolated towards s=0'%expts)
    #END do_frontex

    def do_ext15(self, arg):
        '''Extrapolate a curve f(x) at its end to 1.5*x_end. This extrapolation
    resembles the extrapolation "into the edges of a quadratic SAS pattern"
    SYNTAX: ext15,<pdeg>[2],opi[1]
        <pdeg> [0..5] Degree of the polynomial to fit to the last 5% of the
               curve. special treatment for <pdeg>=0: From the last point a
               line is generated which touches the x-axis at its end.
        <opi>  Index of the operand array which shall be extrapolated'''
        arglist = self._longlist(arg,2) # Two arguments
        pdegs = self._setdefault(arglist[0], u'2')
        pdeg, err = self._toint(pdegs)
        if err != '':
            self.perror( err, traceback_war=False )
            return True # Flush cmd queue and return to prompt
        pdeg = abs(pdeg)   # Handle silly input
        pdeg = min(pdeg,5) # Restrict to a polynomial of degree 5
        opis = self._setdefault(arglist[1], u'1')
        opi, err  = self._toopi( opis )
        if err != '':
            self.perror( err, traceback_war=False )
            return True # Flush cmd queue and return to prompt
        thecurve = self.ops[opi]
        span = thecurve.n
        if thecurve.n < 3:
            self.perror('OP%i is empty'%opi, traceback_war=False)
            return True # Flush cmd queue and return to prompt
        xend = thecurve.x[span-1]
        #TODO the 0.95 could become a parameter
        xbeg = 0.95*xend # favorable interval: [xbeg .. xend]
        if pdeg == 0:
            # The very simple line extrapolation works always
            xbeg = xend
            xend = xbeg*1.5
            xdel = thecurve.x[span-1] - thecurve.x[span-2]
            ybeg = thecurve.y[span-1]
            m = (0 - ybeg)/(xend - xbeg)
            xact = xbeg + xdel
            # Generate the line
            while xact <= xend:
                thecurve.x[span] = xact
                thecurve.y[span] = ybeg + m*(xact - xbeg)
                xact = xact + xdel
                span = span + 1
            many = span - thecurve.n
            thecurve.n = span
            self.poutput('OP%i: %i points extrapolated using a straight line to y=0'%(opi,many))
            return

        good = np.where(thecurve.x[:span] > xbeg) # complex list of all valid indices
        thex = thecurve.x[good] # The x-values in the interval
        pts  = len(thex)
        # where does not rearrange, so it is unnecessary to: they=thecurve.y[span-pts;span]
        if pts < 10:
            self.perror('Less than 10 points between x=%8.3e and %8.3e'%(xbeg,xend), traceback_war=False)
            return True
        thecurve.ext15( pdeg, pts )
        many = thecurve.n - span
        self.poutput('OP%i: %i points extrapolated using a %i-order polynomial'%(opi,many,pdeg))
    #END do_ext15


    def _funexp(self, t, ap, b, c):
        # return ap + b*np.exp( np.exp(c)*t ) # with c=cl
        return ap + b*np.exp(-c*t)
    #END _funexp()
    
    def do_extexp(self, arg):
        '''Extrapolate a curve f(x) by assuming an exponential decay of the tail.
    This extrapolation can be useful, if a short detector has only captured
    the start of Porod's law and a prolongation of s**2 I_1(s) is required
    for the interference function. The model function is
            y(t) = Ap + b * exp( -c*t )      y(0) = b + Ap
    
    SYNTAX: extexp,<s_s>,<s_e>[1.5*s_c],opi[1]
        <s_s> start s. Points above s_s are feeding the nonlinear regression
        <s_e> end s. The fitted function will extend up to s_e.
        <opi>  Index of the operand array which shall be extrapolated
    (s_c: cutoff s of the input curve)
    
    The command sets the attribute(s):
        Ap
    
    Method: The command calls curve_fit from scipy.optimize and inserts
    fitted values for  <s_fs> ... <s_e>  with s_fs = s_s + (s_c - s_s)/2.
    Here s_c is the cutoff of the input curve. Above s_c new points are
    generated by continuing the spacing of the last two points.
    
    Hint for the programmer: The command does not estimate starting values
    for the parameters Ap, b and c, because it converged well up to now.
    In case of problems one may choose a different decay parameter, namely
    cl = log(c) which could make the regression algorithm more robust.'''
        arglist = self._longlist(arg,3) # Three arguments

        ## (3)
        opis = self._setdefault(arglist[2], u'1')
        opi, err  = self._toopi( opis )
        if err != '':
            self.perror( err, traceback_war=False )
            return True # Flush cmd queue and return to prompt
        thecurve = self.ops[opi]
        span = thecurve.n
        if thecurve.n < 3:
            self.perror('OP%i is empty'%opi, traceback_war=False)
            return True # Flush cmd queue and return to prompt
        sc = thecurve.x[span-1] # highest "s" in the curve
        
        ## (1)
        ss, err = self._tonumber(arglist[0])
        if err != '':
            self.perror( err, traceback_war=False )
            return True # Flush cmd queue and return to prompt
        if ss >= sc:
            self.perror( 'Curve shorter than extrapolation start s_s', traceback_war=False )
            return True # Flush cmd queue and return to prompt
        
        ## (2)
        defval = 1.5*sc
        defstr = "%s"%defval
        ses = self._setdefault(arglist[1], defstr)
        se, err = self._tonumber(ses)
        if err != '':
            self.perror( err, traceback_war=False )
            return True # Flush cmd queue and return to prompt
        if se <= sc:
            self.perror( 'No work to do', traceback_war=False )
            return True # Flush cmd queue and return to prompt
                
        # Start of extrapolation
        xv = thecurve.x[:span] # Consider only the curve points
        yv = thecurve.y[:span]
        # Feed arrays
        feedx = xv[np.where(xv >= ss)]
        feedy = yv[np.where(xv >= ss)]
        if len( feedx ) < 10:
            self.perror( 'Insufficient feed for extrapolation', traceback_war=False )
            return True # Flush cmd queue and return to prompt
        with warnings.catch_warnings():
            warnings.simplefilter("error", OptimizeWarning)
            try:
                popt, pcov = curve_fit( self._funexp, feedx, feedy )
            except (OptimizeWarning, Exception) as exi:
                # Catch warnings and all exceptions
                print( exi.message )
                self.perror('curve_fit unsuccessful', traceback_war=False)
                return True # Flush cmd queue and return to prompt
        # Significances
        perr = np.sqrt(np.diag(pcov))
        ap, b, c = popt
        pnames = ('Ap', 'b', 'c' )
        for i in range(len(pnames)):
            print("%3s = %9.2e ± %9.2e"%(pnames[i],popt[i],perr[i]))
            
        # Overwrite the last half of the tail
        so = se + (sc - se)/2.0
        theslots = np.where(xv > so)
        thecurve.y[theslots] = self._funexp(thecurve.x[theslots], ap, b, c )
        # Generate extra points at the end
        dels = xv[-1] - xv[-2]
        start = sc + dels
        stop = se + dels
        extrax = np.arange(start,stop,dels)
        if len(extrax) + span > self._MAXP:
            self.perror( 'Too many extrapolation points requested', traceback_war=False )
            return True # Flush cmd queue and return to prompt
        extray = self._funexp(extrax, ap, b, c )    
        # Insert the extrapolated data
        ne = len(extrax)
        thecurve.x[span:span+ne] = extrax
        thecurve.y[span:span+ne] = extray
        thecurve.n = span + ne
        # Set attribute Ap
        thecurve.puta('Ap',ap)
        self.poutput('OP%i: %i points extrapolated assuming exponential decay'%(opi,ne))
    #END do_extexp()

    def do_morph(self, arg):
        '''Apply a general morphology operator to a curve.
        SYNTAX:   morph,<box>[7],<opi>[1],<task>[erode]
              <box>:  an odd integer which specifies the width of a sliding
                      box
              <opi>:  The index of the curve which shall be median-filtered
              <task>: 'erode', 'median', 'dilate' (short: first letter)
                      erode:  take the data point with the lowest y-value in
                              the box
                      median: take the middle y-value from the box list
                      dilate: take the highest y-value from the box list
        Example: morph,13,2,dilate
        The command calls the curve specific method morph directly. There is
        a separate command &median as a shortcut to the median operator because
        of the importance of the latter.

        The complex morphological operators "opening" and "closing" must be
        mimicked by two successive calls to the morph command:
        Opening : &morph,<mybox>,<op>,erode&morph,<mybox>,<op>,dilate
        Closing : &morph,<mybox>,<op>,dilate&morph,<mybox>,<op>,erode'''
        arglist = self._longlist(arg,3) # Three arguments
        ## (1)
        boxs = self._setdefault(arglist[0], u'7')
        box, err = self._toint(boxs)
        if err != '':
            self.perror( err, traceback_war=False )
            return True # Flush cmd queue and return to prompt
        box = abs(box)   # Handle silly input
        box = max(box,3)
        ## (2)
        opis = self._setdefault(arglist[1], u'1')
        opi, err  = self._toopi( opis )
        if err != '':
            self.perror( err, traceback_war=False )
            return True # Flush cmd queue and return to prompt
        if self.ops[opi].isempty():
            self.perror('OP%i is empty'%opi, traceback_war=False)
            return True # Flush cmd queue and return to prompt
        ## (3)
        task = self._setdefault(arglist[2], u'erode')
        if task[0] == 'd':
            task = 'dilate'
        elif task[0] == 'e':
            task = 'erode'
        elif task[0] == 'm':
            task = 'median'
        else:
            self.perror('Unknown morphing task "%s"'%task, traceback_war=False)

        self.poutput('OP%i: "%s" morphing over %i data points'%(opi,task,box))
        self.ops[opi].morph(box,task)
        pass
    #END do_morph

    def do_median(self, arg):
        '''Smooth noisy curves by median filtering (removal of "salt-and-pepper noise").
        SYNTAX:   median,<box>[7],<opi>[1]
              <box>: an odd integer which specifies the width of a sliding box
              <opi>: The index of the curve which shall be median-filtered
        Example: median,9,2
        Hint: &median is a shortcut to &morph with the task "median".'''
        arglist = self._longlist(arg,2) # Two arguments
        boxs = self._setdefault(arglist[0], u'7')
        box, err = self._toint(boxs)
        if err != '':
            self.perror( err, traceback_war=False )
            return True # Flush cmd queue and return to prompt
        box = abs(box)   # Handle silly input
        box = max(box,3)
        opis = self._setdefault(arglist[1], u'1')
        opi, err  = self._toopi( opis )
        if err != '':
            self.perror( err, traceback_war=False )
            return True # Flush cmd queue and return to prompt
        if self.ops[opi].isempty():
            self.perror('OP%i is empty'%opi, traceback_war=False)
            return True # Flush cmd queue and return to prompt
        self.poutput('OP%i: Smoothing by median filtering over %i data points'%(opi,box))
        self.ops[opi].morph(box,'median')
        pass
    #END do_median

    def do_graphic(self, arg):
        '''Display and manipulate the curves in the operands OP0 to OP2.
   Presented are only the operands which BOTH contain data and for which the
   corresponding bit in self.visicurves is set to 1. self.visicurves is a
   settable variable (e.g. "set visicurves 6" to suppress presentation of OP0)
   SYNTAX: graphic, <tramod>, <backgr>
       <tramod> [0]: 0 .. 4 modus of transformation just for the display
       <backgr> [0]: a number which is a background subtracted only for display
       The transformations which are carried out are (in this sequence):
       (x,y)  --> (x, y-backgr) = (x,y')
       (x,y') --> (x, y'*x^tramod)
   The functionality of this command is realized in the module curvetray,
   which works hand-in-hand with curveclass, the implementation of curves
   and primitive operations on them.

   The interactive part of graphic:
   The built-in interactive commands and the radio buttons of matplotlib/pyplot
   work. If a radio button is active, specific commands cannot be performed and
   a warning (both text and acoustic) is issued.
   Graphical input (GI) is performed by entering vectors. The user can escape
   from an unintended graphical input by pressing and releasing the right mouse
   button.
   Vector: PRESS (and hold) left mouse button, then
           MOVE the mouse to the release position and
              ... while in "move"-mode, the starting point of the line is
                  adjusted while the Ctrl-Key is pressed (move the whole line)
           RELEASE the left mouse button.
    Keys:   | action:
   ---------+--------------------------------------------------------------
    0..4    | change <tramod> by
    b       | GI of the new background (y_RELEASE)
   ---------+--------------------------------------------------------------
    ctrl-0  |
      ...   | select the curve OP[0] ... OP[2]. The curve markers turn red.
    ctrl-2  |
   ---------+--------------------------------------------------------------
    crtrl-3 | deselect a selected curve

    Manipulation operations on selected curves (lower and upper case)
    a : sc(a)le the curve by y_RELEASE/y_PRESS (make y_PRESS --> y_RELEASE)
    x : same in x-direction: compress/stretch curve
    n : li(n)k the tail of OP2 to the front of OP1. x_RELEASE
        becomes the border
    d : (d)elete one or more data points. All points between x_PRESS and
        x_RELEASE are deleted.
    z : (z)ero the points in the selected interval. Now the points are not
        deleted, but their y-values are set to zero.
    i : (i)nsert one point at the position (x_RELEASE, y_RELEASE)
    m : (m)ove the point which is closest to (x_PRESS,y_PRESS) up or down
        to the new height y_RELEASE.'''
        arglist = self._longlist(arg,2) # Two arguments
        tramod = self._setdefault(arglist[0], u'0')
        if tramod not in ('0','1','2','3','4'):
            self.perror('tramod "%s" must be 0, 1, 2, 3 or 4'%tramod, traceback_war=False)
            return True # Flush and call postloop
        tramod = int(tramod)
        backgr = self._setdefault(arglist[1], u'0')
        backgr, Err = self._tonumber( backgr )
        if Err:
            self.perror(Err, traceback_war=False)
            return True

        PresentSomething = False
        markers = ['|','+','x']
        for opi in range(3): # Is at least one of the OPs not empty?
            if (self.ops[opi].n > 0):
                # Shift test-bit 1 into visicurves and check,
                # if the corresponding bit in visicurves is set
                if (self.visicurves & (1 <<opi)) != 0:
                    # Curve shall be presented
                    PresentSomething = True
                else:
                    # suppress painting
                    markers[opi] = ''
        # print( markers ) # ... to check visibility
        if PresentSomething:
            self.ThePlot.ScatKnife(self.ops, markers,self.fontfactor,backgr,tramod)
        pass
    #End do_graphic()

# Define synonyms to the coded commands
    '''
    do_say = do_speak  # now "say" is a synonym for "speak"
    do_orate = do_speak  # another synonym, but this one takes multi-line input
    '''

#+++++++++++++++User configurability+++++++++++++++++++++++++++++++++++++++++++++
    def readprofile( self ):
        '''Configure settable parameters from a profile file'''
        # Generate name of the program's rc file
        extpos = sys.argv[0].rfind('.')
        rcfile = sys.argv[0][0:extpos]+'.rc'
        short = os.path.basename(rcfile) # Without its path
        if os.path.isfile( short ):
            rcfile = short # In current directory
        if os.path.isfile(rcfile):
            print('Retrieving individual configuration from %s'%rcfile)
            f = open(rcfile, 'r')
            for line in f:
                line = line.strip()
                if len(line) > 1: # Skip empty lines
                    if line[0] != '#': # Skip comment
                        order = line
                        commstart = order.find('#')
                        if commstart != -1: # remove trailing comment
                            order = order[0:commstart]
                        param, val = order.split('=')
                        param = param.strip()
                        val = val.strip()
                        try:
                            current_val = getattr(self, param)
                            fine = True
                        except:
                            print('%s: Parameter "%s" unknown. Cannot set it'%(short,param))
                            fine = False
                        if fine:
                            if (val[0] == val[-1]) and val[0] in ("'", '"'):
                                val = val[1:-1]
                            else:
                                val = cast(current_val, val)
                                pass
                            # new val has the correct type: set it!
                            setattr(self, param, val)
                            pass
                        pass
        else:
            #print('No configuration file %s'%rcfile)
            pass
        pass
    #END readprofile()

def cast(current, new):
    """Tries to force a new value into the same type as the current when trying to set
    the value for a parameter.

    :param current: current value for the parameter, type varies
    :param new: str - new value
    :return: new value with same type as current, or the current value if there was
    an error casting
    """
    typ = type(current)
    if typ == bool:
        try:
            return bool(int(new))
        except (ValueError, TypeError):
            pass
        try:
            new = new.lower()
        except AttributeError:
            pass
        if (new == 'on') or (new[0] in ('y', 't')):
            return True
        if (new == 'off') or (new[0] in ('n', 'f')):
            return False
    else:
        try:
            return typ(new)
        except (ValueError, TypeError):
            pass
    print("Problem setting parameter (now %s) to %s; incorrect type?" % (current, new))
    return current

#----------------------------User configurability--------------------------------

#*** Main program
if __name__ == '__main__':
    SaskiaProg = SasCmds()
    SaskiaProg.readprofile() # Configure from individual profile
    SaskiaProg.cmdloop()
