#!/usr/bin/env python
# coding=utf-8
import os, sys
import glob # For filename search with wildcards
import copy # for independent operands

'''Extract attributes from the frames taken during an experiment
and store them in a curve.
SYNTAX: @@attrext, <framefiles>, <outfile>, <keyy>, <keyx>
    <framefiles>: Selector defining the file series
    <outfile>:    The file, into which the resulting curve is written
    <keyy>:       The attribute, whose variation shall be studied
    <keyx>: (optional): x-axis is not the frame number itself, but
                  the specified attribute (e.g. "elapsed" or "T"
                  for elapsed time or the temperature)
    EXAMPLEs:
             @@attrext,sy02_up*,ay02_up_bat,bat  # as function of frame-no.
             @@attrext,sy02_up*,aty02_up_bat,bat,T  # as function of Temp
    extracts all "bat" attribute values of the small-angle experiment y02_up
    in the attribute file ay02_up_bat.dat and takes simply the frame number
    as the abscissa. -- Noisy readings of the PIN-Diode become visible!
    Postprocessing:
        saskia> ops,l # Move PIN-diode curve to OP1
        saskia> average,,100 # average the first 100 frames (const absorption)
        saskia> smp,,,,.9 # smooth the rest (here: automatically with sp=0.9)
        saskia> @@attrinj,sy02_up*,bats # inject the smoothed PIN-data
    Then we are prepared for absorption compensation and subtraction of the
    machine background from all curves of the frame set'''

## Script procedures

def banner():
    # print the name of the running script
    myname = os.path.basename(sys.argv[0])
    self.poutput("Finished: {!r}".format(myname))
    pass
#END banner()

def extractor():
    # *** Check the number of arguments with which the script has been called
    argus = len(sys.argv) - 1
    if argus < 3:
        return "script needs 3 or 4 arguments"
    #argument lister in the next 2 lines
    #for i, arg in enumerate(sys.argv[1:]):
    #    self.poutput("arg {}: {!r}".format(i+1, arg))

    # *** Evaluate the arguments
    # Argument 1: the "all"-files
    alldef = sys.argv[1]

    # If the user has given no extension, then append '.dat'
    test = alldef.split('.')
    if len(test) < 2:
        alldef = alldef+'.dat'

    # Argument 2: The output file
    outfile = sys.argv[2]
    test = alldef.split('.')
    if len(test) < 2:
        outfile = outfile+'.dat'

    # Argument 3: The key whose values shall be extracted
    dictkey = sys.argv[3]

    # Argument 4: Special x-values or ordinary frame number?
    xkey = 'FrameIndex' # Default string
    specialx = argus > 3
    if specialx:
        xkey = sys.argv[4]
    # alphabetically sorted list of the "all"-files
    allfiles = sorted(glob.glob(alldef))
    if len(allfiles) < 1:
        return 'No files matching "%s"'%alldef

    # Make an independent curve operand for the collection
    collcur = copy.deepcopy( self.ops[0] )
    collcur.clean()
    collcur.env = {'axis2' : dictkey, 'axis1' : 'frame', 'units1' : ''}

    actframe = copy.deepcopy( self.ops[0] )

    # Test dictionary env: Are the specified attributes present?
    err = actframe.readascii( allfiles[0] )
    if err:
        return err
    if not (dictkey in actframe.env):
        return 'Attribute "%s" missing in first frame'%dictkey
    if specialx:
        specialx = xkey in actframe.env
        if not specialx:
            self.poutput('Attribute "%s" missing in first frame. Back to frame-index.'%xkey)
    if specialx:
        # valid special abscissa found -- so overwrite "frame"
        collcur.env['axis1'] = xkey

    for i, thefile in enumerate(allfiles):
        err = actframe.readascii( thefile )
        if err:
            return err
        yval = actframe.attribute(dictkey)
        if specialx:
            xval = actframe.attribute(xkey)
        else:
            # May be that the user has erased some of the frames. Then "i" is
            # not the frame index. So let's extract the frame index from the
            # file name!
            body  = thefile.split('.')
            words = body[0].split('_')
            frame = words[-1]
            xval = int(frame)

        self.poutput('%4i. x = %8g, y = %8g'%(i,xval,yval)) # log extracted values
        # Test: x-values must be strictly ascending
        if i > 0:
            ascending = collcur.x[i-1] < xval
            if not ascending:
                topheader = '# From sequence "%s" attributes "%s" vs. "%s"'%(alldef,dictkey,xkey)
                collcur.headers.append(topheader)
                return 'STOP: This x is not bigger than the last one!'
        collcur.x[i] = float(xval)
        collcur.y[i] = yval
        collcur.n = i+1 # number of entered points
        pass # loop end

    # Bookkeeping
    span = collcur.n

    topheader = '# From sequence "%s" attributes "%s" vs. "%s"'%(alldef,dictkey,xkey)
    collcur.headers.append(topheader)
    collcur.writeascii( outfile, True ) # permit overwriting
    self.poutput('Attribute curve written to "%s"'%outfile)
    pass
#END extractor

# The main program of this script
error = extractor()
if error:
    self.perror(error)
    # And exit without signaling the error status
banner()

