#!/usr/bin/env python
# coding=utf-8
import os, sys
import glob # For filename search with wildcards
import numpy as np

'''backcorr.py: Background correction of scattering curves which belong to
           the same frameset. Before the script can be started, the machine
    background must have been loaded into OP0. The results will be stored
    in new curve files with the old file name prepended by a "b" for
    background corrected data. The raw-curve intensity is scaled up in order
    to compensate for the intensity loss by absorption. Thickness correction
    and transformation to absolute units is not done here. The script
    uses and destroys OP1
EXAMPLE: saskia> read,bgsaxs_smo,0
         saskia> @@backcorr,sy02_up*,0.9
explanation:           sy02_up*.dat  -->  bsy02_up*.dat with decreasing the
                       weight of the background absorption factor by
                       multiplication with 0.9. The beam attenuation is
                       fetched from the key "bats" in the attribute dictionary
                       env of the actual curve or, if "bats" is not present,
                       from "bat".
         saskia> @@backcorr,sy02_up*,,bad
                       Here tunefact=1 and the user asks to use the beam
                       absorption values which he has stored under the
                       attribute "bad".
         For wide-angle X-ray scattering data:
         saskia> read,bgwaxs_smo,0
         saskia> @@backcorr,wy02_up_*,0.9,,20
explanation:
                       The waxs background without the metal peaks is loaded
                       into OP0. Then this script (backcorr.py) is called
                       for the frame set wy02_up_*.dat with a bg-tuning factor
                       of 0.9, standard attribute key for beam attenuation and
                       a smoothing parameter of 20 which should remove the
                       metal peaks without deforming the broader WAXS peaks.
SYNTAX: backcorr,<frameset>[''],<tunefact>[1.0],<attkey>[''],<smopar>[''],<bgignore>['']
        <frameset>: a file designator like: sy02_up*
        <tunefact>: a factor by which the computed absorption factor emut
                    is multiplied (to correct error in the PIN-diode
                    reading of the machine background)
        <attkey>:   attribute key from which the reading of the beam
                    attenuation detector shall be fetched. "bats" is the
        standard key for smoothed bat-readings. The fallback key is "bat",
        which should have been set up by the merge-script from the ratio
        pin/beam with beam the primary beam intensity and pin the intensity
        measured in the PIN-diode during the exposure of the actual frame.
        With short-time exposure these attributes will exhibit considerable
        noise. Then a smoothed value should have been made available under
        the key "bats". The script first searches for the user-given key,
        then for "bats" and finally for "bat".
        <smopar>:   A smoothing parameter which removes the metal reflections
                    from the wide-angle X-ray scattering curves which are
                    processed here. It must be tested, if this kind of
                    metal-peak removal is able to keep the (wider) polymer
                    peaks, whereas the (sharp) metal peaks are removed.
                    See example above!
        <bgignore>: Do not subtract the background itself (good for WAXS
                    processing)'''

def banner():
    # print the name of the running script
    myname = os.path.basename(sys.argv[0])
    self.poutput("Finished: {!r}".format(myname))
    pass
#END banner()


def argexpand( specs ):
    '''Expand the argument list, if the user has given less than specs
    arguments'''
    result = sys.argv[1:]
    while len( result ) < specs:
        result.append( '' )
    return result
#END argexpand()

def bestbat( opi, usrattr ):
    '''searches in opi the attribute dictionary for the highest ranking
    key for Beam ATtenuation data.'''
    curdict = self.ops[opi].env
    if usrattr in curdict:
        return usrattr

    attr = 'bats' #smoothed beam attenuation data
    if attr in curdict:
        return attr
    else:
        attr = 'bat' #original beam attenuation data
        if attr in curdict:
            return attr
    return '' # Nothing found
#END bestbat()


def doit():
    '''doit does the work'''
    #argument lister
    #for i, arg in enumerate(sys.argv[1:]):
    #    print("arg {}: {!r}".format(i+1, arg))

    ### Check presence of machine background
    opbg = 0
    bgcurve = self.ops[opbg] # Short for background curve
    sacurve = self.ops[1]    # Short for sample curve
    if bgcurve.isempty():
        return 'No machine background provided in OP%s (needed for beam attenuation)'%opbg

    ### Argument parsing
    arglist = argexpand(4) # The script has 4 arguments
    ## (1) defines the "all"-files
    alldef = arglist[0]
    # If the user has given no extension, then append '.dat'
    test = alldef.split('.')
    if len(test) < 2:
        alldef = alldef+'.dat'
    # alphabetically sorted list of the specified files
    allfiles = sorted(glob.glob(alldef))
    if len(allfiles) < 1:
        return 'No files matching "%s"'%alldef

    ## (2) tunefactor
    tunes = arglist[1]
    tunefact = 1.0
    if tunes != '':
        try:
            tunefact = float(tunes)
        except:
            return 'tunefact: Unable to convert "%s" to a number'%tunes
    tunefact = abs(tunefact)
    if tunefact < 0.1:
        return 'tunefact too low'

    ## (3) attkey
    attkey = arglist[2]
    # end argument parsing
    bgkey = bestbat( opbg, attkey )

    ## (4) smopar
    smos = arglist[3]
    smooth_scat = smos != ''
    if smooth_scat:
        try:
            smopar = float(smos)
        except:
            return 'smopar: Unable to convert "%s" to a number'%smos

    ## (5) bgignore
    bgignore = arglist[4] != ''

    if bgkey == '':
        return 'OP%i: Cannot find a key for the attribute "beam attenuation"'%opbg
    batbg = self.ops[opbg].env[bgkey] # The beam attenuation of the machine background

    # determine bg-related "constants"
    spanb = bgcurve.n


    # process all files
    for i, thefile in enumerate(allfiles):
        # print( thefile )
        # read the curve file into OP1
        self.ops[1].readascii( thefile )
        if smooth_scat:
            smo = self.ops[1]._spline( smopar, 0 ) # Keep points!
            n = smo.n # The number of points
            self.ops[1].x[:n] = smo.x[:n]
            self.ops[1].y[:n] = smo.y[:n]
            self.ops[1].n = n # yes, I know they do not change
        sakey = bestbat( 1, attkey ) # Find the beam attenuation key of the sample
        if sakey == '':
            return '"%s": Cannot find beam-attenuation key "%s"'%(thefile, attkey)
        batsa = sacurve.env[sakey]
        # Compute the absorption factor emut from (noisy?) PIN-diode
        emutp = batsa / batbg
        # Modify by tunefactor
        emut = emutp*tunefact
        # Clip silly absorption (from blue frames?)
        emut = min( emut, 0.999 )
        emut = max( 0.01, emut )
        # Make emut an attribute of the curve and store it
        self.ops[1].env['emut'] = emut
        # In the sample scattering: compensate the intensity loss
        # due to absorption
        span = sacurve.n
        self.ops[1].y[:span] = self.ops[1].y[:span]/emut

        if not bgignore:
            # Now the background scattering can be subtracted. Test, if
            # fast subtraction is possible
            simple = spanb == span # Both curves have the same number of points
            if simple:
                # Are the deviations at each position sufficiently close?
                reldev = (sacurve.x[:span] - bgcurve.x[:span])/bgcurve.x[:span]
                reldev = abs(reldev) # Catching bg higher than sample
                # "simple", if the deviations of limits are smaller than 0.01%
                simple = max(reldev) < 1.0E-4 #WARNING This becomes hazardous,
                # ... if there are very small x-values -- but there is the beamstop.
                if simple:
                    # Subtract y-values directly one-by-one
                    self.ops[1].y[:span] = sacurve.y[:span] - bgcurve.y[:span]
                else:
                    self.poutput('*** Slow interpolation of background data')
                    # General curve subtraction with interpolation
                    self.ops[1].addcurve(self.ops[0],-1.0)
                    pass
                pass

        ## I compute µt from exp(-µt) and store it as the attribute "mut".
        ## This is convenient for normalization to the irradiated volume.
        ## For every frame i the sample thickness in length units can be
        ## computed, if only for one frame "r" the real thickness t_r is
        ## known. Thus measure the thickness before starting the experiment.
        ##                  t_i = t_r * ( µt_i / µt_r )
        ##
        mut = -np.log(emut)
        self.ops[1].env['mut'] = mut # Set the attribute

        # Make the name of the output file
        savfn = 'b'+thefile
        self.poutput('bg-corrected with e**(-µt)=%6.4f in "%s"'%(emut,savfn))
        err = self.ops[1].writeascii( savfn, True ) # Permit overwriting
        if err != '':
            return err
        pass
    #end loop all files
    self.poutput('All files processed. New attributes: "emut"=exp(-µt), "mut"=µt.')
    self.poutput('"mut" is good for later normalization to identical thickness or abs. int.')
    return ''
#END doit()

error = doit()
if error != '':
    self.perror( error, traceback_war=False)
    # No signalling to cmd
banner()



