#!/usr/bin/env python
# coding=utf-8
import os, sys
import glob # For filename search with wildcards
import copy # For true copies of variables

'''dwellchange.py: Change the dwell frames in a set of already merged frames

This script is related to the special experiment plan at ALBA. It is a tool to
change the number of "dwell frames" after reaching the top temperature, during
which the sample is kept at constant temperature before cooling. The reason is,
that there are runs with different top temperatures but the same material. Then
for presentation in a 2D pseudo-color plot it appears suitable to have the same
number of frames for all experimental runs. All experiments with lower top temperature
are treated: The real dwell frames (5) are removed and replaced more interpolated 
dwell frames, so that ultimately all graphs share the same number of frames
(in our case: 365). Then the plots can be stacked vertically, and each column is
related to the same temperature. For clarification one simply paints a white
box on top of the dwell frames in order not to puzzle the reader. Low-temperature
experiments then have a broad white central box, while high-temperature experiments
have a narrow one.

SYNTAX: @@dwellchange,<frameset>,<upend>,<remove>,<insert>
EXAMPLE: @@dwellchange,nmdbwy50,180,0,5 # Only insert 5 dwellframes after 180 frames
    <frameset>: An existent set of curves with names <frameset>_???.dat. The new set
                is named "dummy" for easy checking and overwriting.
    <upend>:  The number of frames before the insertion
    <remove>: The number of old dwell frames which must not be considered
    <insert>: The number of new dwell frames.
'''

## Script procedures

def banner():
    # print the name of the running script
    myname = os.path.basename(sys.argv[0])
    self.poutput("Finished: {!r}".format(myname))
    pass
#END banner()

def datfilesfound( test ):
    # True, if there are files matching the pattern
    missingext = len(test.split('.')) == 1
    if missingext:
        # No dot which indicates a file extension
        # Append default extension for data files
        testx = test+'.dat'
    else:
        testx = test
    listoffiles = glob.glob( testx )
    return len(listoffiles) > 0
#END datfilesfound()

def keyav( fracearly, fraclate, key, dest, early, late ):
    # Interpolation: 
    # Puts the average from early and late weighted by
    # fracearly and fraclate into the dictionary dest
    if key in early:
        if key in late:
            dest[key] = fracearly*early[key] + fraclate*late[key]
            return
    dest[key] = 'n/a' # Indicate failure to average

def changedwellframes():
    '''changedwellframes does the work'''
    #argument lister
    #for i, arg in enumerate(sys.argv[1:]):
    #    print("arg {}: {!r}".format(i+1, arg))

    # Argument parsing of the first argument is postponed, because it
    # may involve the deletion of old files which shall only be done
    # just before the work starts.

    ### Argument parsing
    argus = len(sys.argv) - 1
    if argus != 4:
        return "Script needs exactly 4 arguments"

    ## (1) The set to work
    origbody = sys.argv[1]
    # outbody = '_'+origbody
    outbody = 'dummy'

    origset = origbody+'_*.dat'
    if not datfilesfound( origset ):
        return 'origset: There are no files matching "%s"'%origset

    ## (2) Length of upgoing branch
    upends = sys.argv[2]
    try:
        upend = abs(int( upends ))
    except:
        return 'Error converting upend "%s" to integer'%upends


    ## (3) Number of old dwell frames
    removes = sys.argv[3]
    try:
        remove = abs(int( removes ))
    except:
        return 'Error converting remove "%s" to integer'%removes

    ## (4) Number of new dwell frames
    inserts = sys.argv[4]
    try:
        insert = abs(int( inserts ))
    except:
        return 'Error converting insert "%s" to integer'%inserts

    ### start the work
    outindex = 0 # The actual index into the output set
    # sorted list of the old set
    setfiles = sorted(glob.glob(origset))
    for i, thefile in enumerate(setfiles):
        err = self.ops[0].readascii( thefile )
        if err != '':
            return 'Error reading file "%s" in origset'%thefile
        # Make the output file name
        outis = str(outindex)
        while len(outis) < 4:
            # Fill with leading zeroes
            outis = '0'+outis
        outfile = outbody+'_'+outis

        if i == upend-1:
            # Last curve from up branch is just processed
            # Remember for continuity
            attrib = self.ops[0].env
            uptime = attrib['elapsed']
            cycle = attrib['exposure'] + attrib['latency']
            upframe = attrib['frame']
        # generate the next file in the merged set

        if (i < upend):
            self.poutput('Writing %s'%outfile)
            err = self.ops[0].writeascii( outfile, True ) # overwrite
            if err != '':
                return err
            outindex = outindex + 1
            # Remember the last curve as the "early" dwell frame
            be = copy.deepcopy( self.ops[0] )
            # in between do nothing - just over-read old dwell frames
        elif i == upend + remove:
            # We have reached the first frame after the dwell frames
            # So we must insert the new dwell frames and the reference
            # frame before it becomes easy copying, again
            ghost = copy.deepcopy(self.ops[0])
            span = ghost.n
            sufra = insert+1
            bl = self.ops[0] # late  real frame
            for jj in range(1,sufra):
                uptime = uptime + cycle
                ghost.env['elapsed'] = uptime
                upframe = upframe + 1
                ghost.env['frame']   = upframe
                # Sliding weight of average with fe and fl being the fractions
                # for weighting
                fe = (sufra -jj)/float(sufra)
                fl = jj/float(sufra)
                keyav( fe, fl, 'bats', ghost.env, be.env, bl.env )
                # ghost.env['bats'] = early*be.env['bats'] + late*bl.env['bats']
                keyav( fe, fl, 'bat',  ghost.env, be.env, bl.env )
                keyav( fe, fl, 'emut', ghost.env, be.env, bl.env )
                keyav( fe, fl, 'mut',  ghost.env, be.env, bl.env )
                keyav( fe, fl, 'T',    ghost.env, be.env, bl.env )
                ghost.y[:span] = fe*be.y[:span] + fl*bl.y[:span]
                # x-values remain the same

                # Make the output file name
                outis = str(outindex)
                while len(outis) < 4:
                    # Fill with leading zeroes
                    outis = '0'+outis
                outfile = outbody+'_'+outis
                # generate the next file in the merged set
                self.poutput('Writing %s'%outfile)
                err = ghost.writeascii( outfile, True )
                if err != '':
                    return err
                outindex = outindex + 1
            # Artificial dwell frames have been inserted
            # insert the reference at last
            outis = str(outindex)
            while len(outis) < 4:
                # Fill with leading zeroes
                outis = '0'+outis
            outfile = outbody+'_'+outis
            # Continuity of respective attributes -- i starts from 0
            self.ops[0].env['elapsed'] = uptime
            self.ops[0].env['frame']   = upframe

            # generate the next file in the merged set
            self.poutput('Writing %s %4i %6.1f %s'%(outfile,upframe,uptime,thefile))
            err = self.ops[0].writeascii( outfile, True )
            if err != '':
                return err
        elif (i > upend+remove):
            # Copy the down branch frames (except for the first frame)
            # Continuity of respective attributes -- i starts from 0
            uptime = uptime + cycle
            self.ops[0].env['elapsed'] = uptime
            upframe = upframe + 1
            self.ops[0].env['frame']   = upframe

            self.poutput('Writing %s %4i %6.1f %s'%(outfile,upframe,uptime,thefile))
            err = self.ops[0].writeascii( outfile, True )
            if err != '':
                return err
            outindex = outindex + 1
        # Foot of the loop

    # Erase used operands
    self.ops[0].clean()
    self.ops[1].clean()
    self.poutput('New frameset "%s" generated.'%outbody)
    return ''
#END changedwellframes()

# END script procedures

# Main:
err = changedwellframes()
if err != '':
    self.perror( err, traceback_war=False)
    # No special treatment in case of error
banner()



